@@ -22,6 +22,7 @@
 */
 
 #include <config.h>
+#include <string.h>
 #include <unistd.h>
 #include <sys/types.h>
 #include <sys/stat.h>
@@ -29,7 +30,6 @@
 
 #include <gnome.h>
 #include <bonobo.h>
-#include <gal/widgets/e-unicode.h>
 
 #include "htmlengine-edit-cut-and-paste.h"
 
@@ -47,6 +47,9 @@
 #include "template.h"
 
 static void smiley_cb (BonoboUIComponent *uic, GtkHTMLControlData *cd, const char *cname);
+static void paragraph_style_cb (BonoboUIComponent *uic, GtkHTMLControlData *cd, const char *cname);
+static void paragraph_align_cb (BonoboUIComponent *uic, GtkHTMLControlData *cd, const char *cname);
+static void font_style_cb (BonoboUIComponent *uic, GtkHTMLControlData *cd, const char *cname);
 
 static void
 undo_cb (BonoboUIComponent *uic, GtkHTMLControlData *cd, const char *cname)
@@ -215,7 +218,7 @@ file_dialog_destroy (GtkWidget *w, GtkHTMLControlData *cd)
 static void
 file_dialog_ok (GtkWidget *w, GtkHTMLControlData *cd)
 {
-	gchar *filename;
+	const gchar *filename;
 	gint fd;
 
 	filename = gtk_file_selection_get_filename (GTK_FILE_SELECTION (cd->file_dialog));
@@ -232,17 +235,13 @@ file_dialog_ok (GtkWidget *w, GtkHTMLControlData *cd)
 			gtk_html_write (tmp, stream, "<PRE>", 5);
 		}
 		while ((rb = read (fd, buffer, BUFFER_SIZE - 1)) > 0) {
-			gchar *native;
-
 			buffer [rb] = 0;
 
-			native = e_utf8_from_gtk_string (GTK_WIDGET (cd->html), buffer);
 			if (cd->file_html) {
-				gtk_html_write (tmp, stream, native, -1);
+				gtk_html_write (tmp, stream, buffer, -1);
 			} else {
-				html_engine_paste_text (cd->html->engine, native, g_utf8_strlen (native, -1));
+				html_engine_paste_text (cd->html->engine, buffer, g_utf8_strlen (buffer, -1));
 			}
-			g_free (native);
 		}
 		if (!cd->file_html) {
 			gtk_html_write (tmp, stream, "</PRE>", 6);
@@ -267,14 +266,13 @@ insert_file_dialog (GtkHTMLControlData *cd, gboolean html)
 	cd->file_dialog = gtk_file_selection_new (html ? _("Insert HTML file") : _("Insert text file"));
 	gtk_file_selection_set_filename (GTK_FILE_SELECTION (cd->file_dialog), "~/");
 
-	gtk_signal_connect_object (GTK_OBJECT (GTK_FILE_SELECTION (cd->file_dialog)->cancel_button),
-				   "clicked", GTK_SIGNAL_FUNC (gtk_widget_destroy), GTK_OBJECT (cd->file_dialog));
+	g_signal_connect_object (GTK_FILE_SELECTION (cd->file_dialog)->cancel_button,
+				 "clicked", G_CALLBACK (gtk_widget_destroy), GTK_OBJECT (cd->file_dialog),
+				 G_CONNECT_SWAPPED);
 
-	gtk_signal_connect (GTK_OBJECT (GTK_FILE_SELECTION (cd->file_dialog)->ok_button),
-			    "clicked", GTK_SIGNAL_FUNC (file_dialog_ok), cd);
+	g_signal_connect (GTK_FILE_SELECTION (cd->file_dialog)->ok_button, "clicked", G_CALLBACK (file_dialog_ok), cd);
 
-	gtk_signal_connect (GTK_OBJECT (cd->file_dialog), "destroy",
-			    GTK_SIGNAL_FUNC (file_dialog_destroy), cd);
+	g_signal_connect (cd->file_dialog, "destroy", G_CALLBACK (file_dialog_destroy), cd);
 
 	gtk_widget_show (cd->file_dialog);
 }
@@ -404,6 +402,40 @@ BonoboUIVerb verbs [] = {
 	BONOBO_UI_UNSAFE_VERB ("FormatParagraph", format_paragraph_cb),
 	BONOBO_UI_UNSAFE_VERB ("FormatPage", format_page_cb),
 
+	BONOBO_UI_UNSAFE_VERB ("HeadingNormal", paragraph_style_cb),
+	BONOBO_UI_UNSAFE_VERB ("HeadingPreformat", paragraph_style_cb),
+	BONOBO_UI_UNSAFE_VERB ("HeadingH1", paragraph_style_cb),
+	BONOBO_UI_UNSAFE_VERB ("HeadingH2", paragraph_style_cb),
+	BONOBO_UI_UNSAFE_VERB ("HeadingH3", paragraph_style_cb),
+	BONOBO_UI_UNSAFE_VERB ("HeadingH4", paragraph_style_cb),
+	BONOBO_UI_UNSAFE_VERB ("HeadingH5", paragraph_style_cb),
+	BONOBO_UI_UNSAFE_VERB ("HeadingH6", paragraph_style_cb),
+	BONOBO_UI_UNSAFE_VERB ("HeadingH1", paragraph_style_cb),
+	BONOBO_UI_UNSAFE_VERB ("HeadingAddress", paragraph_style_cb),
+	BONOBO_UI_UNSAFE_VERB ("HeadingBulletedList", paragraph_style_cb),
+	BONOBO_UI_UNSAFE_VERB ("HeadingRomanList", paragraph_style_cb),
+	BONOBO_UI_UNSAFE_VERB ("HeadingNumberedList", paragraph_style_cb),
+	BONOBO_UI_UNSAFE_VERB ("HeadingAlphabeticalList", paragraph_style_cb),
+
+	BONOBO_UI_UNSAFE_VERB ("FontSizeNegTwo", font_style_cb),
+	BONOBO_UI_UNSAFE_VERB ("FontSizeNegOne", font_style_cb),
+	BONOBO_UI_UNSAFE_VERB ("FontSizeZero", font_style_cb),
+	BONOBO_UI_UNSAFE_VERB ("FontSizeOne", font_style_cb),
+	BONOBO_UI_UNSAFE_VERB ("FontSizeTwo", font_style_cb),
+	BONOBO_UI_UNSAFE_VERB ("FontSizeThree", font_style_cb),
+	BONOBO_UI_UNSAFE_VERB ("FontSizeFour", font_style_cb),
+	BONOBO_UI_UNSAFE_VERB ("FormatBold", font_style_cb),
+	BONOBO_UI_UNSAFE_VERB ("FormatItalic", font_style_cb),
+	BONOBO_UI_UNSAFE_VERB ("FormatUnderline", font_style_cb),
+	BONOBO_UI_UNSAFE_VERB ("FormatStrikeout", font_style_cb),
+	BONOBO_UI_UNSAFE_VERB ("FormatFixed", font_style_cb),
+	BONOBO_UI_UNSAFE_VERB ("FormatSubscript", font_style_cb),
+	BONOBO_UI_UNSAFE_VERB ("FormatSuperscript", font_style_cb),
+
+	BONOBO_UI_UNSAFE_VERB ("AlignLeft", paragraph_align_cb),
+	BONOBO_UI_UNSAFE_VERB ("AlignCenter", paragraph_align_cb),
+	BONOBO_UI_UNSAFE_VERB ("AlignRight", paragraph_align_cb),
+
 	BONOBO_UI_VERB_END
 };
 
@@ -411,20 +443,20 @@ static struct {
 	GtkHTMLFontStyle style;
 	const gchar *verb;
 } font_style_assoc[] = {
-	{GTK_HTML_FONT_STYLE_SIZE_1, "/commands/FontSizeNegTwo"},
-	{GTK_HTML_FONT_STYLE_SIZE_2, "/commands/FontSizeNegOne"},
-	{GTK_HTML_FONT_STYLE_SIZE_3, "/commands/FontSizeZero"},
-	{GTK_HTML_FONT_STYLE_SIZE_4, "/commands/FontSizeOne"},
-	{GTK_HTML_FONT_STYLE_SIZE_5, "/commands/FontSizeTwo"},
-	{GTK_HTML_FONT_STYLE_SIZE_6, "/commands/FontSizeThree"},
-	{GTK_HTML_FONT_STYLE_SIZE_7, "/commands/FontSizeFour"},
-	{GTK_HTML_FONT_STYLE_BOLD,    "/commands/FormatBold"},
-	{GTK_HTML_FONT_STYLE_ITALIC, "/commands/FormatItalic"},
-	{GTK_HTML_FONT_STYLE_UNDERLINE, "/commands/FormatUnderline"},
-	{GTK_HTML_FONT_STYLE_STRIKEOUT, "/commands/FormatStrikeout"},
-	{GTK_HTML_FONT_STYLE_FIXED, "/commands/FormatFixed"},
-	{GTK_HTML_FONT_STYLE_SUBSCRIPT, "/commands/FormatSubscript"},
-	{GTK_HTML_FONT_STYLE_SUBSCRIPT, "/commands/FormatSuperscript"},
+	{GTK_HTML_FONT_STYLE_SIZE_1, "FontSizeNegTwo"},
+	{GTK_HTML_FONT_STYLE_SIZE_2, "FontSizeNegOne"},
+	{GTK_HTML_FONT_STYLE_SIZE_3, "FontSizeZero"},
+	{GTK_HTML_FONT_STYLE_SIZE_4, "FontSizeOne"},
+	{GTK_HTML_FONT_STYLE_SIZE_5, "FontSizeTwo"},
+	{GTK_HTML_FONT_STYLE_SIZE_6, "FontSizeThree"},
+	{GTK_HTML_FONT_STYLE_SIZE_7, "FontSizeFour"},
+	{GTK_HTML_FONT_STYLE_BOLD,    "FormatBold"},
+	{GTK_HTML_FONT_STYLE_ITALIC, "FormatItalic"},
+	{GTK_HTML_FONT_STYLE_UNDERLINE, "FormatUnderline"},
+	{GTK_HTML_FONT_STYLE_STRIKEOUT, "FormatStrikeout"},
+	{GTK_HTML_FONT_STYLE_FIXED, "FormatFixed"},
+	{GTK_HTML_FONT_STYLE_SUBSCRIPT, "FormatSubscript"},
+	{GTK_HTML_FONT_STYLE_SUBSCRIPT, "FormatSuperscript"},
 	{0, NULL}
 };
 
@@ -432,19 +464,19 @@ static struct {
 	GtkHTMLParagraphStyle style;
 	const gchar *verb;
 } paragraph_style_assoc[] = {
-	{GTK_HTML_PARAGRAPH_STYLE_NORMAL, "/commands/HeadingNormal"},
-	{GTK_HTML_PARAGRAPH_STYLE_H1, "/commands/HeadingH1"},
-	{GTK_HTML_PARAGRAPH_STYLE_H2, "/commands/HeadingH2"},
-	{GTK_HTML_PARAGRAPH_STYLE_H3, "/commands/HeadingH3"},
-	{GTK_HTML_PARAGRAPH_STYLE_H4, "/commands/HeadingH4"},
-	{GTK_HTML_PARAGRAPH_STYLE_H5, "/commands/HeadingH5"},
-	{GTK_HTML_PARAGRAPH_STYLE_H6, "/commands/HeadingH6"},
-	{GTK_HTML_PARAGRAPH_STYLE_ADDRESS, "/commands/HeadingAddress"},
-	{GTK_HTML_PARAGRAPH_STYLE_PRE, "/commands/HeadingPreformat"},
-	{GTK_HTML_PARAGRAPH_STYLE_ITEMDOTTED, "/commands/HeadingBulletedList"},
-	{GTK_HTML_PARAGRAPH_STYLE_ITEMROMAN, "/commands/HeadingRomanList"},
-	{GTK_HTML_PARAGRAPH_STYLE_ITEMDIGIT, "/commands/HeadingNumberedList"},
-	{GTK_HTML_PARAGRAPH_STYLE_ITEMALPHA, "/commands/HeadingAlphabeticalList"},
+	{GTK_HTML_PARAGRAPH_STYLE_NORMAL, "HeadingNormal"},
+	{GTK_HTML_PARAGRAPH_STYLE_H1, "HeadingH1"},
+	{GTK_HTML_PARAGRAPH_STYLE_H2, "HeadingH2"},
+	{GTK_HTML_PARAGRAPH_STYLE_H3, "HeadingH3"},
+	{GTK_HTML_PARAGRAPH_STYLE_H4, "HeadingH4"},
+	{GTK_HTML_PARAGRAPH_STYLE_H5, "HeadingH5"},
+	{GTK_HTML_PARAGRAPH_STYLE_H6, "HeadingH6"},
+	{GTK_HTML_PARAGRAPH_STYLE_ADDRESS, "HeadingAddress"},
+	{GTK_HTML_PARAGRAPH_STYLE_PRE, "HeadingPreformat"},
+	{GTK_HTML_PARAGRAPH_STYLE_ITEMDOTTED, "HeadingBulletedList"},
+	{GTK_HTML_PARAGRAPH_STYLE_ITEMROMAN, "HeadingRomanList"},
+	{GTK_HTML_PARAGRAPH_STYLE_ITEMDIGIT, "HeadingNumberedList"},
+	{GTK_HTML_PARAGRAPH_STYLE_ITEMALPHA, "HeadingAlphabeticalList"},
 	{0, NULL}
 };
 
@@ -452,59 +484,36 @@ static struct {
 	GtkHTMLParagraphAlignment style;
 	const gchar *verb;
 } paragraph_align_assoc[] = {
-	{GTK_HTML_PARAGRAPH_ALIGNMENT_LEFT, "/commands/AlignLeft"},
-	{GTK_HTML_PARAGRAPH_ALIGNMENT_CENTER, "/commands/AlignCenter"},
-	{GTK_HTML_PARAGRAPH_ALIGNMENT_RIGHT, "/commands/AlignRight"},
+	{GTK_HTML_PARAGRAPH_ALIGNMENT_LEFT, "AlignLeft"},
+	{GTK_HTML_PARAGRAPH_ALIGNMENT_CENTER, "AlignCenter"},
+	{GTK_HTML_PARAGRAPH_ALIGNMENT_RIGHT, "AlignRight"},
 	{0, NULL}
 };
 
 static void 
-paragraph_align_cb (BonoboUIComponent           *component,
-		    const char                  *path,
-		    Bonobo_UIComponent_EventType type,
-		    const char                  *state,
-		    gpointer                     user_data)
-     
+paragraph_align_cb (BonoboUIComponent *uic, GtkHTMLControlData *cd, const char *cname)
 {
-	GtkHTMLControlData *cd = (GtkHTMLControlData *)user_data;
 	int i;
 
-	if (cd->block_font_style_change || !atoi(state))
-		return;
-
 	/* g_warning ("wowee %s :: %s", path, state); */
 	for (i = 0; paragraph_align_assoc[i].verb != NULL; i++) {
-		if (!strcmp (path, paragraph_align_assoc[i].verb + 10)) {
+		if (!strcmp (cname, paragraph_align_assoc[i].verb)) {
 			/* g_warning ("setting style to: %s", 
 			   paragraph_align_assoc[i].verb); */
 
-			gtk_html_set_paragraph_alignment (cd->html, 
-							  paragraph_align_assoc[i].style);
+			gtk_html_set_paragraph_alignment (cd->html, paragraph_align_assoc[i].style);
 			return;
 		}
 	}
 }
 
-static void 
-paragraph_style_cb (BonoboUIComponent           *component,
-		    const char                  *path,
-		    Bonobo_UIComponent_EventType type,
-		    const char                  *state,
-		    gpointer                     user_data)
-     
+static void
+paragraph_style_cb (BonoboUIComponent *uic, GtkHTMLControlData *cd, const char *cname)
 {
-	GtkHTMLControlData *cd = (GtkHTMLControlData *)user_data;
 	int i;
 	
-	if (cd->block_font_style_change)
-		return;
-
-	/* g_warning ("wowee %s :: %s", path, state); */
-	if (!atoi(state))
-		return; 
-
 	for (i = 0; paragraph_style_assoc[i].verb != NULL; i++) {
-		if (!strcmp (path, paragraph_style_assoc[i].verb + 10)) {
+		if (!strcmp (cname, paragraph_style_assoc[i].verb)) {
 			/* g_warning ("setting style to: %s", 
 			   paragraph_style_assoc[i].verb); */
 
@@ -515,152 +524,25 @@ paragraph_style_cb (BonoboUIComponent           *component,
 }
 
 static void 
-font_size_cb (BonoboUIComponent           *component,
-	       const char                  *path,
-	       Bonobo_UIComponent_EventType type,
-	       const char                  *state,
-	       gpointer                     user_data)
-
+font_style_cb (BonoboUIComponent *uic, GtkHTMLControlData *cd, const char *cname)
 {
-	GtkHTMLControlData *cd = (GtkHTMLControlData *)user_data;
 	int i;
 	
-	if (cd->block_font_style_change)
-		return;
-
 	/* g_warning ("wowee %s :: %s", path, state); */
 	for (i = 0; font_style_assoc[i].verb != NULL; i++) {
-		if (!strcmp (path, font_style_assoc[i].verb + 10)) {
-			if (font_style_assoc[i].style > GTK_HTML_FONT_STYLE_MAX) {
-				if (atoi (state)) {
-					gtk_html_set_font_style (cd->html, ~0,
-								 font_style_assoc[i].style); 
-					
-					
-				} else {
-					gint mask = ~0 & ~font_style_assoc[i].style;
-					
-					gtk_html_set_font_style (cd->html, 
-								 mask, 0);
-				}			    
-				
-			} else { 
-				if (atoi (state))
-					gtk_html_set_font_style (cd->html, 
-								 GTK_HTML_FONT_STYLE_MAX 
-								 & ~GTK_HTML_FONT_STYLE_SIZE_MASK, 
-								 font_style_assoc[i].style);
-			}
+		if (!strcmp (cname, font_style_assoc[i].verb)) {
+			if (font_style_assoc[i].style > GTK_HTML_FONT_STYLE_MAX)
+				gtk_html_set_font_style (cd->html, ~0,
+							 font_style_assoc[i].style); 
+			else
+				gtk_html_set_font_style (cd->html, 
+							 GTK_HTML_FONT_STYLE_MAX 
+							 & ~GTK_HTML_FONT_STYLE_SIZE_MASK, 
+							 font_style_assoc[i].style);
 		}
 	}
 }
 
-static void 
-menubar_update_font_style (GtkWidget *widget, 
-			   GtkHTMLFontStyle style, 
-			   GtkHTMLControlData *cd)
-{
-	BonoboUIComponent *uic;
-	int size, i;
-	CORBA_Environment ev;
-	
-	CORBA_exception_init (&ev);
-
-	uic = bonobo_control_get_ui_component (cd->control);
-	
-	g_return_if_fail (uic != NULL);
-
-	size = ((int)style) & GTK_HTML_FONT_STYLE_SIZE_MASK;
-
-	cd->block_font_style_change++;
-	for (i = 0; font_style_assoc[i].verb != NULL; i++) {
-		/* deal with sizes */
-		if (size == font_style_assoc[i].style) {
-			bonobo_ui_component_set_prop (uic, font_style_assoc[i].verb, 
-						      "state", "1", &ev);
-		}
-
-		/* deal with styles */
-		if (font_style_assoc[i].style > GTK_HTML_FONT_STYLE_SIZE_MASK) {
-			char *state = (int)font_style_assoc[i].style & style ? "1" : "0";
-
-			bonobo_ui_component_set_prop (uic, font_style_assoc[i].verb,
-						      "state", state, &ev);
-		}
-	} 
-	cd->block_font_style_change--;
-
-	CORBA_exception_free (&ev);
-}
-
-static void
-menubar_update_paragraph_alignment (GtkHTML *html, 
-				    GtkHTMLParagraphAlignment style, 
-				    GtkHTMLControlData *cd)
-{
-	BonoboUIComponent *uic;
-	char *path = NULL;
-	int i;
-	
-	uic = bonobo_control_get_ui_component (cd->control);
-
-	g_return_if_fail (uic != NULL);
-
-	for (i = 0; paragraph_align_assoc[i].verb != NULL; i++) {
-		if (paragraph_align_assoc[i].style == style) {
-			path = (char *)paragraph_align_assoc[i].verb;
-			break;
-		}
-	}
-
-	if (path) {
-		CORBA_Environment ev;
-
-		CORBA_exception_init (&ev);
-		cd->block_font_style_change ++;
-		bonobo_ui_component_set_prop (uic, path,
-					      "state", "1", &ev);
-		cd->block_font_style_change --;
-		CORBA_exception_free (&ev);	
-	} else {
-		g_warning ("Unknown Paragraph Alignment");
-	}
-}
-
-static void
-menubar_update_paragraph_style (GtkHTML *html, 
-				GtkHTMLParagraphStyle style, 
-				GtkHTMLControlData *cd)
-{
-	BonoboUIComponent *uic;
-	const char *path = NULL;
-	int i;
-
-	uic = bonobo_control_get_ui_component (cd->control);
-
-	g_return_if_fail (uic != NULL);
-
-	for (i = 0; paragraph_style_assoc[i].verb != NULL; i++) {
-		if (paragraph_style_assoc[i].style == style) {
-			path = paragraph_style_assoc[i].verb;
-			break;
-		}
-	}
-
-	if (path) {
-		CORBA_Environment ev;
-
-		CORBA_exception_init (&ev);
-		cd->block_font_style_change ++;
-		bonobo_ui_component_set_prop (uic, path,
-					      "state", "1", &ev);
-		cd->block_font_style_change --;
-		CORBA_exception_free (&ev);	
-	} else {
-		g_warning ("Unknown Paragraph Style");
-	}
-}
-
 void
 menubar_update_format (GtkHTMLControlData *cd)
 {
@@ -669,84 +551,85 @@ menubar_update_format (GtkHTMLControlData *cd)
 	gchar *sensitive;
 
 	uic = bonobo_control_get_ui_component (cd->control);
-
-	g_return_if_fail (uic != NULL);
-
-	sensitive = (cd->format_html ? "1" : "0");
-
-	CORBA_exception_init (&ev);
-
-	bonobo_ui_component_freeze (uic, &ev);
-
-	bonobo_ui_component_set_prop (uic, "/commands/InsertImage",
-				      "sensitive", sensitive, &ev);
-	bonobo_ui_component_set_prop (uic, "/commands/InsertLink",
-				      "sensitive", sensitive, &ev);
-	bonobo_ui_component_set_prop (uic, "/commands/InsertRule",
-				      "sensitive", sensitive, &ev);
-	bonobo_ui_component_set_prop (uic, "/commands/InsertTable",
-				      "sensitive", sensitive, &ev);
-	bonobo_ui_component_set_prop (uic, "/commands/InsertTemplate",
-				      "sensitive", sensitive, &ev);
-	bonobo_ui_component_set_prop (uic, "/commands/InsertTemplate",
-				      "sensitive", sensitive, &ev);
-	bonobo_ui_component_set_prop (uic, "/commands/InsertSmiley1",
-				      "sensitive", sensitive, &ev);
-	bonobo_ui_component_set_prop (uic, "/commands/InsertSmiley2",
-				      "sensitive", sensitive, &ev);
-	bonobo_ui_component_set_prop (uic, "/commands/InsertSmiley3",
-				      "sensitive", sensitive, &ev);
-	bonobo_ui_component_set_prop (uic, "/commands/InsertSmiley4",
-				      "sensitive", sensitive, &ev);
-	bonobo_ui_component_set_prop (uic, "/commands/InsertSmiley5",
-				      "sensitive", sensitive, &ev);
-	bonobo_ui_component_set_prop (uic, "/commands/InsertSmiley6",
-				      "sensitive", sensitive, &ev);
-	bonobo_ui_component_set_prop (uic, "/commands/InsertSmiley8",
-				      "sensitive", sensitive, &ev);
-	bonobo_ui_component_set_prop (uic, "/commands/InsertSmiley9",
-				      "sensitive", sensitive, &ev);
-	bonobo_ui_component_set_prop (uic, "/commands/InsertSmiley10",
-				      "sensitive", sensitive, &ev);
-	bonobo_ui_component_set_prop (uic, "/commands/InsertSmiley11",
-				      "sensitive", sensitive, &ev);
-
-	bonobo_ui_component_set_prop (uic, "/commands/FormatBold",
-				      "sensitive", sensitive, &ev);
-	bonobo_ui_component_set_prop (uic, "/commands/FormatItalic",
-				      "sensitive", sensitive, &ev);
-	bonobo_ui_component_set_prop (uic, "/commands/FormatUnderline",
-				      "sensitive", sensitive, &ev);
-	bonobo_ui_component_set_prop (uic, "/commands/FormatStrikeout",
-				      "sensitive", sensitive, &ev);
-	bonobo_ui_component_set_prop (uic, "/commands/FormatPlain",
-				      "sensitive", sensitive, &ev);
 	
-	bonobo_ui_component_set_prop (uic, "/commands/AlignLeft",
-				      "sensitive", sensitive, &ev);		
-	bonobo_ui_component_set_prop (uic, "/commands/AlignRight",
-				      "sensitive", sensitive, &ev);	
-	bonobo_ui_component_set_prop (uic, "/commands/AlignCenter",
-				      "sensitive", sensitive, &ev);	
+	if (uic != CORBA_OBJECT_NIL && bonobo_ui_component_get_container (uic) != CORBA_OBJECT_NIL) {
 
-	bonobo_ui_component_set_prop (uic, "/commands/HeadingH1",
-				      "sensitive", sensitive, &ev);	
-	bonobo_ui_component_set_prop (uic, "/commands/HeadingH2",
-				      "sensitive", sensitive, &ev);	
-	bonobo_ui_component_set_prop (uic, "/commands/HeadingH3",
-				      "sensitive", sensitive, &ev);
-	bonobo_ui_component_set_prop (uic, "/commands/HeadingH4",
-				      "sensitive", sensitive, &ev);
-	bonobo_ui_component_set_prop (uic, "/commands/HeadingH5",
-				      "sensitive", sensitive, &ev);
-	bonobo_ui_component_set_prop (uic, "/commands/HeadingH6",
-				      "sensitive", sensitive, &ev);
-	bonobo_ui_component_set_prop (uic, "/commands/HeadingAddress",
-				      "sensitive", sensitive, &ev);
+		sensitive = (cd->format_html ? "1" : "0");
 
-	bonobo_ui_component_thaw (uic, &ev);	
+		CORBA_exception_init (&ev);
 
-	CORBA_exception_free (&ev);	
+		bonobo_ui_component_freeze (uic, &ev);
+
+		bonobo_ui_component_set_prop (uic, "/commands/InsertImage",
+					      "sensitive", sensitive, &ev);
+		bonobo_ui_component_set_prop (uic, "/commands/InsertLink",
+					      "sensitive", sensitive, &ev);
+		bonobo_ui_component_set_prop (uic, "/commands/InsertRule",
+					      "sensitive", sensitive, &ev);
+		bonobo_ui_component_set_prop (uic, "/commands/InsertTable",
+					      "sensitive", sensitive, &ev);
+		bonobo_ui_component_set_prop (uic, "/commands/InsertTemplate",
+					      "sensitive", sensitive, &ev);
+		bonobo_ui_component_set_prop (uic, "/commands/InsertTemplate",
+					      "sensitive", sensitive, &ev);
+		bonobo_ui_component_set_prop (uic, "/commands/InsertSmiley1",
+					      "sensitive", sensitive, &ev);
+		bonobo_ui_component_set_prop (uic, "/commands/InsertSmiley2",
+					      "sensitive", sensitive, &ev);
+		bonobo_ui_component_set_prop (uic, "/commands/InsertSmiley3",
+					      "sensitive", sensitive, &ev);
+		bonobo_ui_component_set_prop (uic, "/commands/InsertSmiley4",
+					      "sensitive", sensitive, &ev);
+		bonobo_ui_component_set_prop (uic, "/commands/InsertSmiley5",
+					      "sensitive", sensitive, &ev);
+		bonobo_ui_component_set_prop (uic, "/commands/InsertSmiley6",
+					      "sensitive", sensitive, &ev);
+		bonobo_ui_component_set_prop (uic, "/commands/InsertSmiley8",
+					      "sensitive", sensitive, &ev);
+		bonobo_ui_component_set_prop (uic, "/commands/InsertSmiley9",
+					      "sensitive", sensitive, &ev);
+		bonobo_ui_component_set_prop (uic, "/commands/InsertSmiley10",
+					      "sensitive", sensitive, &ev);
+		bonobo_ui_component_set_prop (uic, "/commands/InsertSmiley11",
+					      "sensitive", sensitive, &ev);
+
+		bonobo_ui_component_set_prop (uic, "/commands/FormatBold",
+					      "sensitive", sensitive, &ev);
+		bonobo_ui_component_set_prop (uic, "/commands/FormatItalic",
+					      "sensitive", sensitive, &ev);
+		bonobo_ui_component_set_prop (uic, "/commands/FormatUnderline",
+					      "sensitive", sensitive, &ev);
+		bonobo_ui_component_set_prop (uic, "/commands/FormatStrikeout",
+					      "sensitive", sensitive, &ev);
+		bonobo_ui_component_set_prop (uic, "/commands/FormatPlain",
+					      "sensitive", sensitive, &ev);
+	
+		bonobo_ui_component_set_prop (uic, "/commands/AlignLeft",
+					      "sensitive", sensitive, &ev);		
+		bonobo_ui_component_set_prop (uic, "/commands/AlignRight",
+					      "sensitive", sensitive, &ev);	
+		bonobo_ui_component_set_prop (uic, "/commands/AlignCenter",
+					      "sensitive", sensitive, &ev);	
+
+		bonobo_ui_component_set_prop (uic, "/commands/HeadingH1",
+					      "sensitive", sensitive, &ev);	
+		bonobo_ui_component_set_prop (uic, "/commands/HeadingH2",
+					      "sensitive", sensitive, &ev);	
+		bonobo_ui_component_set_prop (uic, "/commands/HeadingH3",
+					      "sensitive", sensitive, &ev);
+		bonobo_ui_component_set_prop (uic, "/commands/HeadingH4",
+					      "sensitive", sensitive, &ev);
+		bonobo_ui_component_set_prop (uic, "/commands/HeadingH5",
+					      "sensitive", sensitive, &ev);
+		bonobo_ui_component_set_prop (uic, "/commands/HeadingH6",
+					      "sensitive", sensitive, &ev);
+		bonobo_ui_component_set_prop (uic, "/commands/HeadingAddress",
+					      "sensitive", sensitive, &ev);
+
+		bonobo_ui_component_thaw (uic, &ev);	
+
+		CORBA_exception_free (&ev);
+	}
 }
 
 void
@@ -763,7 +646,7 @@ menubar_set_languages (GtkHTMLControlData *cd, const gchar *lstr)
 	cd->block_language_changes = TRUE;
 	for (i = 0; i < cd->languages->_length; i ++) {
 		enabled = strstr (lstr, cd->languages->_buffer [i].abrev) != NULL;
-		g_string_sprintf (str, "/commands/SpellLanguage%d", i + 1);
+		g_string_printf (str, "/commands/SpellLanguage%d", i + 1);
 		bonobo_ui_component_set_prop (cd->uic, str->str, "state", enabled ? "1" : "0", NULL);
 	}
 	cd->block_language_changes = FALSE;
@@ -811,30 +694,8 @@ menubar_setup (BonoboUIComponent  *uic,
 	g_return_if_fail (GTK_IS_HTML (cd->html));
 	g_return_if_fail (BONOBO_IS_UI_COMPONENT (uic));
 
-	gtk_signal_connect (GTK_OBJECT (cd->html), "current_paragraph_style_changed",
-			    GTK_SIGNAL_FUNC (menubar_update_paragraph_style), cd);
-
-	gtk_signal_connect (GTK_OBJECT (cd->html), "current_paragraph_alignment_changed",
-			    GTK_SIGNAL_FUNC (menubar_update_paragraph_alignment), cd);
-
-	gtk_signal_connect (GTK_OBJECT (cd->html), "insertion_font_style_changed",
-			    GTK_SIGNAL_FUNC (menubar_update_font_style), cd);
-
 	bonobo_ui_component_add_verb_list_with_data (uic, verbs, cd);
 
-	for (i = 0; paragraph_style_assoc[i].verb != NULL; i++) {
-		bonobo_ui_component_add_listener (uic, paragraph_style_assoc[i].verb + 10, paragraph_style_cb, cd);
-	}
-
-	for (i = 0; paragraph_align_assoc[i].verb != NULL; i++) {
-		bonobo_ui_component_add_listener (uic, paragraph_align_assoc[i].verb + 10, paragraph_align_cb, cd);
-	}
-
-	for (i = 0; font_style_assoc[i].verb != NULL; i++) {
-		bonobo_ui_component_add_listener (uic, font_style_assoc[i].verb + 10, font_size_cb, cd);
-	}
-
-	bonobo_ui_util_set_ui (uic, GNOMEDATADIR,
-			       "GNOME_GtkHTML_Editor.xml",
-			       "GNOME_GtkHTML_Editor");
+	bonobo_ui_util_set_ui (uic, GTKHTML_DATADIR, "GNOME_GtkHTML_Editor.xml", "GNOME_GtkHTML_Editor", NULL);
+	menubar_update_format (cd);
 }
