@@ -24,8 +24,6 @@
 #include <config.h>
 #endif
 
-#include <gal/widgets/e-unicode.h>
-
 #include "gtkhtml.h"
 
 #include "htmlclueflow.h"
@@ -52,7 +50,6 @@
 #include "table.h"
 #include "text.h"
 
-/* #define DEBUG */
 #ifdef DEBUG
 #include "gtkhtmldebug.h"
 #endif
@@ -259,7 +256,7 @@ show_prop_dialog (GtkHTMLControlData *cd, GtkHTMLEditPropertyType start)
 static void
 prop_dialog (GtkWidget *mi, GtkHTMLControlData *cd)
 {
-	show_prop_dialog (cd, GPOINTER_TO_INT (gtk_object_get_data (GTK_OBJECT (mi), "type")));
+	show_prop_dialog (cd, GPOINTER_TO_INT (g_object_get_data (G_OBJECT (mi), "type")));
 }
 
 static void
@@ -283,11 +280,11 @@ link_prop_dialog (GtkWidget *mi, GtkHTMLControlData *cd)
 static void
 spell_suggest (GtkWidget *mi, GtkHTMLControlData *cd)
 {
-	HTMLEngine *e = cd->html->engine;
+	/* HTMLEngine *e = cd->html->engine;
 
-	/* gtk_signal_emit_by_name (GTK_OBJECT (cd->html), "spell_suggestion_request",
+	   gtk_signal_emit_by_name (GTK_OBJECT (cd->html), "spell_suggestion_request",
 	   e->spell_checker, html_engine_get_word (e)); */
-	spell_suggestion_request (cd->html, html_engine_get_spell_word (e), cd);
+	spell_suggestion_request (cd->html, cd);
 }
 
 static void
@@ -345,39 +342,36 @@ insert_html (GtkWidget *mi, GtkHTMLControlData *cd)
 #endif
 
 #define ADD_ITEM_BASE(f,t) \
-                gtk_object_set_data (GTK_OBJECT (menuitem), "type", GINT_TO_POINTER (GTK_HTML_EDIT_PROPERTY_ ## t)); \
-		gtk_menu_append (GTK_MENU (menu), menuitem); \
+                g_object_set_data (G_OBJECT (menuitem), "type", GINT_TO_POINTER (GTK_HTML_EDIT_PROPERTY_ ## t)); \
+		gtk_menu_shell_append (GTK_MENU_SHELL (menu), menuitem); \
 		gtk_widget_show (menuitem); \
-		gtk_signal_connect (GTK_OBJECT (menuitem), "activate", GTK_SIGNAL_FUNC (f), cd); \
+		g_signal_connect (menuitem, "activate", G_CALLBACK (f), cd); \
                 (*items)++; items_sep++
 
-#define ADD_ITEM(l,f,t) \
-		menuitem = gtk_menu_item_new_with_label (l); \
-                ADD_ITEM_BASE (f,t)
-
 #define ADD_ITEM_SENSITIVE(l,f,t,s) \
 		menuitem = gtk_menu_item_new_with_label (l); \
                 ADD_ITEM_BASE (f,t); \
                 gtk_widget_set_sensitive (menuitem, s);
 
-#define ADD_ITEM_UTF8(l,f,t) \
-                menuitem = e_utf8_gtk_menu_item_new_with_label (GTK_MENU (menu), l); \
+#define ADD_ITEM(l,f,t) \
+		menuitem = gtk_menu_item_new_with_label (l); \
                 ADD_ITEM_BASE (f,t)
 
 #define ADD_SEP \
         if (items_sep) { \
                 menuitem = gtk_menu_item_new (); \
-                gtk_menu_append (GTK_MENU (menu), menuitem); \
+                gtk_menu_shell_append (GTK_MENU_SHELL (menu), menuitem); \
                 gtk_widget_show (menuitem); \
 		items_sep = 0; \
         }
 
 #define ADD_PROP(x) \
-        cd->properties_types = g_list_append (cd->properties_types, GINT_TO_POINTER (GTK_HTML_EDIT_PROPERTY_ ## x))
+        cd->properties_types = g_list_append (cd->properties_types, GINT_TO_POINTER (GTK_HTML_EDIT_PROPERTY_ ## x)); \
+        (*props) ++;
 
 #define SUBMENU(l) \
 		        menuitem = gtk_menu_item_new_with_label (_(l)); \
-			gtk_menu_append (GTK_MENU (menu), menuitem); \
+			gtk_menu_shell_append (GTK_MENU_SHELL (menu), menuitem); \
 			gtk_widget_show (menuitem); \
 			(*items)++; items_sep++; \
 			submenu = gtk_menu_new (); \
@@ -390,7 +384,7 @@ insert_html (GtkWidget *mi, GtkHTMLControlData *cd)
 			menu = menuparent
 
 static GtkWidget *
-prepare_properties_and_menu (GtkHTMLControlData *cd, guint *items)
+prepare_properties_and_menu (GtkHTMLControlData *cd, guint *items, guint *props)
 {
 	HTMLEngine *e = cd->html->engine;
 	HTMLObject *obj;
@@ -402,6 +396,8 @@ prepare_properties_and_menu (GtkHTMLControlData *cd, guint *items)
 
 	obj  = cd->html->engine->cursor->object;
 	menu = gtk_menu_new ();
+	*items = 0;
+	*props = 0;
 
 	if (cd->properties_types) {
 		g_list_free (cd->properties_types);
@@ -438,27 +434,21 @@ prepare_properties_and_menu (GtkHTMLControlData *cd, guint *items)
 
 	if (!active && obj && html_object_is_text (obj)
 	    && !html_engine_spell_word_is_valid (e)) {
-		gchar *spell, *word, *check_utf8, *add_utf8, *ignore_utf8, *ignore, *add;
+		gchar *spell, *word, *ignore, *add;
 
 		word   = html_engine_get_spell_word (e);
-		check_utf8 = e_utf8_from_locale_string (_("Check '%s' spelling..."));
-		spell  = g_strdup_printf (check_utf8, word);
-		g_free (check_utf8);
-		add_utf8 = e_utf8_from_locale_string (_("Add '%s' to dictionary"));
-		add    = g_strdup_printf (add_utf8, word);
-		g_free (add_utf8);
-		ignore_utf8 = e_utf8_from_locale_string (_("Ignore '%s'"));
-		ignore = g_strdup_printf (ignore_utf8, word);
-		g_free (ignore_utf8);
+		spell  = g_strdup_printf (_("Check '%s' spelling..."), word);
+		add    = g_strdup_printf (_("Add '%s' to dictionary"), word);
+		ignore = g_strdup_printf (_("Ignore '%s'"), word);
 		ADD_SEP;
 		SUBMENU (N_("Spell checker"));
 		if (cd->has_spell_control) {
-			ADD_ITEM_UTF8 (spell, spell_check_cb, NONE);
+			ADD_ITEM (spell, spell_check_cb, NONE);
 		} else {
 			ADD_ITEM (_("Suggest word"), spell_suggest, NONE);
 		}
-		ADD_ITEM_UTF8 (add, spell_add, NONE);
-		ADD_ITEM_UTF8 (ignore, spell_ignore, NONE);
+		ADD_ITEM (add, spell_add, NONE);
+		ADD_ITEM (ignore, spell_ignore, NONE);
 		END_SUBMENU;
 
 		g_free (spell);
@@ -495,11 +485,8 @@ prepare_properties_and_menu (GtkHTMLControlData *cd, guint *items)
 			ADD_ITEM (_("Paragraph..."), prop_dialog, PARAGRAPH);
 			ADD_PROP (PARAGRAPH);
 			break;
-		case HTML_TYPE_TABLE:
-			ADD_SEP;
-			ADD_PROP (TABLE);
-			ADD_ITEM (_("Table..."), prop_dialog, TABLE);
 		default:
+			;
 		}
 		if (obj->parent && obj->parent->parent && HTML_IS_TABLE_CELL (obj->parent->parent)) {
 			if (cd->format_html) {
@@ -535,7 +522,11 @@ prepare_properties_and_menu (GtkHTMLControlData *cd, guint *items)
 		ADD_ITEM (_("Page..."), prop_dialog, BODY);
 	}
 
-	gtk_widget_show (menu);
+	if (*items == 0) {
+		gtk_object_sink (GTK_OBJECT (menu));
+		menu = NULL;
+	} else
+		gtk_widget_show (menu);
 
 	return menu;
 }
@@ -544,19 +535,19 @@ gint
 popup_show (GtkHTMLControlData *cd, GdkEventButton *event)
 {
 	GtkWidget *menu;
-	guint items = 0;
+	guint items, props;
+
+	menu = prepare_properties_and_menu (cd, &items, &props);
 
-	menu = prepare_properties_and_menu (cd, &items);
 	if (items)
-		gtk_menu_popup (GTK_MENU (menu), NULL, NULL, NULL, NULL, 
+		gtk_menu_popup (GTK_MENU (menu), NULL, NULL, NULL, NULL,
 				event ? event->button : 0, event ? event->time : 0);
-	gtk_widget_unref (menu);
 
 	return (items > 0);
 }
 
 static void
-set_position (GtkMenu *menu, gint *x, gint *y, gpointer data)
+set_position (GtkMenu *menu, gint *x, gint *y, gboolean *push_in, gpointer data)
 {
 	GtkHTMLControlData *cd = (GtkHTMLControlData *) data;
 	HTMLEngine *e = cd->html->engine;
@@ -572,13 +563,12 @@ gint
 popup_show_at_cursor (GtkHTMLControlData *cd)
 {
 	GtkWidget *menu;
-	guint items = 0;
+	guint items, props;
 
-	menu = prepare_properties_and_menu (cd, &items);
+	menu = prepare_properties_and_menu (cd, &items, &props);
 	gtk_widget_show (menu);
 	if (items)
 		gtk_menu_popup (GTK_MENU (menu), NULL, NULL, set_position, cd, 0, 0);
-	gtk_widget_unref (menu);
 
 	return (items > 0);
 }
@@ -586,9 +576,11 @@ popup_show_at_cursor (GtkHTMLControlData *cd)
 void
 property_dialog_show (GtkHTMLControlData *cd)
 {
-	guint items = 0;
+	GtkWidget *menu;
+	guint items, props;
 
-	gtk_widget_unref (prepare_properties_and_menu (cd, &items));
-	if (items)
+	menu = prepare_properties_and_menu (cd, &items, &props);
+	gtk_object_sink (GTK_OBJECT (menu));
+	if (props)
 		show_prop_dialog (cd, GTK_HTML_EDIT_PROPERTY_NONE);
 }
