@@ -28,52 +28,41 @@
 #include "htmlengine.h"
 
 struct _GtkHTMLReplaceAskDialog {
-	GnomeDialog *dialog;
-	HTMLEngine  *engine;
+	GtkDialog  *dialog;
+	HTMLEngine *engine;
 };
 
 typedef struct _GtkHTMLReplaceAskDialog GtkHTMLReplaceAskDialog;
 
 struct _GtkHTMLReplaceDialog {
-	GnomeDialog *dialog;
+	GtkDialog   *dialog;
 	GtkHTML     *html;
 	GtkWidget   *entry_search;
 	GtkWidget   *entry_replace;
 	GtkWidget   *backward;
 	GtkWidget   *case_sensitive;
 
-	GtkHTMLReplaceAskDialog *ask_dialog;
+	GtkHTMLControlData *cd;
 };
 
 static void
-replace_do (GtkHTMLReplaceAskDialog *d, HTMLReplaceQueryAnswer answer)
+ask_dialog_response (GtkDialog *dialog, gint response_id, GtkHTMLReplaceAskDialog *d)
 {
-	gnome_dialog_close (d->dialog);
-	html_engine_replace_do (d->engine, answer);
-}
-
-static void
-replace_cb (GtkWidget *button, GtkHTMLReplaceAskDialog *d)
-{
-	replace_do (d, RQA_Replace);
-}
-
-static void
-replace_all_cb (GtkWidget *button, GtkHTMLReplaceAskDialog *d)
-{
-	replace_do (d, RQA_ReplaceAll);
-}
-
-static void
-next_cb (GtkWidget *button, GtkHTMLReplaceAskDialog *d)
-{
-	replace_do (d, RQA_Next);
-}
-
-static void
-cancel_cb (GtkWidget *button, GtkHTMLReplaceAskDialog *d)
-{
-	replace_do (d, RQA_Cancel);
+	switch (response_id) {
+	case GTK_RESPONSE_DELETE_EVENT:
+	case GTK_RESPONSE_CLOSE:
+		html_engine_replace_do (d->engine, RQA_Cancel);
+		break;
+	case 0: /* Replace */
+		html_engine_replace_do (d->engine, RQA_Replace);
+		break;
+	case 1: /* Replace All */
+		html_engine_replace_do (d->engine, RQA_ReplaceAll);
+		break;
+	case 2: /* Next */
+		html_engine_replace_do (d->engine, RQA_Next);
+		break;
+	}
 }
 
 static GtkHTMLReplaceAskDialog *
@@ -82,21 +71,16 @@ ask_dialog_new (HTMLEngine *e)
 	GtkHTMLReplaceAskDialog *d;
 
 	d = g_new (GtkHTMLReplaceAskDialog, 1);
-	d->dialog = GNOME_DIALOG (gnome_dialog_new (_("Replace confirmation"),
-						    _("Replace"), _("Replace all"), _("Next"),
-						    GNOME_STOCK_BUTTON_CANCEL, NULL));
+	d->dialog = GTK_DIALOG (gtk_dialog_new_with_buttons (_("Replace confirmation"), NULL, 0,
+							     _("Replace all"), 1,
+							     _("Next"), 2,
+							     GTK_STOCK_CLOSE, GTK_RESPONSE_CLOSE,
+							     _("Replace"), 0,
+							     NULL));
 	d->engine = e;
 
 	gnome_window_icon_set_from_file (GTK_WINDOW (d->dialog), ICONDIR "/search-and-replace-24.png");
-	gnome_dialog_button_connect (d->dialog, 0, replace_cb, d);
-	gnome_dialog_button_connect (d->dialog, 1, replace_all_cb, d);
-	gnome_dialog_button_connect (d->dialog, 2, next_cb, d);
-	gnome_dialog_button_connect (d->dialog, 3, cancel_cb, d);
-
-	gnome_dialog_close_hides (d->dialog, TRUE);
-	/* gnome_dialog_set_close (d->dialog, TRUE); */
-
-	gnome_dialog_set_default (d->dialog, 0);
+	g_signal_connect (d->dialog, "response", G_CALLBACK (ask_dialog_response), d);
 
 	return d;
 }
@@ -104,50 +88,71 @@ ask_dialog_new (HTMLEngine *e)
 static void
 ask (HTMLEngine *e, gpointer data)
 {
-	GtkHTMLReplaceDialog *d = (GtkHTMLReplaceDialog *) data;
+	GtkHTMLReplaceAskDialog *ask_dialog;
 
-	if (!d->ask_dialog) {
-		d->ask_dialog = ask_dialog_new (e);
-		gnome_dialog_run (d->ask_dialog->dialog);
-	} else {
-		gtk_widget_show (GTK_WIDGET (d->ask_dialog->dialog));
-		gdk_window_raise (GTK_WIDGET (d->ask_dialog->dialog)->window);
-	}	
+	ask_dialog = ask_dialog_new (e);
+	gtk_dialog_run (ask_dialog->dialog);
+	gtk_widget_destroy (GTK_WIDGET (ask_dialog->dialog));
+	g_free (ask_dialog);
 }
 
 static void
-button_replace_cb (GtkWidget *but, GtkHTMLReplaceDialog *d)
+entry_search_changed (GtkWidget *entry, GtkHTMLReplaceDialog *d)
 {
-	gnome_dialog_close  (d->dialog);	
-	html_engine_replace (d->html->engine,
-			     gtk_entry_get_text (GTK_ENTRY (d->entry_search)),
-			     gtk_entry_get_text (GTK_ENTRY (d->entry_replace)),
-			     GTK_TOGGLE_BUTTON (d->case_sensitive)->active,
-			     GTK_TOGGLE_BUTTON (d->backward)->active == 0, FALSE,
-			     ask, d);
+	if (d->cd->replace_text_search)
+		g_free (d->cd->replace_text_search);
+	d->cd->replace_text_search = g_strdup (gtk_entry_get_text (GTK_ENTRY (d->entry_search)));
 }
 
 static void
-entry_changed (GtkWidget *entry, GtkHTMLReplaceDialog *d)
+entry_replace_changed (GtkWidget *entry, GtkHTMLReplaceDialog *d)
 {
+	if (d->cd->replace_text_replace)
+		g_free (d->cd->replace_text_replace);
+	d->cd->replace_text_replace = g_strdup (gtk_entry_get_text (GTK_ENTRY (d->entry_replace)));
 }
 
 static void
 entry_activate (GtkWidget *entry, GtkHTMLReplaceDialog *d)
 {
-	button_replace_cb (NULL, d);
+	gtk_dialog_response (GTK_DIALOG (d->dialog), 0);
+}
+
+static void
+replace_dialog_response (GtkDialog *dialog, gint response_id, GtkHTMLReplaceDialog *d)
+{
+	switch (response_id) {
+	case 0: /* Replace */
+		gtk_widget_hide (GTK_WIDGET (d->dialog));
+		html_engine_replace (d->html->engine,
+				     gtk_entry_get_text (GTK_ENTRY (d->entry_search)),
+				     gtk_entry_get_text (GTK_ENTRY (d->entry_replace)),
+				     GTK_TOGGLE_BUTTON (d->case_sensitive)->active,
+				     GTK_TOGGLE_BUTTON (d->backward)->active == 0, FALSE,
+				     ask, d);
+		break;
+	}
+}
+
+void
+gtk_html_replace_dialog_destroy (GtkHTMLReplaceDialog *d)
+{
+	gtk_widget_destroy (GTK_WIDGET (d->dialog));
+	g_free (d);
 }
 
 GtkHTMLReplaceDialog *
-gtk_html_replace_dialog_new (GtkHTML *html)
+gtk_html_replace_dialog_new (GtkHTML *html, GtkHTMLControlData *cd)
 {
 	GtkHTMLReplaceDialog *dialog = g_new (GtkHTMLReplaceDialog, 1);
 	GtkWidget *hbox;
 	GtkWidget *table;
 	GtkWidget *label;
 
-	dialog->dialog         = GNOME_DIALOG (gnome_dialog_new (NULL, _("Replace"),
-								 GNOME_STOCK_BUTTON_CANCEL, NULL));
+	dialog->dialog         = GTK_DIALOG (gtk_dialog_new_with_buttons (_("Replace"), NULL, 0,
+									  GTK_STOCK_CLOSE, GTK_RESPONSE_CLOSE,
+									  _("Replace"), 0,
+									  NULL));
 
 	table = gtk_table_new (2, 2, FALSE);
 	dialog->entry_search   = gtk_entry_new ();
@@ -155,7 +160,12 @@ gtk_html_replace_dialog_new (GtkHTML *html)
 	dialog->backward       = gtk_check_button_new_with_label (_("search backward"));
 	dialog->case_sensitive = gtk_check_button_new_with_label (_("case sensitive"));
 	dialog->html           = html;
-	dialog->ask_dialog     = NULL;
+	dialog->cd             = cd;
+
+	if (cd->replace_text_search)
+		gtk_entry_set_text (GTK_ENTRY (dialog->entry_search), cd->replace_text_search);
+	if (cd->replace_text_replace)
+		gtk_entry_set_text (GTK_ENTRY (dialog->entry_replace), cd->replace_text_replace);
 
 	gtk_table_set_col_spacings (GTK_TABLE (table), 3);
 	label = gtk_label_new (_("Replace"));
@@ -180,34 +190,21 @@ gtk_html_replace_dialog_new (GtkHTML *html)
 
 	gnome_window_icon_set_from_file (GTK_WINDOW (dialog->dialog), ICONDIR "/search-and-replace-24.png");
 
-	gnome_dialog_button_connect (dialog->dialog, 0, button_replace_cb, dialog);
-	gnome_dialog_close_hides (dialog->dialog, TRUE);
-	gnome_dialog_set_close (dialog->dialog, TRUE);
-
-	gnome_dialog_set_default (dialog->dialog, 0);
 	gtk_widget_grab_focus (dialog->entry_search);
 
-	gtk_signal_connect (GTK_OBJECT (dialog->entry_search), "changed",
-			    entry_changed, dialog);
-	gtk_signal_connect (GTK_OBJECT (dialog->entry_search), "activate",
-			    entry_activate, dialog);
-	gtk_signal_connect (GTK_OBJECT (dialog->entry_replace), "activate",
-			    entry_activate, dialog);
+	g_signal_connect (dialog->dialog, "response", G_CALLBACK (replace_dialog_response), dialog);
+	g_signal_connect (dialog->entry_search, "changed", G_CALLBACK (entry_search_changed), dialog);
+	g_signal_connect (dialog->entry_search, "activate", G_CALLBACK (entry_activate), dialog);
+	g_signal_connect (dialog->entry_replace, "changed", G_CALLBACK (entry_replace_changed), dialog);
+	g_signal_connect (dialog->entry_replace, "activate", G_CALLBACK (entry_activate), dialog);
 
 	return dialog;
 }
 
-void
-gtk_html_replace_dialog_destroy (GtkHTMLReplaceDialog *d)
-{
-	g_free (d);
-}
-
 void
 replace (GtkHTMLControlData *cd)
 {
 	RUN_DIALOG (replace, _("Replace"));
-
-	if (cd->replace_dialog)
-		gtk_widget_grab_focus (cd->replace_dialog->entry_search);
+	gtk_html_replace_dialog_destroy (cd->replace_dialog);
+	cd->replace_dialog = NULL;
 }
