@@ -30,8 +30,6 @@
 #include <string.h>
 #include <stdlib.h>
 
-#include <gal/widgets/e-unicode.h>
-
 #include "gtkhtml-properties.h"
 
 #include "htmlcolor.h"
@@ -51,22 +49,18 @@
 #include "htmltablecell.h"
 #include "htmltext.h"
 #include "htmltextslave.h"	/* FIXME */
-#include "htmlvspace.h"
 
-
 HTMLClueFlowClass html_clueflow_class;
 static HTMLClueClass *parent_class = NULL;
 
 #define HCF_CLASS(x) HTML_CLUEFLOW_CLASS (HTML_OBJECT (x)->klass)
-#define HTML_IS_PLAIN_PAINTER(obj)              (GTK_CHECK_TYPE ((obj), HTML_TYPE_PLAIN_PAINTER))
 
-inline HTMLHAlignType html_clueflow_get_halignment (HTMLClueFlow *flow);
-static gchar * get_item_number_str (HTMLClueFlow *flow);
-static void update_items_after_indentation_change (HTMLClueFlow *flow);
-
-#define CLUEFLOW_INDENT             "        "
-#define CLUEFLOW_BLOCKQUOTE_CITE    "> "
-          
+inline HTMLHAlignType html_clueflow_get_halignment          (HTMLClueFlow *flow);
+static gchar *        get_item_number_str                   (HTMLClueFlow *flow);
+static guint          get_post_padding                      (HTMLClueFlow *flow, 
+							     guint pad);
+static int            get_similar_depth                     (HTMLClueFlow *self, 
+							     HTMLClueFlow *neighbor);
 
 static void
 copy_levels (GByteArray *dst, GByteArray *src)
@@ -94,10 +88,22 @@ is_levels_equal (HTMLClueFlow *me, HTMLClueFlow *you)
 	return !memcmp (me->levels->data, you->levels->data, you->levels->len);
 }
 
+static inline gboolean
+is_item (HTMLClueFlow *flow)
+{
+	return flow && flow->style == HTML_CLUEFLOW_STYLE_LIST_ITEM;
+}
+
 static void
 destroy (HTMLObject *self)
 {
-	g_byte_array_free (HTML_CLUEFLOW (self)->levels, TRUE);
+	HTMLClueFlow *flow = HTML_CLUEFLOW (self);
+
+	g_byte_array_free (flow->levels, TRUE);
+	if (flow->item_color) {
+		html_color_unref (flow->item_color);
+		flow->item_color = NULL;
+	}
 
 	(* HTML_OBJECT_CLASS (parent_class)->destroy) (self);
 }
@@ -113,12 +119,11 @@ copy (HTMLObject *self,
 	HTML_CLUEFLOW (dest)->item_type = HTML_CLUEFLOW (self)->item_type;
 	HTML_CLUEFLOW (dest)->item_number = HTML_CLUEFLOW (self)->item_number;
 	HTML_CLUEFLOW (dest)->clear = HTML_CLUEFLOW (self)->clear;
-}
-
-static inline gboolean
-is_item (HTMLClueFlow *flow)
-{
-	return flow && flow->style == HTML_CLUEFLOW_STYLE_LIST_ITEM;
+	HTML_CLUEFLOW (dest)->item_color = HTML_CLUEFLOW (self)->item_color;
+	HTML_CLUEFLOW (dest)->indent_width = HTML_CLUEFLOW (self)->indent_width;
+	
+	if (HTML_CLUEFLOW (dest)->item_color)
+		html_color_ref (HTML_CLUEFLOW (dest)->item_color);
 }
 
 static inline gboolean
@@ -158,7 +163,9 @@ get_prev_relative_item (HTMLObject *self)
 	prev = self->prev;
 	while (prev 
 	       && HTML_IS_CLUEFLOW (prev) 
-	       && HTML_CLUEFLOW (prev)->levels->len > HTML_CLUEFLOW (self)->levels->len
+	       && (HTML_CLUEFLOW (prev)->levels->len > HTML_CLUEFLOW (self)->levels->len
+		   || (HTML_CLUEFLOW (prev)->levels->len == HTML_CLUEFLOW (self)->levels->len
+		       && !is_item (HTML_CLUEFLOW (prev))))
 	       && !memcmp (HTML_CLUEFLOW (prev)->levels->data,
 			   HTML_CLUEFLOW (self)->levels->data,
 			   HTML_CLUEFLOW (self)->levels->len))
@@ -174,8 +181,11 @@ get_next_relative_item (HTMLObject *self)
 	HTMLObject *next;
 
 	next = self->next;
-	while (next && HTML_IS_CLUEFLOW (next)
-	       && HTML_CLUEFLOW (next)->levels->len > HTML_CLUEFLOW (self)->levels->len
+	while (next 
+	       && HTML_IS_CLUEFLOW (next)
+	       && (HTML_CLUEFLOW (next)->levels->len > HTML_CLUEFLOW (self)->levels->len
+		   || (HTML_CLUEFLOW (next)->levels->len == HTML_CLUEFLOW (self)->levels->len
+		       && !is_item (HTML_CLUEFLOW (next))))
 	       && !memcmp (HTML_CLUEFLOW (next)->levels->data,
 			   HTML_CLUEFLOW (self)->levels->data, 
 			   HTML_CLUEFLOW (self)->levels->len))
@@ -186,22 +196,26 @@ get_next_relative_item (HTMLObject *self)
 }
 
 static void
-update_item_number (HTMLObject *self)
+update_item_number (HTMLObject *self, HTMLEngine *e)
 {
 	HTMLObject *prev, *next;
 
 	if (!self || !is_item (HTML_CLUEFLOW (self)))
 		return;
 
+	/* printf ("update_item_number\n"); */
+
 	prev = get_prev_relative_item (self);
 	if (items_are_relative (prev, self))
 		HTML_CLUEFLOW (self)->item_number = HTML_CLUEFLOW (prev)->item_number + 1;
 	else if (is_item (HTML_CLUEFLOW (self)))
 		HTML_CLUEFLOW (self)->item_number = 1;
+	html_engine_queue_draw (e, self);
 
 	next = self;
 	while ((next = get_next_relative_item (next)) && items_are_relative (self, next)) {
 		HTML_CLUEFLOW (next)->item_number = HTML_CLUEFLOW (self)->item_number + 1;
+		html_engine_queue_draw (e, next);
 		self = next;
 	}
 }
@@ -229,13 +243,13 @@ op_helper (HTMLObject *self, HTMLEngine *e, GList *from, GList *to, GList *left,
 		   e->cursor->position --; */
 	}
 	if (cut)
-		html_clueflow_remove_text_slaves (HTML_CLUEFLOW (self));
+		html_clue_remove_text_slaves (HTML_CLUE (self));
 	o = cut
 		? (*HTML_OBJECT_CLASS (parent_class)->op_cut) (self, e, from, to, left, right, len)
 		: (*HTML_OBJECT_CLASS (parent_class)->op_copy) (self, NULL, e, from, to, len);
 
 	if (!cut && o) {
-		html_clueflow_remove_text_slaves (HTML_CLUEFLOW (o));
+		html_clue_remove_text_slaves (HTML_CLUE (o));
 	}
 
 	return o;
@@ -257,15 +271,15 @@ op_cut (HTMLObject *self, HTMLEngine *e, GList *from, GList *to, GList *left, GL
 
 	rv = op_helper (self, e, from, to, left, right, len, TRUE);
 
-	if (prev) {
-		update_item_number (prev);
+	if (prev && from) {
+		update_item_number (prev, e);
 		if (prev->next == self)
-			update_item_number (self);
+			update_item_number (self, e);
 	}
-	if (next) {
+	if (next && to) {
 		if (next->prev == self)
-			update_item_number (self);
-		update_item_number (next);
+			update_item_number (self, e);
+		update_item_number (next, e);
 	}
 
 	return rv;
@@ -301,34 +315,31 @@ static void
 split (HTMLObject *self, HTMLEngine *e, HTMLObject *child, gint offset, gint level, GList **left, GList **right)
 {
 	set_around_size (child);
-	html_clueflow_remove_text_slaves (HTML_CLUEFLOW (self));
+	html_clue_remove_text_slaves (HTML_CLUE (self));
 	(*HTML_OBJECT_CLASS (parent_class)->split) (self, e, child, offset, level, left, right);
 
-	update_item_number (self);
+	update_item_number (self, e);
 }
 
 static gboolean
 merge (HTMLObject *self, HTMLObject *with, HTMLEngine *e, GList **left, GList **right, HTMLCursor *cursor)
 {
 	HTMLClueFlow *cf1, *cf2;
+	HTMLObject *cf2_next_relative;
 	gboolean rv;
 
 	cf1 = HTML_CLUEFLOW (self);
 	cf2 = HTML_CLUEFLOW (with);
 
-	html_clueflow_remove_text_slaves (cf1);
-	html_clueflow_remove_text_slaves (cf2);
+	html_clue_remove_text_slaves (HTML_CLUE (cf1));
+	html_clue_remove_text_slaves (HTML_CLUE (cf2));
+
+	cf2_next_relative = get_next_relative_item (with);
 
 	set_tail_size (self);
 	set_head_size (with);
 
-	/* printf ("merge flows\n"); */
-
 	if (html_clueflow_is_empty (cf1)) {
-		cf1->style = cf2->style;
-		copy_levels (cf1->levels, cf2->levels);
-		cf1->item_type = cf2->item_type;
-		cf1->item_number = cf2->item_number - 1;
 		self->x = with->x;
 		self->y = with->y;
 		self->width = with->width;
@@ -341,43 +352,45 @@ merge (HTMLObject *self, HTMLObject *with, HTMLEngine *e, GList **left, GList **
 
 	rv = (* HTML_OBJECT_CLASS (parent_class)->merge) (self, with, e, left, right, cursor);
 
-	update_item_number (self);
-	cf1->item_number --;
-	update_item_number (with);
-	cf1->item_number ++;
+	if (rv) {
+		if (is_item (cf1)) {
+			/* cf2 will be removed, update item numbers around
+			   as if it was already removed - it has to have
+			   the same item style as cf1 to not break item lists*/
+			g_byte_array_free (cf2->levels, TRUE);
+			cf2->levels = html_clueflow_dup_levels (cf1);
+			cf2->style = cf1->style;
+			cf2->item_type = cf1->item_type;
 
+			update_item_number (self, e);
+			cf1->item_number --;
+			update_item_number (with, e);
+			cf1->item_number ++;
+
+			if (cf2_next_relative)
+				update_item_number (cf2_next_relative, e);
+		}
+	
+	}
+		
 	return rv;
 }
 
 static guint
 calc_padding (HTMLPainter *painter)
 {
-	guint ascent, descent;
-
-	/* FIXME maybe this should depend on the style.  */
-	/* FIXME move this into the painter.             */
-	if (HTML_IS_PLAIN_PAINTER (painter)) {
-		ascent = 0;
-		descent = 0;
-	} else {
-		ascent = html_painter_calc_ascent (painter, GTK_HTML_FONT_STYLE_SIZE_3, NULL);
-		descent = html_painter_calc_descent (painter, GTK_HTML_FONT_STYLE_SIZE_3, NULL);
+	if (!HTML_IS_PLAIN_PAINTER (painter)) {
+		return html_painter_get_space_width (painter, GTK_HTML_FONT_STYLE_SIZE_3, NULL);
 	}
-	return ascent + descent;
+	return 0;
 }
 
 static guint
 calc_bullet_size (HTMLPainter *painter)
 {
-	guint ascent, descent;
-
-	ascent = html_painter_calc_ascent (painter, GTK_HTML_FONT_STYLE_SIZE_3, NULL);
-	descent = html_painter_calc_descent (painter, GTK_HTML_FONT_STYLE_SIZE_3, NULL);
-
-	return (ascent + descent) / 3;
+	return html_painter_get_space_width (painter, GTK_HTML_FONT_STYLE_SIZE_3, NULL) / 2;
 }
 
-
 static gboolean
 is_cite (HTMLClueFlow *flow, gint level)
 {
@@ -386,7 +399,7 @@ is_cite (HTMLClueFlow *flow, gint level)
 
 	return FALSE;
 }
-
+  
 static gboolean
 is_header (HTMLClueFlow *flow)
 {
@@ -403,7 +416,44 @@ is_header (HTMLClueFlow *flow)
 	}
 }
 
-static guint get_post_padding (HTMLClueFlow *flow, guint pad);
+static gboolean
+need_blockquote_padding  (HTMLClueFlow *flow, HTMLClueFlow *prev)
+{
+	int i = get_similar_depth (flow, prev);
+	
+	/* 
+	 * If the levels don't match up the the current flow
+	 * level the padding should be handled the other way.
+	 */
+	if (i < flow->levels->len || flow->levels->len == 0) {
+		if (i < prev->levels->len)
+			return TRUE;
+                else 
+			return FALSE;
+	}
+	
+	i = prev->levels->len - i;
+	
+	/*
+	 * now we check each level greater than the current flow level
+	 * and see if it is a blockquote and therefore needs padding
+	 */
+	while (i > 0) {
+		HTMLListType type;
+		
+		type = prev->levels->data [prev->levels->len - i];
+
+		if (is_blockquote (type)) {
+			return TRUE;
+		}
+		i--;
+	}
+
+	/*
+	 * If all the levels were items we don't need padding
+	 */
+	return FALSE;
+}
 
 static guint
 get_pre_padding (HTMLClueFlow *flow, guint pad)
@@ -420,19 +470,14 @@ get_pre_padding (HTMLClueFlow *flow, guint pad)
 		if (get_post_padding (HTML_CLUEFLOW (prev_object), 1))
 			return 0;
 
-		if (is_item (HTML_CLUEFLOW (prev_object))) {
-			if (is_item (flow) || is_levels_equal (HTML_CLUEFLOW (prev_object), flow)) {
-				return 0;
-			} else {
-				return pad;
-			}
-		} else if (is_item (flow)) {
-			return pad;
-		}
-
 		prev = HTML_CLUEFLOW (prev_object);
-		if (prev->levels->len > flow->levels->len)
-			return pad;
+
+		if (!is_levels_equal (flow, prev)) {
+			if (need_blockquote_padding (flow, prev))
+				return pad;
+			else 
+				return 0;
+  		}
 
 		if (flow->style == HTML_CLUEFLOW_STYLE_PRE
 		    && prev->style != HTML_CLUEFLOW_STYLE_PRE
@@ -447,7 +492,7 @@ get_pre_padding (HTMLClueFlow *flow, guint pad)
 
 	if (! is_header (flow) && flow->levels->len == 0)
 		return 0;
-
+	
 	return pad;
 }
 
@@ -464,13 +509,14 @@ get_post_padding (HTMLClueFlow *flow,
 	if (HTML_OBJECT_TYPE (next_object) == HTML_TYPE_CLUEFLOW) {
 		HTMLClueFlow *next;
 
-		if (is_item (flow) && is_item (HTML_CLUEFLOW (next_object)))
-			return 0;
-
 		next = HTML_CLUEFLOW (next_object);
-		if ((next->levels->len >= flow->levels->len)
-		    && !is_levels_equal (next, flow))
-			return pad;
+
+		if (!is_levels_equal (next, flow)) {
+			if (need_blockquote_padding (flow, next)) 
+				return pad;
+			else 
+				return 0;
+		}
 
 		if (flow->style == HTML_CLUEFLOW_STYLE_PRE
 		    && next->style != HTML_CLUEFLOW_STYLE_PRE
@@ -518,37 +564,34 @@ get_level_indent (HTMLClueFlow *flow,
 		  gint level,
 		  HTMLPainter *painter)
 {
-	GtkHTMLFontStyle style;
 	int line_offset = 0;
 	guint indent = 0;
 	gint i = 0;
-	
-	style = html_clueflow_get_default_font_style (flow);
 
 	if (flow->levels->len > 0 || ! is_item (flow)) {
+		guint cite_width, indent_width;
+
+		cite_width   = html_painter_get_block_cite_width (painter, GTK_HTML_FONT_STYLE_SIZE_3, NULL);
+		indent_width = html_painter_get_block_indent_width (painter, GTK_HTML_FONT_STYLE_SIZE_3, NULL);
+		
 		while (i <= level) {
 			switch (flow->levels->data[i]) {
 			case HTML_LIST_TYPE_BLOCKQUOTE_CITE:
-				indent += html_painter_calc_text_width (painter, 
-									CLUEFLOW_BLOCKQUOTE_CITE, 
-									strlen (CLUEFLOW_BLOCKQUOTE_CITE), &line_offset,
-									GTK_HTML_FONT_STYLE_SIZE_3, 
-									NULL);
+				indent += cite_width;
 				break;
 			case HTML_LIST_TYPE_GLOSSARY_DL:
 				indent += 0;
 				break;
 			default:
-				indent += html_painter_calc_text_width (painter, 
-									CLUEFLOW_INDENT, 
-									strlen (CLUEFLOW_INDENT), &line_offset,
-									GTK_HTML_FONT_STYLE_SIZE_3, 
-									NULL);
+				indent += indent_width;
 				break;
 			}
 			i++;
 		}
 	} else {
+		GtkHTMLFontStyle style;
+		style = html_clueflow_get_default_font_style (flow);
+
 		indent = 4 * html_painter_get_space_width (painter, style, NULL);
 	}
 
@@ -559,9 +602,12 @@ static guint
 get_indent (HTMLClueFlow *flow,
 	    HTMLPainter *painter)
 {
-	return get_level_indent (flow, flow->levels->len - 1, painter);
-}
+	if (flow->indent_width < 0 )
+		flow->indent_width = get_level_indent (flow, flow->levels->len -1, painter);
 
+	return flow->indent_width;
+} 
+ 
 /* HTMLObject methods.  */
 
 static void
@@ -601,7 +647,7 @@ calc_min_width (HTMLObject *o,
 			w += add
 				? html_object_calc_preferred_width (cur, painter)
 				: html_object_calc_min_width (cur, painter);
-			if (!add || cur->flags & HTML_OBJECT_FLAG_NEWLINE || !cur->next) {
+			if (!add || !cur->next) {
 				if (min_width < w) min_width = w;
 				w = 0;
 			}
@@ -663,7 +709,7 @@ static gint
 object_nb_width (HTMLObject *o, HTMLPainter *painter, gboolean lineBegin)
 {
 	if (HTML_IS_TEXT_SLAVE (o))
-		return html_text_slave_nb_width (HTML_TEXT_SLAVE (o), painter, lineBegin);
+		return html_text_slave_get_nb_width (HTML_TEXT_SLAVE (o), painter, lineBegin);
 		
 	return html_object_calc_min_width (o, painter);
 }
@@ -728,7 +774,7 @@ update_top_height (HTMLObject *begin, HTMLObject *end, gint *a, gint *d, gint *h
 static inline void
 update_line_positions (HTMLObject *clue, HTMLObject *begin, HTMLObject *end, gint left, gint a, gint d, gint height)
 {
-	gint xinc;
+	gint xinc = 0;
 
 	switch (html_clueflow_get_halignment (HTML_CLUEFLOW (clue))) {
 	case HTML_HALIGN_NONE:
@@ -903,7 +949,7 @@ static gboolean
 calc_size (HTMLObject *o, HTMLPainter *painter, GList **changed_objs)
 {
 	HTMLClueFlow *cf = HTML_CLUEFLOW (o);
-	gint oa, od, ow, padding;
+	gint oa, od, ow, oi, padding;
 	gboolean leaf_children_changed_size = FALSE;
 	gboolean changed, changed_size = FALSE;
 
@@ -911,6 +957,9 @@ calc_size (HTMLObject *o, HTMLPainter *painter, GList **changed_objs)
 	oa = o->ascent;
 	od = o->descent;
 	ow = o->width;
+
+	cf->indent_width = -1;
+
 	o->ascent = 0;
 	o->descent = 0;
 	o->width = MAX (o->max_width, html_object_calc_min_width (o, painter));
@@ -956,16 +1005,13 @@ calc_preferred_width (HTMLObject *o,
 	next = NULL;
 
 	for (obj = HTML_CLUE (o)->head; obj != 0; obj = obj->next) {
-		if (!(obj->flags & HTML_OBJECT_FLAG_NEWLINE)) {
-			w += html_object_calc_preferred_width (obj, painter);
-		}
+		w += html_object_calc_preferred_width (obj, painter);
 
-		if (obj->flags & HTML_OBJECT_FLAG_NEWLINE || !(next = html_object_next_not_slave (obj))) {
-			HTMLObject *eol = (obj->flags & HTML_OBJECT_FLAG_NEWLINE) ? html_object_prev_not_slave (obj) : obj;
+		if (!(next = html_object_next_not_slave (obj))) {
 
 			/* remove trailing space width on the end of line which is not on end of paragraph */
-			if (next && html_object_is_text (eol))
-				w -= html_text_trail_space_width (HTML_TEXT (eol), painter);
+			if (next && html_object_is_text (obj))
+				w -= html_text_trail_space_width (HTML_TEXT (obj), painter);
 
 			if (w > maxw)
 				maxw = w;
@@ -1055,17 +1101,47 @@ get_item_number_str (HTMLClueFlow *flow)
 	}
 }
 
+static void
+draw_gt_line (HTMLObject *cur, HTMLPainter *p, gint offset, gint x, gint y)
+{
+	gint cy, w, a, d, line_offset = 0;
+
+	/* FIXME: cache items and glyphs? */
+	html_painter_calc_text_size (p, HTML_BLOCK_CITE, 
+				     strlen (HTML_BLOCK_CITE), NULL, NULL, &line_offset,
+				     GTK_HTML_FONT_STYLE_SIZE_3, NULL,
+				     &w, &a, &d);
+
+	cy = offset;
+	while (cy + a <= cur->ascent) {
+		/* FIXME: cache items and glyphs? */
+		html_painter_draw_text (p, x, y + cur->y - cy,
+					HTML_BLOCK_CITE, 1, NULL, NULL, 0);
+		cy += a + d;
+	}
+
+	cy = - offset + a + d;
+	while (cy + d <= cur->descent) {
+		/* FIXME: cache items and glyphs? */
+		html_painter_draw_text (p, x, y + cur->y + cy,
+					HTML_BLOCK_CITE, 1, NULL, NULL, 0);
+		cy += a + d;
+	}
+}
+		
 static void
 draw_quotes (HTMLObject *self, HTMLPainter *painter, 
 	     gint x, gint y, gint width, gint height,
 	     gint tx, gint ty)
 {
 	HTMLClueFlow *flow;
-	ArtIRect paint, area, clip;
+	GdkRectangle paint, area, clip;
 	int i;
 	int indent = 0;
 	int last_indent = 0;
-
+	gint pixel_size = html_painter_get_pixel_size (painter);
+	gboolean is_plain = HTML_IS_PLAIN_PAINTER (painter);
+	
 	flow = HTML_CLUEFLOW (self);
 
 	for (i = 0; i < flow->levels->len; i++, last_indent = indent) {
@@ -1073,43 +1149,42 @@ draw_quotes (HTMLObject *self, HTMLPainter *painter,
 
 		html_painter_set_pen (painter, &html_colorset_get_color_allocated (painter, HTMLLinkColor)->color);
 		if (is_cite (flow, i)) {
-			if (!HTML_IS_PLAIN_PAINTER (painter)) {
-				area.x0 = self->x + indent - 5;
-				area.x1 = area.x0 + 2;
-				area.y0 = self->y - self->ascent;
-				area.y1 = self->y + self->descent;
+			if (!is_plain) {
+				area.x = self->x + indent - 5 * pixel_size;
+				area.width = 2 * pixel_size;
+				area.y = self->y - self->ascent;
+				area.height = self->ascent + self->descent;
 				
-				clip.x0 = x;
-				clip.x1 = x + width;
-				clip.y0 = y;
-				clip.y1 = y + height;
+				clip.x = x;
+				clip.width = width;
+				clip.y = y;
+				clip.height = height;
 				
-				art_irect_intersect (&paint, &clip, &area);
-				if (art_irect_empty (&paint))
+				if (!gdk_rectangle_intersect (&clip, &area, &paint))
 					return;
-				
-				html_painter_fill_rect (painter, 
-							paint.x0 + tx, paint.y0 + ty,
-							paint.x1 - paint.x0, paint.y1 - paint.y0);
-			} else {
-				/* draw "> " quote characters in the plain case */ 
-				HTMLObject *cur = HTML_CLUE (self)->head;
-				gint last_y = 0;
-			       
 
+				html_painter_fill_rect (painter, 
+							paint.x + tx, paint.y + ty,
+							paint.width, paint.height);
+			} else {
+				HTMLObject *cur = HTML_CLUE (self)->head;
+				gint baseline = 0;
 				while (cur) {
-					if (cur->y != last_y) {
-						html_painter_set_font_style (painter, 
-								     html_clueflow_get_default_font_style (flow));
+					if (cur->y != 0) {
+						baseline = cur->y;
+						break;
+  					}
+  					cur = cur->next;
+  				}
+
+
+				/* draw "> " quote characters in the plain case */ 
+				html_painter_set_font_style (painter, 
+							     html_clueflow_get_default_font_style (flow));
 						
-						html_painter_set_font_face  (painter, NULL);
-						html_painter_draw_text (painter, self->x + tx + last_indent,
-									self->y - self->ascent + cur->y + ty,
-									CLUEFLOW_BLOCKQUOTE_CITE, 1, 0);
-					}
-					last_y = cur->y;
-					cur = cur->next;
-				}
+				html_painter_set_font_face  (painter, NULL);
+				draw_gt_line (self, painter, self->ascent - baseline,
+					      self->x + tx + last_indent, ty);
 			}
 		}
 	}
@@ -1127,15 +1202,20 @@ draw_item (HTMLObject *self, HTMLPainter *painter, gint x, gint y, gint width, g
 		first = first->next;
 
 	flow = HTML_CLUEFLOW (self);
-	html_painter_set_pen (painter, &html_colorset_get_color_allocated (painter, HTMLTextColor)->color);
 
-	indent = get_level_indent (flow, flow->levels->len - 1, painter);
+	if (flow->item_color) {
+		html_color_alloc (flow->item_color, painter);
+		html_painter_set_pen (painter, &flow->item_color->color);
+	} else
+		html_painter_set_pen (painter, &html_colorset_get_color_allocated (painter, HTMLTextColor)->color);
+
+	indent = get_indent (flow, painter);
 	if (flow->item_type == HTML_LIST_TYPE_UNORDERED) {
 		guint bullet_size;
 		gint xp, yp;
 		bullet_size = MAX (3, calc_bullet_size (painter));
 
-		xp = self->x + indent - 2 * bullet_size;
+		xp = self->x + indent - 2 * bullet_size;	
 		yp = self->y - self->ascent 
 			+ (first->y - first->ascent) 
 			+ (first->ascent + first->descent)/2 
@@ -1157,17 +1237,19 @@ draw_item (HTMLObject *self, HTMLPainter *painter, gint x, gint y, gint width, g
 
 		number = get_item_number_str (flow);
 		if (number) {
-			gint width, len, line_offset = 0;
+			gint width, len, line_offset = 0, asc, dsc;
 
 			len   = strlen (number);
-			width = html_painter_calc_text_width (painter, number, len, &line_offset,
-							      html_clueflow_get_default_font_style (flow), NULL);
-
+			/* FIXME: cache items and glyphs? */
+			html_painter_calc_text_size (painter, number, len, NULL, NULL, &line_offset,
+						     html_clueflow_get_default_font_style (flow), NULL, &width, &asc, &dsc);
+			width += html_painter_get_space_width (painter, html_clueflow_get_default_font_style (flow), NULL);
 			html_painter_set_font_style (painter, html_clueflow_get_default_font_style (flow));
 			html_painter_set_font_face  (painter, NULL);
-			html_painter_draw_text (painter, self->x + indent - width + tx,
+			/* FIXME: cache items and glyphs? */
+			html_painter_draw_text (painter, self->x + first->x - width + tx,
 						self->y - self->ascent + first->y + ty,
-						number, strlen (number), 0);
+						number, strlen (number), NULL, NULL, 0);
 		}
 		g_free (number);
 	}
@@ -1311,15 +1393,6 @@ check_point (HTMLObject *self,
 	return NULL;
 }
 
-static void
-append_selection_string (HTMLObject *self,
-			 GString *buffer)
-{
-        (*HTML_OBJECT_CLASS (parent_class)->append_selection_string) (self, buffer);
-	if (self->selected)
-		g_string_append_c (buffer, '\n');
-}
-
 
 /* Saving support.  */
 
@@ -1694,12 +1767,12 @@ plain_padding (HTMLClueFlow *flow, GString *out, gboolean firstline)
 		for (i = 0; i < flow->levels->len; i++) {
 			switch (flow->levels->data[i]) {
 			case HTML_LIST_TYPE_BLOCKQUOTE_CITE:
-				APPEND_PLAIN (CLUEFLOW_BLOCKQUOTE_CITE);
+				APPEND_PLAIN (HTML_BLOCK_CITE);
 				break;
 			case HTML_LIST_TYPE_GLOSSARY_DL:
 				break;
 			default:
-				APPEND_PLAIN (CLUEFLOW_INDENT);
+				APPEND_PLAIN (HTML_BLOCK_INDENT);
 				break;
 			}
 		}
@@ -1719,6 +1792,18 @@ plain_padding (HTMLClueFlow *flow, GString *out, gboolean firstline)
 	return pad_len;
 }
 
+static void
+append_selection_string (HTMLObject *self,
+			 GString *buffer)
+{
+        (*HTML_OBJECT_CLASS (parent_class)->append_selection_string) (self, buffer);
+
+	if (self->selected) {
+		g_string_append_c (buffer, '\n');
+		plain_padding (HTML_CLUEFLOW (self), buffer, TRUE);
+	}
+}
+
 static gboolean
 save_plain (HTMLObject *self,
 	    HTMLEngineSaveState *state,
@@ -1736,7 +1821,8 @@ save_plain (HTMLObject *self,
 	flow = HTML_CLUEFLOW (self);
 
 	pad = plain_padding (flow, NULL, FALSE);
-	buffer_state = html_engine_save_buffer_new (state->engine);
+	buffer_state = html_engine_save_buffer_new (state->engine, 
+						    state->inline_frames);
 
 	max_len = MAX (requested_width - pad, 0);
 	/* buffer the paragraph's content into the save buffer */
@@ -1969,7 +2055,7 @@ search_text (HTMLObject **beg, HTMLSearch *info)
 				gint rv;
 #ifndef HAVE_GNU_REGEX
 				regmatch_t match;
-				guchar *p=par+pos;
+				/* guchar *p=par+pos; */
 
 				/* FIXME UTF8
 				   replace &nbsp;'s with spaces
@@ -2105,7 +2191,7 @@ relayout (HTMLObject *self,
 	if (mw <= self->max_width)
 		return (*HTML_OBJECT_CLASS (parent_class)->relayout) (self, engine, child);
 	html_engine_calc_size (engine, FALSE);
-	html_engine_draw (engine, 0, 0, engine->width, engine->height);
+	html_engine_draw (engine, engine->x_offset, engine->y_offset, engine->width, engine->height);
 
 	return TRUE;
 }
@@ -2161,7 +2247,7 @@ html_clueflow_class_init (HTMLClueFlowClass *klass,
 void
 html_clueflow_init (HTMLClueFlow *clueflow, HTMLClueFlowClass *klass,
 		    HTMLClueFlowStyle style, GByteArray *levels, HTMLListType item_type, gint item_number,
-		    HTMLClearType clear)
+ 		    HTMLClearType clear)
 {
 	HTMLObject *object;
 	HTMLClue *clue;
@@ -2178,16 +2264,17 @@ html_clueflow_init (HTMLClueFlow *clueflow, HTMLClueFlowClass *klass,
 
 	clueflow->style = style;
 	clueflow->levels = levels; 
+	clueflow->indent_width = -1;
 
 	clueflow->item_type   = item_type;
 	clueflow->item_number = item_number;
+	clueflow->item_color = NULL;
 
 	clueflow->clear = clear;
 }
 
 HTMLObject *
-html_clueflow_new (HTMLClueFlowStyle style, GByteArray *levels,
-		   HTMLListType item_type, gint item_number, HTMLClearType clear)
+html_clueflow_new (HTMLClueFlowStyle style, GByteArray *levels, HTMLListType item_type, gint item_number, HTMLClearType clear)
 {
 	HTMLClueFlow *clueflow;
 
@@ -2202,7 +2289,7 @@ html_clueflow_new_from_flow (HTMLClueFlow *flow)
 {
 	HTMLObject *o;
 
-	o = html_clueflow_new (flow->style, html_clueflow_dup_levels (flow),
+	o = html_clueflow_new (flow->style, html_clueflow_dup_levels (flow), 
 			       flow->item_type, flow->item_number, flow->clear);
 	html_object_copy_data_from_object (o, HTML_OBJECT (flow));
 
@@ -2341,8 +2428,14 @@ html_clueflow_set_levels (HTMLClueFlow *flow,
 			  HTMLEngine *engine,
 			  GByteArray *levels)
 {
+	HTMLObject *next_relative;
+
+	next_relative = get_next_relative_item (HTML_OBJECT (flow));
 	copy_levels (flow->levels, levels);
-	update_items_after_indentation_change (flow);
+
+	update_item_number (HTML_OBJECT (flow), engine);
+	if (next_relative)
+		update_item_number (next_relative, engine);
 
 	relayout_with_siblings (flow, engine);
 }
@@ -2363,9 +2456,9 @@ html_clueflow_set_item_type (HTMLClueFlow *flow,
 
 	flow->item_type = item_type;
 
-	update_item_number (HTML_OBJECT (flow));
+	update_item_number (HTML_OBJECT (flow), engine);
 	if (!items_are_relative (HTML_OBJECT (flow), HTML_OBJECT (flow)->next) && HTML_OBJECT (flow)->next)
-		update_item_number (HTML_OBJECT (flow)->next);
+		update_item_number (HTML_OBJECT (flow)->next, engine);
 
 	html_engine_schedule_update (engine);
 	/* FIXME - make it more effective: relayout_with_siblings (flow, engine); */
@@ -2418,28 +2511,20 @@ html_clueflow_get_halignment (HTMLClueFlow *flow)
 		return HTML_CLUE (flow)->halign;
 }
 
-static void
-update_items_after_indentation_change (HTMLClueFlow *flow)
-{
-	if (is_item (flow)) {
-		/* FIXME levels */
-		update_item_number (HTML_OBJECT (flow));
-		if (HTML_OBJECT (flow)->next && is_item (HTML_CLUEFLOW (HTML_OBJECT (flow)->next)))
-			update_item_number (HTML_OBJECT (flow)->next);
-	}
-}
-
 void
 html_clueflow_modify_indentation_by_delta (HTMLClueFlow *flow,
 					   HTMLEngine *engine,
 					   gint indentation_delta,
 					   guint8 *indentation_levels)
 {
+	HTMLObject *next_relative;
 	gint indentation;
 	g_return_if_fail (flow != NULL);
 	g_return_if_fail (engine != NULL);
 	g_return_if_fail (HTML_IS_ENGINE (engine));
 
+	next_relative = get_next_relative_item (HTML_OBJECT (flow));
+
 	indentation = flow->levels->len + indentation_delta;
 	indentation = indentation < 0 ? 0 : indentation;
 
@@ -2454,7 +2539,9 @@ html_clueflow_modify_indentation_by_delta (HTMLClueFlow *flow,
 		}
 	}
 
-	update_items_after_indentation_change (flow);
+	update_item_number (HTML_OBJECT (flow), engine);
+	if (next_relative)
+		update_item_number (next_relative, engine);
 	relayout_with_siblings (flow, engine);
 }
 
@@ -2464,6 +2551,7 @@ html_clueflow_set_indentation (HTMLClueFlow *flow,
 			       gint indentation,
 			       guint8 *indentation_levels)
 {
+	HTMLObject *next_relative;
 	int i;
 	g_return_if_fail (flow != NULL);
 	g_return_if_fail (engine != NULL);
@@ -2472,13 +2560,17 @@ html_clueflow_set_indentation (HTMLClueFlow *flow,
 	if (indentation < 0)
 		indentation = 0;
 
+	next_relative = get_next_relative_item (HTML_OBJECT (flow));
+
 	g_byte_array_set_size (flow->levels, indentation);
 
 	i = indentation;
 	while (i--)
 		flow->levels->data[i] = indentation_levels[i];
 
-	update_items_after_indentation_change (flow);
+	update_item_number (HTML_OBJECT (flow), engine);
+	if (next_relative)
+		update_item_number (next_relative, engine);
 	relayout_with_siblings (flow, engine);
 }
 
@@ -2528,28 +2620,6 @@ html_clueflow_get_properties (HTMLClueFlow *flow,
 		*alignment_return = HTML_CLUE (flow)->halign;
 }
 #endif
-
-
-void
-html_clueflow_remove_text_slaves (HTMLClueFlow *flow)
-{
-	HTMLClue *clue;
-	HTMLObject *p;
-	HTMLObject *pnext;
-
-	g_return_if_fail (flow != NULL);
-
-	clue = HTML_CLUE (flow);
-	for (p = clue->head; p != NULL; p = pnext) {
-		pnext = p->next;
-
-		if (HTML_OBJECT_TYPE (p) == HTML_TYPE_TEXTSLAVE) {
-			html_clue_remove (clue, p);
-			html_object_destroy (p);
-		}
-	}
-}
-
 /* spell checking */
 
 #include "htmlinterval.h"
@@ -2682,9 +2752,12 @@ end_of_word (gchar *ct, gboolean cited)
 	gboolean cited2;
 
 	cited2 = FALSE;
-	while (*ct && (cn = e_unicode_get_utf8 (ct, &uc))
+	while (*ct
+	       && (uc = g_utf8_get_char (ct))
+	       && (cn = g_utf8_next_char (ct))
 	       && (html_selection_spell_word (uc, &cited2)
-		   || (!cited && cited2 && (cnn = e_unicode_get_utf8 (cn, &ucn)) && g_unichar_isalpha (ucn)))) {
+		   || (!cited && cited2
+		       && *cn && (cnn = g_utf8_next_char (cn)) && (ucn = g_utf8_get_char (cnn)) && g_unichar_isalpha (ucn)))) {
 		ct = cn;
 		cited2 = FALSE;
 	}
@@ -2716,7 +2789,7 @@ html_clueflow_spell_check (HTMLClueFlow *flow, HTMLEngine *e, HTMLInterval *inte
 
 	clue = HTML_CLUE (flow);
 	if (!e->widget->editor_api
-	    || !GTK_HTML_CLASS (GTK_OBJECT (e->widget)->klass)->properties->live_spell_check
+	    || !GTK_HTML_CLASS (GTK_WIDGET_GET_CLASS (e->widget))->properties->live_spell_check
 	    || !clue || !clue->tail)
 		return;
 
@@ -2836,3 +2909,25 @@ html_clueflow_tabs (HTMLClueFlow *flow, HTMLPainter *p)
 	return (flow && HTML_IS_CLUEFLOW (flow) && flow->style == HTML_CLUEFLOW_STYLE_PRE) || HTML_IS_PLAIN_PAINTER (p)
 		? TRUE : FALSE;
 }
+
+void
+html_clueflow_set_item_color (HTMLClueFlow *flow, HTMLColor *color)
+{
+	if (flow->item_color)
+		html_color_unref (flow->item_color);
+	if (color)
+		html_color_ref (color);
+	flow->item_color = color;
+}
+
+gboolean
+html_clueflow_style_equals (HTMLClueFlow *cf1, HTMLClueFlow *cf2)
+{
+	if (!cf1 || !cf2
+	    || !HTML_IS_CLUEFLOW (cf1) || !HTML_IS_CLUEFLOW (cf2)
+	    || cf1->style != cf2->style
+	    || (cf1->style == HTML_CLUEFLOW_STYLE_LIST_ITEM && cf1->item_type != cf2->item_type)
+	    || !is_levels_equal (cf1, cf2))
+		return FALSE;
+	return TRUE;
+}
