@@ -28,6 +28,8 @@
 #include "gtkhtml.h"
 #include "gtkhtml-embedded.h"
 #include "htmlembedded.h"
+#include "htmlframe.h"
+#include "htmliframe.h"
 #include "htmlpainter.h"
 #include "htmlengine.h"
 
@@ -68,18 +70,20 @@ draw (HTMLObject *o,
 	if (!element->widget)
 		return;
 
-	if (element->widget) {
-		new_x = GTK_LAYOUT (element->parent)->hadjustment->value + o->x + tx;
-		new_y = GTK_LAYOUT (element->parent)->vadjustment->value + o->y + ty - o->ascent;
+	if (element->parent) {
+		new_x = o->x + tx;
+		new_y = o->y + ty - o->ascent;
 		
 		if (element->widget->parent) {
 			if (new_x != element->abs_x || new_y != element->abs_y)
 				gtk_layout_move (GTK_LAYOUT(element->parent), element->widget, new_x, new_y);
-			else
+			else if (!GTK_HTML (element->parent)->engine->expose)
 				gtk_widget_queue_draw (element->widget);
 		}
+	
 		element->abs_x = new_x;
 		element->abs_y = new_y;
+		
 		if (!element->widget->parent)
 			gtk_layout_put (GTK_LAYOUT(element->parent), element->widget, new_x, new_y);
 	}
@@ -100,10 +104,10 @@ destroy (HTMLObject *o)
 		g_free(element->value);
 	if(element->widget) {
 		gtk_widget_hide (element->widget);
-		gtk_signal_disconnect_by_data (GTK_OBJECT (element->widget), element);
+		g_signal_handlers_disconnect_matched (element->widget, G_SIGNAL_MATCH_DATA, 0, 0, NULL, NULL, element);
 		if (element->changed_id > 0)
-			gtk_signal_disconnect (GTK_OBJECT (element->widget), element->changed_id);
-		gtk_object_set_data (GTK_OBJECT (element->widget), "embeddedelement", NULL);
+			g_signal_handler_disconnect (element->widget, element->changed_id);
+		g_object_set_data (G_OBJECT (element->widget), "embeddedelement", NULL);
 		if (element->widget->parent && element->parent) {
 			g_assert (element->widget->parent == element->parent);
 			gtk_container_remove (GTK_CONTAINER (element->parent), element->widget);
@@ -178,11 +182,10 @@ calc_size (HTMLObject *self, HTMLPainter *painter, GList **changed_objs)
 	return FALSE;
 }
 
-
-void
-html_embedded_reset (HTMLEmbedded *e)
+static gboolean
+accepts_cursor (HTMLObject *o)
 {
-	HTML_EMBEDDED_CLASS (HTML_OBJECT (e)->klass)->reset (e);
+	return TRUE;
 }
 
 static gchar *
@@ -190,6 +193,12 @@ encode (HTMLEmbedded *e)
 {
 	return g_strdup ("");
 }
+
+void
+html_embedded_reset (HTMLEmbedded *e)
+{
+	HTML_EMBEDDED_CLASS (HTML_OBJECT (e)->klass)->reset (e);
+}
 
 gchar *
 html_embedded_encode (HTMLEmbedded *e)
@@ -197,6 +206,12 @@ html_embedded_encode (HTMLEmbedded *e)
 	return HTML_EMBEDDED_CLASS (HTML_OBJECT (e)->klass)->encode (e);
 }
 
+void
+html_embedded_reparent (HTMLEmbedded *e, GtkWidget *new_parent)
+{
+	HTML_EMBEDDED_CLASS (HTML_OBJECT (e)->klass)->reparent (e, new_parent);
+}
+
 void
 html_embedded_set_form (HTMLEmbedded *e, HTMLForm *form)
 {
@@ -204,7 +219,7 @@ html_embedded_set_form (HTMLEmbedded *e, HTMLForm *form)
 }
 
 gchar *
-html_embedded_encode_string (gchar *str)
+html_embedded_encode_string (const gchar *str)
 {
         static gchar *safe = "$-._!*(),"; /* RFC 1738 */
         unsigned pos = 0;
@@ -273,6 +288,7 @@ html_embedded_class_init (HTMLEmbeddedClass *klass,
 	object_class->destroy = destroy;
 	object_class->copy = copy;
 	object_class->draw = draw;
+	object_class->accepts_cursor = accepts_cursor;
 	object_class->calc_size = calc_size;
 	object_class->calc_min_width = calc_min_width;
 
@@ -322,7 +338,7 @@ html_embedded_object_changed (GtkHTMLEmbedded *eb, HTMLEngine *e)
 {
 	HTMLObject *object;
 
-	object = gtk_object_get_data(GTK_OBJECT(eb), "embeddedelement");
+	object = HTML_OBJECT (g_object_get_data (G_OBJECT (eb), "embeddedelement"));
 	if (object)
 		html_object_calc_size (object, e->painter, FALSE);
 	
@@ -341,9 +357,8 @@ html_embedded_new_widget (GtkWidget *parent, GtkHTMLEmbedded *eb, HTMLEngine *en
 	/* pass em as the user_data so that the handler will disconnect 
 	 * when the object is destoyed
 	 */
-	gtk_signal_connect(GTK_OBJECT(eb), "button_press_event",
-			   GTK_SIGNAL_FUNC(html_embedded_grab_cursor), em);
-	em->changed_id = gtk_signal_connect (GTK_OBJECT (eb), "changed", html_embedded_object_changed, engine);
+	g_signal_connect(eb, "button_press_event", G_CALLBACK (html_embedded_grab_cursor), em);
+	em->changed_id = g_signal_connect (eb, "changed", G_CALLBACK (html_embedded_object_changed), engine);
 	/* printf ("id %u\n", em->changed_id); */
 
 	return em;
@@ -370,10 +385,9 @@ html_embedded_set_widget (HTMLEmbedded *emb, GtkWidget *w)
 	emb->widget = w;
 	
 	gtk_widget_show (w);
-	
-	gtk_object_set_data (GTK_OBJECT (w), "embeddedelement", emb);
-	gtk_signal_connect (GTK_OBJECT (w), "size_allocate",
-			    GTK_SIGNAL_FUNC (html_embedded_allocate), emb);
+
+	g_object_set_data (G_OBJECT (w), "embeddedelement", emb);
+	g_signal_connect (w, "size_allocate", G_CALLBACK (html_embedded_allocate), emb);
 }
 
 GtkWidget *
@@ -382,4 +396,33 @@ html_embedded_get_widget (HTMLEmbedded *e)
 	return e->widget;
 }
 
+gboolean
+html_object_is_embedded (HTMLObject *o)
+{
+	gboolean rv = FALSE;
 
+	switch (o->klass->type) {
+	case HTML_TYPE_EMBEDDED:
+	case HTML_TYPE_TEXTINPUT:
+	case HTML_TYPE_BUTTON:
+	case HTML_TYPE_IMAGEINPUT:
+	case HTML_TYPE_TEXTAREA:
+	case HTML_TYPE_HIDDEN:
+	case HTML_TYPE_RADIO:
+	case HTML_TYPE_CHECKBOX:
+	case HTML_TYPE_SELECT:
+	case HTML_TYPE_IFRAME:
+	case HTML_TYPE_FRAME:
+		rv = TRUE;
+	default:
+		;
+	}
+
+	return rv;
+}
+
+gboolean
+html_object_is_frame (HTMLObject *o)
+{
+	return HTML_IS_FRAME (o) || HTML_IS_IFRAME (o);
+}
