@@ -24,22 +24,23 @@
 #ifndef _HTMLENGINE_H_
 #define _HTMLENGINE_H_
 
+#include <glib-object.h>
 #include <gdk/gdktypes.h>
-#include <gtk/gtkobject.h>
 #include <gtk/gtkwidget.h>
 #include "gtkhtml-types.h"
 
 #include "htmltypes.h"
 #include "htmlenums.h"
 #include "htmlcursor.h"
-#include "htmlfontmanager.h"
 
+
 #define HTML_TYPE_ENGINE                 (html_engine_get_type ())
-#define HTML_ENGINE(obj)                 (GTK_CHECK_CAST ((obj), HTML_TYPE_ENGINE, HTMLEngine))
-#define HTML_ENGINE_CLASS(klass)         (GTK_CHECK_CLASS_CAST ((klass), HTML_TYPE_ENGINE, HTMLEngineClass))
-#define HTML_IS_ENGINE(obj)              (GTK_CHECK_TYPE ((obj), HTML_TYPE_ENGINE))
-#define HTML_IS_ENGINE_CLASS(klass)      (GTK_CHECK_CLASS_TYPE ((klass), HTML_TYPE_ENGINE))
+#define HTML_ENGINE(obj)                 (G_TYPE_CHECK_INSTANCE_CAST ((obj), HTML_TYPE_ENGINE, HTMLEngine))
+#define HTML_ENGINE_CLASS(klass)         (G_TYPE_CHECK_CLASS_CAST ((klass), HTML_TYPE_ENGINE, HTMLEngineClass))
+#define HTML_IS_ENGINE(obj)              (G_TYPE_CHECK_INSTANCE_TYPE ((obj), HTML_TYPE_ENGINE))
+#define HTML_IS_ENGINE_CLASS(klass)      (G_TYPE_CHECK_CLASS_TYPE ((klass), HTML_TYPE_ENGINE))
 
+
 /* FIXME extreme hideous ugliness in the following lines.  */
 
 #define LEFT_BORDER 10
@@ -50,7 +51,7 @@
 /* FIXME this needs splitting.  */
 
 struct _HTMLEngine {
-	GtkObject parent;
+	GObject parent;
 	HTMLDrawQueue *draw_queue;
 
 	HTMLPainter *painter;
@@ -69,6 +70,9 @@ struct _HTMLEngine {
 	guint       clipboard_len;
 	GList      *clipboard_stack;
 
+	HTMLObject *primary;
+	guint       primary_len;
+
 	/* Freeze counter.  When greater than zero, we never trigger relayouts
            nor repaints.  When going from nonzero to zero, we relayout and
            repaint everything.  */
@@ -114,12 +118,12 @@ struct _HTMLEngine {
 	gboolean eat_space;
 	gboolean allow_frameset;
 	gboolean newPage;
+	gboolean begin;
  
-	HTMLStack *font_style_stack; /* Font style stack, elements are GtkHTMLFontStyles.  */
-	HTMLStack *font_face_stack;
-	HTMLStack *color_stack;	/* Color stack, elements are GdkColors.  */
+	HTMLStack *span_stack;
 	HTMLStack *clueflow_style_stack; /* Clueflow style stack, elements are HTMLClueFlowStyles.  */
 	HTMLStack *frame_stack;
+	HTMLStack *body_stack;
 
 	gchar *url;
 	gchar *target;
@@ -153,8 +157,6 @@ struct _HTMLEngine {
 
         gpointer image_factory;
 
-	HTMLStack *glossaryStack; /* HTMLGlossaryEntry */
-
 	/*
 	 * This list holds strings which are displayed in the view,
 	 * but are not actually contained in the HTML source.
@@ -266,16 +268,16 @@ struct _HTMLEngine {
 
 	gboolean save_data;
 	gint saved_step_count;
+
+	gboolean expose;
 };
 
 /* must be forward referenced *sigh* */
 struct _HTMLEmbedded;
 
 struct _HTMLEngineClass {
-	GtkObjectClass parent_class;
-
-	HTMLFontManager font_manager [HTML_FONT_MANAGER_ID_N];
-
+	GObjectClass parent_class;
+	
 	void (* title_changed) (HTMLEngine *engine);
 	void (* set_base) (HTMLEngine *engine, const gchar *base);
 	void (* set_base_target) (HTMLEngine *engine, const gchar *base_target);
@@ -295,6 +297,8 @@ void        html_engine_realize       (HTMLEngine *engine,
 				       GdkWindow  *window);
 void        html_engine_unrealize     (HTMLEngine *engine);
 
+void      html_engine_saved     (HTMLEngine *e);
+gboolean  html_engine_is_saved  (HTMLEngine *e);
 
 /* Editability control.  */
 void      html_engine_set_editable  (HTMLEngine *e,
@@ -328,6 +332,8 @@ void  html_engine_draw                 (HTMLEngine *e,
 					gint        y,
 					gint        width,
 					gint        height);
+void  html_engine_expose               (HTMLEngine *e,
+					GdkEventExpose *event);
 void  html_engine_draw_background      (HTMLEngine *e,
 					gint        x,
 					gint        y,
@@ -429,7 +435,7 @@ HTMLObject *html_engine_get_object_by_id    (HTMLEngine  *e,
 					     const gchar *id);
 
 HTMLEngine *html_engine_get_top_html_engine (HTMLEngine *e);
-void        html_engine_thaw_idle_reset     (HTMLEngine *e);
+void        html_engine_thaw_idle_flush     (HTMLEngine *e);
 
 /* class data */
 const gchar *html_engine_get_class_data        (HTMLEngine  *e,
@@ -454,7 +460,8 @@ void  html_engine_add_expose  (HTMLEngine *e,
 			       gint        x,
 			       gint        y,
 			       gint        width,
-			       gint        height);
+			       gint        height,
+			       gboolean    expose);
 void html_engine_redraw_selection (HTMLEngine *e);
 
 gboolean    html_engine_focus              (HTMLEngine       *e,
@@ -464,16 +471,11 @@ void        html_engine_set_focus_object   (HTMLEngine       *e,
 					    HTMLObject       *o);
 void        html_engine_draw_focus_object  (HTMLEngine       *e);
 
-void      html_engine_saved     (HTMLEngine *e);
-gboolean  html_engine_is_saved  (HTMLEngine *e);
-
 HTMLMap *html_engine_get_map  (HTMLEngine  *e,
 			       const gchar *name);
 
-HTMLFontManager *html_engine_gdk_font_manager           (HTMLEngine  *e);
-HTMLFontManager *html_engine_font_manager               (HTMLEngine  *e);
-HTMLFontManager *html_engine_font_manager_with_painter  (HTMLEngine  *e,
-							 HTMLPainter *p);
-HTMLFontManager *html_engine_class_gdk_font_manager     (void);
+gboolean html_engine_selection_contains_object_type (HTMLEngine *e,
+						     HTMLType obj_type);
+gboolean html_engine_selection_contains_link        (HTMLEngine *e);
 
 #endif /* _HTMLENGINE_H_ */
