@@ -36,8 +36,8 @@
 #include "htmlsettings.h"
 #include "htmltokenizer.h"
 #include "htmlembedded.h"
-#include <libgnome/gnome-i18n.h>
 
+#define USE_SCROLLED_WINDOW 1
 #ifndef USE_SCROLLED_WINDOW
 #include <gal/widgets/e-scroll-frame.h>
 #endif
@@ -61,8 +61,7 @@ iframe_url_requested (GtkHTML *html, const char *url, GtkHTMLStream *handle, gpo
 	HTMLIFrame *iframe = HTML_IFRAME (data);
 	GtkHTML *parent = GTK_HTML (HTML_EMBEDDED(iframe)->parent);
 
-	gtk_signal_emit_by_name (GTK_OBJECT (parent->engine), "url_requested",
-				 url, handle);
+	g_signal_emit_by_name (parent->engine, "url_requested", url, handle);
 }
 
 static void
@@ -82,7 +81,7 @@ iframe_object_requested (GtkHTML *html, GtkHTMLEmbedded *eb, gpointer data)
 	gboolean ret_val;
 
 	ret_val = FALSE;
-	gtk_signal_emit_by_name (GTK_OBJECT (parent), "object_requested", eb, &ret_val);
+	g_signal_emit_by_name (parent, "object_requested", eb, &ret_val);
 	return ret_val;
 }
 
@@ -90,10 +89,10 @@ static void
 iframe_set_gdk_painter (HTMLIFrame *iframe, HTMLPainter *painter)
 {
 	if (painter)
-		gtk_object_ref (GTK_OBJECT (painter));
+		g_object_ref (G_OBJECT (painter));
 	
 	if (iframe->gdk_painter)
-		gtk_object_unref (GTK_OBJECT (iframe->gdk_painter));
+		g_object_unref (G_OBJECT (iframe->gdk_painter));
 
 	iframe->gdk_painter = painter;
 }
@@ -179,12 +178,12 @@ draw (HTMLObject *o,
 {
 	HTMLIFrame   *iframe  = HTML_IFRAME (o);
 	HTMLEngine   *e       = GTK_HTML (iframe->html)->engine;
-	ArtIRect paint;
+	GdkRectangle paint;
 
 	if (GTK_OBJECT_TYPE (e->painter) == HTML_TYPE_PRINTER) {
 		gint pixel_size = html_painter_get_pixel_size (e->painter);
-		html_object_calc_intersection (o, &paint, x, y, width, height);
-		if (art_irect_empty (&paint))
+
+		if (!html_object_intersect (o, &paint, x, y, width, height))
 			return;
 
 		html_object_draw (e->clue, e->painter,
@@ -202,12 +201,12 @@ set_painter (HTMLObject *o, HTMLPainter *painter)
 	HTMLIFrame *iframe;
 
 	iframe = HTML_IFRAME (o);
-	if (GTK_OBJECT_TYPE (GTK_HTML (iframe->html)->engine->painter) != HTML_TYPE_PRINTER) {
+	if (G_OBJECT_TYPE (GTK_HTML (iframe->html)->engine->painter) != HTML_TYPE_PRINTER) {
 		iframe_set_gdk_painter (iframe, GTK_HTML (iframe->html)->engine->painter);
 	}
 	
 	html_engine_set_painter (GTK_HTML (iframe->html)->engine,
-				 GTK_OBJECT_TYPE (painter) != HTML_TYPE_PRINTER ? iframe->gdk_painter : painter);
+				 G_OBJECT_TYPE (painter) != HTML_TYPE_PRINTER ? iframe->gdk_painter : painter);
 }
 
 static void
@@ -261,8 +260,6 @@ copy (HTMLObject *self,
 	d->html = gtk_html_new ();
 
 	d->gdk_painter = NULL;
-	d->old_painter = NULL;
-	d->parent_painter = NULL;
 
 	d->url = g_strdup (s->url);
 	d->width = s->width;
@@ -313,7 +310,7 @@ html_iframe_set_margin_height (HTMLIFrame *iframe, gint margin_height)
 void
 html_iframe_set_scrolling (HTMLIFrame *iframe, GtkPolicyType scroll)
 {
-#if E_USE_SCROLLED_WINDOW
+#if USE_SCROLLED_WINDOW
 	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (iframe->scroll),
 					scroll, scroll);
 #else
@@ -347,8 +344,7 @@ calc_size (HTMLObject *o, HTMLPainter *painter, GList **changed_objs)
 		height = html_engine_get_doc_height (e);
 		width = html_engine_get_doc_width (e);
 
-		/* FIXME: fix frames for height > 32767 */
-		gtk_widget_set_usize (iframe->scroll, width, MIN (height, 32767));
+		gtk_widget_set_size_request (iframe->scroll, width, height);
 		gtk_widget_queue_resize (iframe->scroll);
 		
 		html_iframe_set_scrolling (iframe, GTK_POLICY_NEVER);
@@ -566,13 +562,7 @@ destroy (HTMLObject *o)
 	g_free (iframe->url);
 
 	if (iframe->html) {
-		if (iframe->old_painter)
-			gtk_object_unref (GTK_OBJECT (iframe->old_painter));
-		if (iframe->parent_painter)
-			gtk_object_unref (GTK_OBJECT (iframe->parent_painter));
-		iframe->old_painter = iframe->parent_painter = NULL;
-
-		gtk_signal_disconnect_by_data (GTK_OBJECT (iframe->html), o);
+		g_signal_handlers_disconnect_matched (iframe->html, G_SIGNAL_MATCH_DATA, 0, 0, NULL, NULL, o);
 		iframe->html = NULL;
 	}
 
@@ -611,6 +601,11 @@ html_iframe_init (HTMLIFrame *iframe,
 					border ? GTK_SHADOW_IN : GTK_SHADOW_NONE);
 
 #endif
+	/*
+	 * FIXME
+	 * are we missing:
+	 * gtk_widget_show (scrolled_window); here?
+	 */
 
 	iframe->scroll = scrolled_window;
 	html_iframe_set_scrolling (iframe, GTK_POLICY_AUTOMATIC);
@@ -618,16 +613,10 @@ html_iframe_init (HTMLIFrame *iframe,
 	new_widget = gtk_html_new ();
 	new_html = GTK_HTML (new_widget);
 
-	/* use parent_html font_manager */
-	gtk_object_ref (GTK_OBJECT (new_html->engine->painter));
-	gtk_object_ref (GTK_OBJECT (parent_html->engine->painter));
-	iframe->old_painter = new_html->engine->painter;
-	iframe->parent_painter = parent_html->engine->painter;
-
 	new_tokenizer = html_tokenizer_clone (parent_html->engine->ht);
 
 	html_engine_set_tokenizer (new_html->engine, new_tokenizer);
-	gtk_object_unref (GTK_OBJECT (new_tokenizer));
+	g_object_unref (G_OBJECT (new_tokenizer));
 	new_tokenizer = NULL;
 
 	gtk_html_set_default_content_type (new_html,
@@ -647,10 +636,8 @@ html_iframe_init (HTMLIFrame *iframe,
 	handle = gtk_html_begin (new_html);
 
 	new_html->engine->clue->parent = HTML_OBJECT (iframe);
-
-	gtk_signal_connect (GTK_OBJECT (new_html), "url_requested",
-			    GTK_SIGNAL_FUNC (iframe_url_requested),
-			    (gpointer)iframe);
+	
+	g_signal_connect (new_html, "url_requested", G_CALLBACK (iframe_url_requested), iframe);
 #if 0
 	/* NOTE: because of peculiarities of the frame/gtkhtml relationship
 	 * on_url and link_clicked are emitted from the toplevel widget not
@@ -663,37 +650,24 @@ html_iframe_init (HTMLIFrame *iframe,
 			    GTK_SIGNAL_FUNC (iframe_link_clicked),
 			    (gpointer)iframe);	
 #endif 
-	gtk_signal_connect (GTK_OBJECT (new_html), "size_changed",
-			    GTK_SIGNAL_FUNC (iframe_size_changed),
-			    (gpointer)iframe);	
-	gtk_signal_connect (GTK_OBJECT (new_html), "set_base",
-			    GTK_SIGNAL_FUNC (iframe_set_base),
-			    (gpointer)iframe);	
-	gtk_signal_connect (GTK_OBJECT (new_html), "object_requested",
-			    GTK_SIGNAL_FUNC (iframe_object_requested),
-			    (gpointer)iframe);	
+	g_signal_connect (new_html, "size_changed", G_CALLBACK (iframe_size_changed), iframe);	
+	g_signal_connect (new_html, "set_base", G_CALLBACK (iframe_set_base), iframe);	
+	g_signal_connect (new_html, "object_requested", G_CALLBACK (iframe_object_requested), iframe);	
 
 	/*
 	  gtk_signal_connect (GTK_OBJECT (html), "button_press_event",
 	  GTK_SIGNAL_FUNC (iframe_button_press_event), iframe);
 	*/
+	if (depth < 10)
+		g_signal_emit_by_name (parent_html->engine, "url_requested", src, handle);
+	
+	gtk_widget_set_size_request (scrolled_window, width, height);
 
-	if (depth < 10) {
-		gtk_signal_emit_by_name (GTK_OBJECT (parent_html->engine), 
-					 "url_requested", src, handle);
-	} else {
-		gtk_html_stream_printf (handle, "%s", _("Error: maximium frame depth exceded"));
-		gtk_html_stream_close (handle, GTK_HTML_STREAM_OK);
-	}
-
-	/* FIXME: fix frames for height > 32767 */
-	gtk_widget_set_usize (scrolled_window, width, MIN (height, 32767));
 	gtk_widget_show (scrolled_window);	
 
 	html_embedded_set_widget (em, scrolled_window);
 
-	gtk_signal_connect(GTK_OBJECT (scrolled_window), "button_press_event",
-			   GTK_SIGNAL_FUNC (html_iframe_grab_cursor), NULL);
+	g_signal_connect (scrolled_window, "button_press_event", G_CALLBACK (html_iframe_grab_cursor), NULL);
 
 	/* inherit the current colors from our parent */
 	html_colorset_set_unchanged (new_html->engine->defaultSettings->color_set,
