@@ -21,14 +21,6 @@
    the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
    Boston, MA 02111-1307, USA.
 
-   TODO:
-
-     - implement proper animation loading (now it loops thru loaded frames
-       and does not stop when loading is in progress and we are out of frames)
-     - look at gdk-pixbuf to make gdk_pixbuf_compose work (look also
-       on gk_pixbuf_render_to_drawable_alpha)
-     - take care about all the frame->action values
-
 */
 
 #include <config.h>
@@ -42,6 +34,7 @@
 #include "gtkhtml-properties.h"
 #include "gtkhtml-stream.h"
 
+#include "htmlclueflow.h"
 #include "htmlcolor.h"
 #include "htmlcolorset.h"
 #include "htmldrawqueue.h"
@@ -53,6 +46,7 @@
 #include "htmlmap.h"
 #include "htmlprinter.h"
 #include "htmlgdkpainter.h"
+#include "htmlplainpainter.h"
 
 /* HTMLImageFactory stuff.  */
 
@@ -63,49 +57,43 @@ struct _HTMLImageFactory {
 
 
 #define DEFAULT_SIZE 48
-#define ANIMATIONS(i) GTK_HTML_CLASS (GTK_OBJECT (i->image_ptr->factory->engine->widget)->klass)->properties->animations
+#define ANIMATIONS(i) GTK_HTML_CLASS (GTK_WIDGET_GET_CLASS (i->image_ptr->factory->engine->widget))->properties->animations
 #define STRDUP_HELPER(i,j) if (i != j) {char *tmp = g_strdup (j); g_free(i); i = tmp;}
 
 HTMLImageClass html_image_class;
 static HTMLObjectClass *parent_class = NULL;
 
-static HTMLImageAnimation *html_image_animation_new     (HTMLImage *image);
-static void                html_image_animation_destroy (HTMLImageAnimation *anim);
-static HTMLImagePointer   *html_image_pointer_new       (const char *filename, HTMLImageFactory *factory);
-static void                html_image_pointer_ref       (HTMLImagePointer *ip);
-static void                html_image_pointer_unref     (HTMLImagePointer *ip);
-static gboolean            html_image_pointer_timeout   (HTMLImagePointer *ip);
-static void                render_cur_frame             (HTMLImage *image, gint nx, gint ny, const GdkColor *highlight_color);
-static guint               get_actual_height            (HTMLImage *image, HTMLPainter *painter);
-
+static HTMLImagePointer   *html_image_pointer_new               (const char *filename, HTMLImageFactory *factory);
+static void                html_image_pointer_ref               (HTMLImagePointer *ip);
+static void                html_image_pointer_unref             (HTMLImagePointer *ip);
+static gboolean            html_image_pointer_timeout           (HTMLImagePointer *ip);
+static gint                html_image_pointer_run_animation     (HTMLImagePointer *ip);
+static void                html_image_pointer_start_animation   (HTMLImagePointer *ip);
 
-static guint
-get_actual_width (HTMLImage *image,
-		  HTMLPainter *painter)
+guint
+html_image_get_actual_width (HTMLImage *image, HTMLPainter *painter)
 {
-	GdkPixbuf *pixbuf = image->image_ptr->pixbuf;
 	GdkPixbufAnimation *anim = image->image_ptr->animation;
+	gint pixel_size = painter ? html_painter_get_pixel_size (painter) : 1;
 	gint width;
 
 	if (image->percent_width) {
 		/* The cast to `gdouble' is to avoid overflow (eg. when
                    printing).  */
-		width = ((gdouble) html_engine_get_view_width (image->image_ptr->factory->engine)
+		width = ((gdouble) HTML_OBJECT (image)->max_width
 			 * image->specified_width) / 100;
 	} else if (image->specified_width > 0) {
-		width = image->specified_width * html_painter_get_pixel_size (painter);
-	} else if (image->image_ptr == NULL || pixbuf == NULL) {
-		width = DEFAULT_SIZE * html_painter_get_pixel_size (painter);
+		width = image->specified_width * pixel_size;
+	} else if (image->image_ptr == NULL || anim == NULL) {
+		width = DEFAULT_SIZE * pixel_size;
 	} else {
-		width = (((anim) ? gdk_pixbuf_animation_get_width (anim) : gdk_pixbuf_get_width (pixbuf))
-			  * html_painter_get_pixel_size (painter));
+		width = gdk_pixbuf_animation_get_width (anim) * pixel_size;
 
 		if (image->specified_height > 0 || image->percent_height) {
 			double scale;
 
-			scale =  ((double) get_actual_height (image, painter)) 
-				/ (((anim) ? gdk_pixbuf_animation_get_height (anim) : gdk_pixbuf_get_height (pixbuf))
-				   * html_painter_get_pixel_size (painter));
+			scale =  ((double) html_image_get_actual_height (image, painter)) 
+				/ (gdk_pixbuf_animation_get_height (anim) * pixel_size);
 			
 			width *= scale;
 		}
@@ -115,13 +103,11 @@ get_actual_width (HTMLImage *image,
 	return width;
 }
 
-
-static guint
-get_actual_height (HTMLImage *image,
-		   HTMLPainter *painter)
+guint
+html_image_get_actual_height (HTMLImage *image, HTMLPainter *painter)
 {
-	GdkPixbuf *pixbuf = image->image_ptr->pixbuf;
 	GdkPixbufAnimation *anim = image->image_ptr->animation;
+	gint pixel_size = painter ? html_painter_get_pixel_size (painter) : 1;
 	gint height;
 		
 	if (image->percent_height) {
@@ -130,19 +116,17 @@ get_actual_height (HTMLImage *image,
 		height = ((gdouble) html_engine_get_view_height (image->image_ptr->factory->engine)
 			  * image->specified_height) / 100;
 	} else if (image->specified_height > 0) {
-		height = image->specified_height * html_painter_get_pixel_size (painter);
-	} else if (image->image_ptr == NULL || pixbuf == NULL) {
-		height = DEFAULT_SIZE * html_painter_get_pixel_size (painter);
+		height = image->specified_height * pixel_size;
+	} else if (image->image_ptr == NULL || anim == NULL) {
+		height = DEFAULT_SIZE * pixel_size;
 	} else {
-		height = (((anim) ? gdk_pixbuf_animation_get_height (anim) : gdk_pixbuf_get_height (pixbuf))
-			  * html_painter_get_pixel_size (painter));
+		height = gdk_pixbuf_animation_get_height (anim) * pixel_size;
 
 		if (image->specified_width > 0 || image->percent_width) {
 			double scale;
 			
-			scale = ((double) get_actual_width (image, painter))
-				/ (((anim) ? gdk_pixbuf_animation_get_width (anim) : gdk_pixbuf_get_width (pixbuf))
-				   * html_painter_get_pixel_size (painter));
+			scale = ((double) html_image_get_actual_width (image, painter))
+				/ (gdk_pixbuf_animation_get_width (anim) * pixel_size);
 			
 			height *= scale;
 		} 
@@ -166,9 +150,6 @@ destroy (HTMLObject *o)
 	html_image_factory_unregister (image->image_ptr->factory,
 				       image->image_ptr, HTML_IMAGE (image));
 
-	if (image->animation)
-		html_image_animation_destroy (image->animation);
-
 	g_free (image->url);
 	g_free (image->target);
 	g_free (image->alt);
@@ -193,7 +174,6 @@ copy (HTMLObject *self,
 	(* HTML_OBJECT_CLASS (parent_class)->copy) (self, dest);
 
 	dimg->image_ptr = simg->image_ptr;
-	dimg->animation = NULL;          /* don't bother with animation copying now. TODO */
 	dimg->color = simg->color;
 	if (simg->color)
 		html_color_ref (dimg->color);
@@ -237,8 +217,8 @@ image_update_url (HTMLImage *image, gint x, gint y)
 	 * it is sick, I know.
 	 */
 	if (image->usemap != NULL) {
-		map = (HTMLMap *)html_engine_get_object_by_id (image->image_ptr->factory->engine, 
-							       image->usemap + 1);
+		map = html_engine_get_map (image->image_ptr->factory->engine, 
+					   image->usemap + 1);
 		
 		if (map) {
 			url = html_map_calc_point (map, x - o->x , y - (o->y - o->ascent));
@@ -291,7 +271,7 @@ calc_min_width (HTMLObject *o,
 	if (image->percent_width || image->percent_height)
 		min_width = pixel_size;
 	else
-		min_width = get_actual_width (HTML_IMAGE (o), painter);
+		min_width = html_image_get_actual_width (HTML_IMAGE (o), painter);
 
 	min_width += (image->border * 2 + 2 * image->hspace) * pixel_size;
 
@@ -305,7 +285,7 @@ calc_preferred_width (HTMLObject *o,
 	HTMLImage *image = HTML_IMAGE (o);
 	guint width;
 
-	width = get_actual_width (HTML_IMAGE (o), painter)
+	width = html_image_get_actual_width (HTML_IMAGE (o), painter)
 		+ (image->border * 2 + 2 * image->hspace) * html_painter_get_pixel_size (painter);
 
 	return width;
@@ -327,12 +307,23 @@ calc_size (HTMLObject *o, HTMLPainter *painter, GList **changed_objs)
 
 	pixel_size = html_painter_get_pixel_size (painter);
 
-	width = get_actual_width (image, painter);
-	height = get_actual_height (image, painter);
+	if (o->parent && HTML_IS_CLUEFLOW (o->parent)
+	    && HTML_IS_PLAIN_PAINTER (painter) && image->alt && *image->alt) {
+		GtkHTMLFontStyle style;
+		gint lo = 0;
 
-	o->width  = width + (image->border + image->hspace) * 2 * pixel_size;
-	o->ascent = height + (image->border + image->vspace) * 2 * pixel_size;
-	o->descent = 0;
+		style = html_clueflow_get_default_font_style (HTML_CLUEFLOW (o->parent));
+		/* FIXME: cache items and glyphs? */
+		html_painter_calc_text_size (painter, image->alt, g_utf8_strlen (image->alt, -1), NULL, NULL, &lo,
+					     style, NULL, &o->width, &o->ascent, &o->descent);
+	} else {
+		width = html_image_get_actual_width (image, painter);
+		height = html_image_get_actual_height (image, painter);
+
+		o->width  = width + (image->border + image->hspace) * 2 * pixel_size;
+		o->ascent = height + (image->border + image->vspace) * 2 * pixel_size;
+		o->descent = 0;
+	}
 
 	if (o->descent != old_descent
 	    || o->ascent != old_ascent
@@ -342,6 +333,18 @@ calc_size (HTMLObject *o, HTMLPainter *painter, GList **changed_objs)
 	return FALSE;
 }
 
+static void
+draw_plain (HTMLObject *o, HTMLPainter *p, gint x, gint y, gint width, gint height, gint tx, gint ty)
+{
+	HTMLImage *img = HTML_IMAGE (o);
+
+	if (img->alt && *img->alt) {
+		html_painter_set_pen (p, &html_colorset_get_color_allocated (p, HTMLTextColor)->color);
+		/* FIXME: cache items and glyphs? */
+		html_painter_draw_text (p, o->x + tx, o->y + ty - o->ascent, img->alt, g_utf8_strlen (img->alt, -1), NULL, NULL, 0);
+	}
+}
+
 static void
 draw (HTMLObject *o,
       HTMLPainter *painter,
@@ -350,22 +353,37 @@ draw (HTMLObject *o,
       gint tx, gint ty)
 {
 	HTMLImage *image;
+	HTMLImagePointer *ip;
 	GdkPixbuf *pixbuf;
 	gint base_x, base_y;
 	gint scale_width, scale_height;
 	GdkColor *highlight_color;
 	guint pixel_size;
-	ArtIRect paint;
+	GdkRectangle paint;
 
 	/* printf ("Image::draw\n"); */
 
-	html_object_calc_intersection (o, &paint, x, y, width, height);
-	if (art_irect_empty (&paint))
+	if (!html_object_intersect (o, &paint, x, y, width, height))
 		return;
 
+	if (HTML_IS_PLAIN_PAINTER (painter)) {
+	draw_plain (o, painter, x, y, width, height, tx, ty);
+		return;
+	}
+
 	image = HTML_IMAGE (o);
+	ip = image->image_ptr;
+
+	if (ip->animation) {
+		if (HTML_IS_GDK_PAINTER (painter) && !gdk_pixbuf_animation_is_static_image (ip->animation)) {
+			pixbuf = gdk_pixbuf_animation_iter_get_pixbuf (ip->iter);
+		} else {
+			pixbuf = gdk_pixbuf_animation_get_static_image (ip->animation);
+		}
+	} else {
+		pixbuf = NULL;
+	}
 
-	pixbuf = image->image_ptr->pixbuf;
 	pixel_size = html_painter_get_pixel_size (painter);
 
 	if (o->selected) {
@@ -400,11 +418,12 @@ draw (HTMLObject *o,
 					 GTK_HTML_ETCH_IN, 1);
 		return;
 	}
+
 	base_x = o->x + tx + (image->border + image->hspace) * pixel_size;
 	base_y = o->y + ty + (image->border + image->vspace) * pixel_size - o->ascent;
 
-	scale_width = get_actual_width (image, painter);
-	scale_height = get_actual_height (image, painter);
+	scale_width = html_image_get_actual_width (image, painter);
+	scale_height = html_image_get_actual_height (image, painter);
 
 	if (image->border) {
 		if (image->have_color) {
@@ -421,20 +440,12 @@ draw (HTMLObject *o,
 					 GTK_HTML_ETCH_NONE, image->border);
 		
 	}
-	if (ANIMATIONS (image) &&  image->animation && ! HTML_IS_PRINTER (painter)) {
-		image->animation->active = TRUE;
-		image->animation->x = base_x;
-		image->animation->y = base_y;
-		image->animation->ex = image->image_ptr->factory->engine->x_offset;
-		image->animation->ey = image->image_ptr->factory->engine->y_offset;
-
-		render_cur_frame (image, base_x, base_y, highlight_color);
-	} else {
-		html_painter_draw_pixmap (painter, pixbuf,
-					  base_x, base_y,
-					  scale_width, scale_height,
-					  highlight_color);
-	}
+	
+	image->animation_active = TRUE;
+	html_painter_draw_pixmap (painter, pixbuf,
+				  base_x, base_y,
+				  scale_width, scale_height,
+				  highlight_color);
 }
 
 gchar *
@@ -444,19 +455,22 @@ html_image_resolve_image_url (GtkHTML *html, gchar *image_url)
 
 	/* printf ("html_image_resolve_image_url %p\n", html->editor_api); */
 	if (html->editor_api) {
-		GtkArg *args [2];
+		GValue  *iarg = g_new0 (GValue, 1);
+		GValue  *oarg;
 
-		args [0] = gtk_arg_new (GTK_TYPE_STRING);
-		GTK_VALUE_STRING (*args [0]) = image_url;
+		g_value_init (iarg, G_TYPE_STRING);
+		g_value_set_string (iarg, image_url);
 
-		args [1] = (* html->editor_api->event) (html, GTK_HTML_EDITOR_EVENT_IMAGE_URL, args, html->editor_data);
-		gtk_arg_free (args [0], FALSE);
+		oarg = (* html->editor_api->event) (html, GTK_HTML_EDITOR_EVENT_IMAGE_URL, iarg, html->editor_data);
 
-		if (args [1]) {
-			if (args [1]->type == GTK_TYPE_STRING)
-				url = GTK_VALUE_STRING (*args [1]);
-			gtk_arg_free (args [1], FALSE);
+		if (oarg) {
+			if (G_VALUE_TYPE (oarg) == G_TYPE_STRING)
+				url = (gchar *) g_value_get_string (oarg);
+			g_value_unset (iarg);	
+			g_free (oarg);
 		}
+		g_value_unset (iarg);
+		g_free (iarg);
 	}
 	if (!url)
 		url = g_strdup (image_url);
@@ -471,13 +485,22 @@ save (HTMLObject *self,
 {
 	HTMLImage *image;
 	gchar *url;
-	gboolean result;
+	gboolean result, link = FALSE;
 
 	g_return_val_if_fail (self != NULL, FALSE);
 	g_return_val_if_fail (state != NULL, FALSE);
 
 	image  = HTML_IMAGE (self);
 
+	if (image->url && *image->url) {
+		url  = g_strconcat (image->url, image->target ? "#" : "", image->target, NULL);
+		link = TRUE;
+		result = html_engine_save_output_string (state, "<A HREF=\"%s\">", url);
+		g_free (url);
+		if (!result)
+			return FALSE;	
+	}
+
 	url    = html_image_resolve_image_url (state->engine->widget, image->image_ptr->url);
 	result = html_engine_save_output_string (state, "<IMG SRC=\"%s\"", url);
 	g_free (url);
@@ -535,6 +558,8 @@ save (HTMLObject *self,
 
 	if (!html_engine_save_output_string (state, ">"))
 		return FALSE;
+	if (link && !html_engine_save_output_string (state, "</A>"))
+		return FALSE;
 	
 	return TRUE;
 }
@@ -589,6 +614,15 @@ set_link (HTMLObject *self, HTMLColor *color, const gchar *url, const gchar *tar
 
 	STRDUP_HELPER (image->url, url);
 	STRDUP_HELPER (image->target, target);
+	if (image->have_color)
+		html_color_unref (image->color);
+	image->color = color;
+	if (color) {
+		html_color_ref (color);
+		image->have_color = TRUE;
+	} else {
+		image->have_color = FALSE;
+	}
 
 	return NULL;
 }
@@ -616,9 +650,12 @@ select_range (HTMLObject *self,
 	      gint length,
 	      gboolean queue_draw)
 {
+	/* printf ("IMAGE: select range\n"); */
 	if ((*parent_class->select_range) (self, engine, offset, length, queue_draw)) {
-		if (queue_draw)
+		if (queue_draw) {
 			html_engine_queue_draw (engine, self);
+			/* printf ("IMAGE: draw queued\n"); */
+		}
 		return TRUE;
 	} else
 		return FALSE;
@@ -705,7 +742,6 @@ html_image_init (HTMLImage *image,
 		image->have_color = FALSE;
 	}
 
-	image->animation = NULL;
 	image->alt = NULL;
 
 	image->hspace = 0;
@@ -868,16 +904,50 @@ html_image_set_size (HTMLImage *image, gint w, gint h, gboolean pw, gboolean ph)
 	}
 }
 
-char *image_content_types[] = {"image/*", NULL};
+static char *fallback_image_content_types[] = {"image/*", NULL};
 
 static char **
 html_image_factory_types (GtkHTMLStream *stream,
 			  gpointer user_data)
 {
-	/* FIXME: this should use the not currently existant function
-	 * in gdk-pixbuf that tells us what mime types are supported 
-	 * by the modules it has loaded.
-	 */
+	static char**image_content_types = NULL;
+	
+#if 0
+	/* this code should work in gtk+-2.2 but it is untested */
+	if (image_content_types == NULL) {
+		GSList *formats;
+		GSList *cur;
+		GSList *types = NULL;
+		gint i;
+
+		formats = gdk_pixbuf_get_formats ();
+		
+		for (cur = formats; cur; cur = cur->next) {
+			GdkPixbufFormat *format = cur->data;
+			char **mime;
+
+			mime = gdk_pixbuf_formats_get_mime_types ();
+			for (i = 0; mime && mime[i]; i++)
+				g_slist_prepend (types, g_strdup (mime[i]));
+
+		}
+		g_slist_free (formats);
+
+		if (types) {
+			image_content_types = g_new0 (char *, g_slist_length (types) + 1);
+			
+			for (cur = types, i = 0; cur; cur = cur->next, i++) {
+				image_content_types[i] = cur->data;
+			}
+			g_slist_free (types);
+		} else {
+			image_content_types = fallback_image_content_types;
+		}
+	}
+#else 
+	image_content_types = fallback_image_content_types;
+#endif
+
 	return image_content_types;
 }
 
@@ -895,9 +965,9 @@ update_or_redraw (HTMLImagePointer *ip)
 			gint pixel_size = html_painter_get_pixel_size (ip->factory->engine->painter);
 			gint w, h;
 
-			w = get_actual_width (image, ip->factory->engine->painter)
+			w = html_image_get_actual_width (image, ip->factory->engine->painter)
 				+ (image->border * 2 + 2 * image->hspace) * pixel_size;
-			h = get_actual_height (image, ip->factory->engine->painter)
+			h = html_image_get_actual_height (image, ip->factory->engine->painter)
 				+ (image->border * 2 + 2 * image->vspace) * pixel_size;
 
 			/* printf ("%dx%d  <-->  %dx%d\n", w, h, HTML_OBJECT (list->data)->width,
@@ -911,10 +981,13 @@ update_or_redraw (HTMLImagePointer *ip)
 		}
 	}
 
+	if (ip->factory->engine->block && ip->factory->engine->opened_streams)
+		return;
+
 	if (!update) {
 		/* printf ("REDRAW\n"); */
 		for (list = ip->interests; list; list = list->next)
-			if (list->data) // && html_object_is_visible (HTML_OBJECT (list->data)))
+			if (list->data) /* && html_object_is_visible (HTML_OBJECT (list->data))) */
 				html_engine_queue_draw (ip->factory->engine, HTML_OBJECT (list->data));
 		if (ip->interests)
 			html_engine_flush_draw_queue (ip->factory->engine);
@@ -931,16 +1004,25 @@ html_image_factory_end_pixbuf (GtkHTMLStream *stream,
 {
 	HTMLImagePointer *ip = user_data;
 
-	gdk_pixbuf_loader_close (ip->loader);
-	if (!ip->animation && !ip->pixbuf) {
-		ip->pixbuf = gdk_pixbuf_loader_get_pixbuf (ip->loader);
+	gdk_pixbuf_loader_close (ip->loader, NULL);
 
-		if (ip->pixbuf)
-			gdk_pixbuf_ref (ip->pixbuf);
+	if (!ip->animation) {
+		ip->animation = gdk_pixbuf_loader_get_animation (ip->loader);
+
+		if (ip->animation)
+			g_object_ref (ip->animation);
 	}
+	html_image_pointer_start_animation (ip);
+
+	g_object_unref (ip->loader);
+	ip->loader = NULL;
 
 	update_or_redraw (ip);
-	
+	if (ip->factory->engine->opened_streams)
+		ip->factory->engine->opened_streams --;
+	/* printf ("IMAGE(%p) opened streams: %d\n", ip->factory->engine, ip->factory->engine->opened_streams); */
+	if (ip->factory->engine->opened_streams == 0 && ip->factory->engine->block)
+		html_engine_schedule_update (ip->factory->engine);
 	html_image_pointer_unref (ip);
 }
 
@@ -953,7 +1035,7 @@ html_image_factory_write_pixbuf (GtkHTMLStream *stream,
 	HTMLImagePointer *p = user_data;
 
 	/* FIXME ! Check return value */
-	gdk_pixbuf_loader_write (p->loader, buffer, size);
+	gdk_pixbuf_loader_write (p->loader, buffer, size, NULL);
 }
 
 static void
@@ -966,179 +1048,69 @@ static void
 html_image_factory_area_prepared (GdkPixbufLoader *loader, HTMLImagePointer *ip)
 {
 	if (!ip->animation) {
-		ip->pixbuf = gdk_pixbuf_loader_get_pixbuf (ip->loader);
-		g_assert (ip->pixbuf);
-
-		gdk_pixbuf_ref (ip->pixbuf);
-		update_or_redraw (ip);
+		ip->animation = gdk_pixbuf_loader_get_animation (loader);
+		g_object_ref (ip->animation);
+		
+		html_image_pointer_start_animation (ip);
 	}
+	update_or_redraw (ip);
 }
 
 static void
-render_cur_frame (HTMLImage *image, gint nx, gint ny, const GdkColor *highlight_color)
+html_image_pointer_queue_animation (HTMLImagePointer *ip)
 {
-	GdkPixbufFrame    *frame;
-	HTMLPainter       *painter;
-	HTMLImageAnimation *anim = image->animation;
-	GdkPixbufAnimation *ganim = image->image_ptr->animation;
-	GList *cur = gdk_pixbuf_animation_get_frames (ganim);
-	gint w, h;
-	gboolean saved_alpha = TRUE;
+	gint delay = gdk_pixbuf_animation_iter_get_delay_time (ip->iter);
 
-	painter = image->image_ptr->factory->engine->painter;
-
-	frame = (GdkPixbufFrame *) anim->cur_frame->data;
-	/* printf ("w: %d h: %d action: %d\n", w, h, frame->action); */
-
-	/* FIXME this is hack to turn off alpha blending while rending 
-	 * animations.  This breaks nothing since gdk-pixbuf doesn't support animations
-	 * for anything but gif and it makes a huge difference if there are lots of
-	 * frames.  We really should add a parameter to the html_painter_draw_pixmap
-	 * call and track wether each image has full alpha or bilevel alpha, but that
-	 * will wait for a bit.
-	 */  
-	if (HTML_IS_GDK_PAINTER (painter)) {
-		saved_alpha = HTML_GDK_PAINTER (painter)->alpha;
-		HTML_GDK_PAINTER (painter)->alpha = FALSE;
-	}
-
-	do {
-		frame = (GdkPixbufFrame *) cur->data;
-		w = gdk_pixbuf_get_width (gdk_pixbuf_frame_get_pixbuf (frame));
-		h = gdk_pixbuf_get_height (gdk_pixbuf_frame_get_pixbuf (frame));
-		if (anim->cur_frame == cur) {
-			html_painter_draw_pixmap (painter, gdk_pixbuf_frame_get_pixbuf (frame),
-						  nx + gdk_pixbuf_frame_get_x_offset (frame),
-						  ny + gdk_pixbuf_frame_get_y_offset (frame),
-						  w, h,
-						  highlight_color);
-			break;
-		} else if (gdk_pixbuf_frame_get_action (frame) == GDK_PIXBUF_FRAME_RETAIN) {
-			html_painter_draw_pixmap (painter, gdk_pixbuf_frame_get_pixbuf (frame),
-						  nx + gdk_pixbuf_frame_get_x_offset (frame),
-						  ny + gdk_pixbuf_frame_get_y_offset (frame),
-						  w, h,
-						  highlight_color);
-		} 
-		cur = cur->next;
-	} while (1);
-
-	if (HTML_IS_GDK_PAINTER (painter)) {
-		HTML_GDK_PAINTER (painter)->alpha = saved_alpha;
-	}
+	if (delay >= 0)
+		ip->animation_timeout = g_timeout_add (delay, 
+						       (GtkFunction) html_image_pointer_run_animation, 
+						       (gpointer) ip);
 
 }
 
 static gint
-html_image_animation_timeout (HTMLImage *image)
+html_image_pointer_run_animation (HTMLImagePointer *ip)
 {
-	HTMLImageAnimation *anim = image->animation;
-	GdkPixbufAnimation *ganim = image->image_ptr->animation;
-	GdkPixbufFrame    *frame;
-	HTMLEngine        *engine;
-	gint nx, ny, nex, ney;
-
+	GdkPixbufAnimation      *ganim = ip->animation;
+	GdkPixbufAnimationIter  *iter = ip->iter;
+	HTMLEngine              *engine = ip->factory->engine;
+	gint nx, ny;
+	
 	/* printf ("animation_timeout\n"); */
-	anim->cur_frame = anim->cur_frame->next;
-	if (!anim->cur_frame)
-		anim->cur_frame = gdk_pixbuf_animation_get_frames (image->image_ptr->animation);
+	if (gdk_pixbuf_animation_iter_advance (iter, NULL)) {
+		GSList *cur;
 
-	frame = (GdkPixbufFrame *) anim->cur_frame->data;
-	
-	/* draw only if animation is active - onscreen */
+		for (cur = ip->interests; cur; cur = cur->next) {
+			HTMLImage           *image = cur->data;
 
-	engine = image->image_ptr->factory->engine;
+			if (image && image->animation_active)
+				html_engine_queue_draw (engine, HTML_OBJECT (image));
 
-	nex = engine->x_offset;
-	ney = engine->y_offset;
-
-	nx = anim->x - (nex - anim->ex);
-	ny = anim->y - (ney - anim->ey);
-	
-	if (anim->active) {
-		gint aw, ah;
-
-		aw = gdk_pixbuf_animation_get_width (ganim);
-		ah = gdk_pixbuf_animation_get_height (ganim);
-
-		if (MAX(0, nx) < MIN(engine->width, nx+aw)
-		    && MAX(0, ny) < MIN(engine->height, ny+ah)) {
-			html_engine_draw (engine,
-					  nx, ny,
-					  aw, ah);
-			/* html_engine_queue_draw (engine, HTML_OBJECT (image)); */
-			
 		}
-		
 	}
-	/* printf ("timeout: %d\n", gdk_pixbuf_frame_get_delay_time (frame)); */
-	anim->timeout = g_timeout_add (10 * (gdk_pixbuf_frame_get_delay_time (frame) > 0
-					     ? gdk_pixbuf_frame_get_delay_time (frame) : 1),
-				       (GtkFunction) html_image_animation_timeout, (gpointer) image);
-
+		
+	html_image_pointer_queue_animation (ip);
 	return FALSE;
 }
 
 static void
-html_image_animation_start (HTMLImage *image)
+html_image_pointer_start_animation (HTMLImagePointer *ip)
 {
-	HTMLImageAnimation *anim = image->animation;
+	if (ip->animation && !gdk_pixbuf_animation_is_static_image (ip->animation)) {
+		if (!ip->iter)
+			ip->iter = gdk_pixbuf_animation_get_iter (ip->animation, NULL);
 
-	if (anim && gdk_pixbuf_animation_get_num_frames (image->image_ptr->animation) > 1) {
-		if (anim->timeout == 0) {
-			GList *frames = gdk_pixbuf_animation_get_frames (image->image_ptr->animation);
-
-			anim->cur_frame = frames->next;
-			anim->cur_n = 1;
-			anim->timeout = g_timeout_add (10 * gdk_pixbuf_frame_get_delay_time
-						       ((GdkPixbufFrame *) frames->data),
-						       (GtkFunction) html_image_animation_timeout, (gpointer) image);
-		}
+		html_image_pointer_queue_animation (ip);
 	}
 }
 
 static void
-html_image_animation_stop (HTMLImageAnimation *anim)
+html_image_pointer_stop_animation (HTMLImagePointer *ip)
 {
-	if (anim->timeout) {
-		g_source_remove (anim->timeout);
-		anim->timeout = 0;
+	if (ip->animation_timeout) {
+		g_source_remove (ip->animation_timeout);
+		ip->animation_timeout = 0;
 	}
-	anim->active  = 0;
-}
-
-static void
-html_image_factory_frame_done (GdkPixbufLoader *loader, GdkPixbufFrame *frame, HTMLImagePointer *ip)
-{
-	/* printf ("frame done\n"); */
-
-	if (!ip->animation) {
-		ip->animation = gdk_pixbuf_loader_get_animation (loader);
-		gdk_pixbuf_animation_ref (ip->animation);
-	}
-	g_assert (ip->animation);
-
-	if (gdk_pixbuf_animation_get_num_frames (ip->animation) > 1) {
-		GSList *cur = ip->interests;
-		HTMLImage *image;
-
-		while (cur) {
-			if (cur->data) {
-				image = HTML_IMAGE (cur->data);
-				if (!image->animation) {
-					image->animation = html_image_animation_new (image);
-				}
-				html_image_animation_start (image);
-			}
-			cur = cur->next;
-		}
-	}
-}
-
-static void
-html_image_factory_animation_done (GdkPixbufLoader *loader, HTMLImagePointer *ip)
-{
-	g_print ("animation done\n");
 }
 
 HTMLImageFactory *
@@ -1159,7 +1131,6 @@ cleanup_images (gpointer key, gpointer value, gpointer user_data)
 	gboolean retval = FALSE;
 
 	ptr = value;
-
 	/* user data means: NULL only clean, non-NULL free */
 	if (user_data){
 		if (ptr->interests != NULL) {
@@ -1194,35 +1165,6 @@ html_image_factory_free (HTMLImageFactory *factory)
 	g_free (factory);
 }
 
-static HTMLImageAnimation *
-html_image_animation_new (HTMLImage *image)
-{
-	HTMLImageAnimation *animation;
-
-	animation = g_new (HTMLImageAnimation, 1);
-	animation->cur_frame = gdk_pixbuf_animation_get_frames (image->image_ptr->animation);
-	animation->cur_n = 0;
-	animation->x = 0;
-	animation->y = 0;
-	animation->ex = 0;
-	animation->ey = 0;
-	animation->timeout = 0;
-	animation->pixbuf = gdk_pixbuf_new (GDK_COLORSPACE_RGB, TRUE, 8,
-					    gdk_pixbuf_animation_get_width (image->image_ptr->animation),
-					    gdk_pixbuf_animation_get_height (image->image_ptr->animation));
-	animation->active = FALSE;
-
-	return animation;
-}
-
-static void
-html_image_animation_destroy (HTMLImageAnimation *anim)
-{
-	html_image_animation_stop (anim);
-	gdk_pixbuf_unref (anim->pixbuf);
-	g_free (anim);
-}
-
 #define STALL_INTERVAL 1000
 
 static HTMLImagePointer *
@@ -1234,7 +1176,7 @@ html_image_pointer_new (const char *filename, HTMLImageFactory *factory)
 	retval->refcount = 1;
 	retval->url = g_strdup (filename);
 	retval->loader = gdk_pixbuf_loader_new ();
-	retval->pixbuf = NULL;
+	retval->iter = NULL;
 	retval->animation = NULL;
 	retval->interests = NULL;
 	retval->factory = factory;
@@ -1242,6 +1184,7 @@ html_image_pointer_new (const char *filename, HTMLImageFactory *factory)
 	retval->stall_timeout = gtk_timeout_add (STALL_INTERVAL, 
 						 (GtkFunction)html_image_pointer_timeout,
 						 retval);
+	retval->animation_timeout = 0;
 	return retval;
 }
 
@@ -1258,7 +1201,7 @@ html_image_pointer_timeout (HTMLImagePointer *ip)
 	 * draw the frame now that we decided they've had enough time to
 	 * load the image
 	 */
-	if (ip->pixbuf == NULL) {
+	if (ip->animation == NULL) {
 		while (list) {
 			image = (HTMLImage *)list->data;
 
@@ -1269,7 +1212,7 @@ html_image_pointer_timeout (HTMLImagePointer *ip)
 			list = list->next;
 		}
 	}
-	ip->stall_timeout = 0;
+	//ip->stall_timeout = 0;
 	return FALSE;
 }
 
@@ -1283,18 +1226,14 @@ static void
 free_image_ptr_data (HTMLImagePointer *ip)
 {
 	if (ip->loader) {
-		gtk_object_unref (GTK_OBJECT (ip->loader));
+		gdk_pixbuf_loader_close (ip->loader, NULL);
+		g_object_unref (ip->loader);
 		ip->loader = NULL;
 	}
 	if (ip->animation) {
-		gdk_pixbuf_animation_unref (ip->animation);
+		g_object_unref (ip->animation);
 		ip->animation = NULL;
 	}
-	if (ip->pixbuf) {
-		gdk_pixbuf_unref (ip->pixbuf);
-		ip->pixbuf = NULL;
-	}
-
 }
 
 static void
@@ -1312,85 +1251,76 @@ html_image_pointer_unref (HTMLImagePointer *ip)
 	g_return_if_fail (ip != NULL);
 
 	ip->refcount--;
+	/* printf ("unref(%p) %s --> %d\n", ip, ip->url, ip->refcount); */
 	if (ip->refcount <= 0) {
+		/* printf ("freeing %s\n", ip->url); */
 		html_image_pointer_remove_stall (ip);
+		html_image_pointer_stop_animation (ip);
 		g_free (ip->url);
 		free_image_ptr_data (ip);
 		g_free (ip);
 	}
 }
 
-static void
+GtkHTMLStream *
 html_image_pointer_load (HTMLImagePointer *ip)
 {
-	GtkHTMLStream *handle;
-
 	html_image_pointer_ref (ip);
 
-	handle = gtk_html_stream_new (GTK_HTML (ip->factory->engine->widget),
-				      html_image_factory_types,
-				      html_image_factory_write_pixbuf,
-				      html_image_factory_end_pixbuf,
-				      ip);
-
-	/* This is a bit evil, I think.  But it's a lot better here
-	   than in the HTMLImage object.  FIXME anyway -- ettore  */
-	gtk_signal_emit_by_name (GTK_OBJECT (ip->factory->engine), "url_requested", ip->url, handle);
+	ip->factory->engine->opened_streams ++;
+	return gtk_html_stream_new (GTK_HTML (ip->factory->engine->widget),
+				    html_image_factory_types,
+				    html_image_factory_write_pixbuf,
+				    html_image_factory_end_pixbuf,
+				    ip);
 }
 
 HTMLImagePointer *
-html_image_factory_register (HTMLImageFactory *factory, HTMLImage *i, const char *filename, gboolean reload)
+html_image_factory_register (HTMLImageFactory *factory, HTMLImage *i, const char *url, gboolean reload)
 {
-	HTMLImagePointer *retval;
+	HTMLImagePointer *ip;
+	GtkHTMLStream *stream = NULL;
 
 	g_return_val_if_fail (factory, NULL);
-	g_return_val_if_fail (filename, NULL);
+	g_return_val_if_fail (url, NULL);
 
-	retval = g_hash_table_lookup (factory->loaded_images, filename);
+	ip = g_hash_table_lookup (factory->loaded_images, url);
 
-	if (!retval){
-		retval = html_image_pointer_new (filename, factory);
-		if (*filename) {
-			gtk_signal_connect (GTK_OBJECT (retval->loader), "area_prepared",
-					    GTK_SIGNAL_FUNC (html_image_factory_area_prepared),
-					    retval);
+	if (!ip) {
+		ip = html_image_pointer_new (url, factory);
+		if (*url) {
+			g_signal_connect (G_OBJECT (ip->loader), "area_prepared",
+					  G_CALLBACK (html_image_factory_area_prepared),
+					  ip);
 
-			gtk_signal_connect (GTK_OBJECT (retval->loader), "area_updated",
-					    GTK_SIGNAL_FUNC (html_image_factory_area_updated),
-					    retval);
+			g_signal_connect (G_OBJECT (ip->loader), "area_updated",
+					  G_CALLBACK (html_image_factory_area_updated),
+					  ip);
 
-			gtk_signal_connect (GTK_OBJECT (retval->loader), "frame_done",
-					    GTK_SIGNAL_FUNC (html_image_factory_frame_done),
-					    retval);
-
-			gtk_signal_connect (GTK_OBJECT (retval->loader), "animation_done",
-					    GTK_SIGNAL_FUNC (html_image_factory_animation_done),
-					    retval);
-
-			g_hash_table_insert (factory->loaded_images, retval->url, retval);
-			html_image_pointer_load (retval);
+			g_hash_table_insert (factory->loaded_images, ip->url, ip);
+			stream = html_image_pointer_load (ip);
+		}
+	} else {
+		if (reload) {
+			free_image_ptr_data (ip);
+			ip->loader = gdk_pixbuf_loader_new ();
+			stream = html_image_pointer_load (ip);
 		}
-	} else if (reload) {
-		free_image_ptr_data (retval);
-		retval->loader = gdk_pixbuf_loader_new ();
-		html_image_pointer_load (retval);
 	}
 
-	html_image_pointer_ref (retval);
+	if (stream)
+		g_signal_emit_by_name (factory->engine, "url_requested", ip->url, stream);
+
+	html_image_pointer_ref (ip);
 
 	/* we add also NULL ptrs, as we dont want these to be cleaned out */
-	retval->interests = g_slist_prepend (retval->interests, i);
+	ip->interests = g_slist_prepend (ip->interests, i);
 
 	if (i) {
-		i->image_ptr      = retval;
-
-		if (retval->animation && gdk_pixbuf_animation_get_num_frames (retval->animation) > 1) {
-			i->animation = html_image_animation_new (i);
-			html_image_animation_start (i);
-		}
+		i->image_ptr = ip;
 	}
 
-	return retval;
+	return ip;
 }
 
 #if 0
@@ -1409,6 +1339,7 @@ html_image_factory_unregister (HTMLImageFactory *factory, HTMLImagePointer *poin
 	html_image_pointer_unref (pointer);
 	if (pointer->refcount == 1) {
 		g_assert (pointer->interests == NULL);
+		/* printf ("remove %s\n", pointer->url); */
 		g_hash_table_remove (factory->loaded_images, pointer->url);
 		html_image_pointer_unref (pointer);
 	}
@@ -1418,20 +1349,8 @@ static void
 stop_anim (gpointer key, gpointer value, gpointer user_data)
 {
 	HTMLImagePointer *ip = value;
-	GSList *cur = ip->interests;
-	HTMLImage *image;
-
 	html_image_pointer_remove_stall (ip);
-
-	while (cur) {
-		if (cur->data) {
-			image = (HTMLImage *) cur->data;
-			if (image->animation) {
-				html_image_animation_stop (image->animation);
-			}
-		}
-		cur = cur->next;
-	}
+	html_image_pointer_stop_animation (ip);
 }
 
 void
@@ -1448,7 +1367,7 @@ move_image_pointers (gpointer key, gpointer value, gpointer data)
 
 	ip->factory = dst;
 	g_hash_table_insert (dst->loaded_images, ip->url, ip);
-	html_image_pointer_load (ip);
+	g_signal_emit_by_name (ip->factory->engine, "url_requested", ip->url, html_image_pointer_load (ip));
 
 	return TRUE;
 }
@@ -1465,13 +1384,12 @@ deactivate_anim (gpointer key, gpointer value, gpointer user_data)
 	HTMLImagePointer *ip = value;
 	GSList *cur = ip->interests;
 	HTMLImage *image;
+	
 
 	while (cur) {
 		if (cur->data) {
 			image = (HTMLImage *) cur->data;
-			if (image->animation) {
-				image->animation->active = 0;
-			}
+			image->animation_active = FALSE;
 		}
 		cur = cur->next;
 	}
@@ -1482,3 +1400,60 @@ html_image_factory_deactivate_animations (HTMLImageFactory *factory)
 {
 	g_hash_table_foreach (factory->loaded_images, deactivate_anim, NULL);
 }
+
+static void
+ref_image_ptr (gpointer key, gpointer val, gpointer data)
+{
+	html_image_pointer_ref (HTML_IMAGE_POINTER (val));
+	/* printf ("ref(%p) %s --> %d\n", val, HTML_IMAGE_POINTER (val)->url, HTML_IMAGE_POINTER (val)->refcount); */
+}
+
+static void
+unref_image_ptr (gpointer key, gpointer val, gpointer data)
+{
+	html_image_pointer_unref (HTML_IMAGE_POINTER (val));
+}
+
+void
+html_image_factory_ref_all_images (HTMLImageFactory *factory)
+{
+	if (!factory->loaded_images)
+		return;
+
+	g_hash_table_foreach (factory->loaded_images, ref_image_ptr, NULL);
+}
+
+void
+html_image_factory_unref_all_images (HTMLImageFactory *factory)
+{
+	if (!factory->loaded_images)
+		return;
+
+	g_hash_table_foreach (factory->loaded_images, unref_image_ptr, NULL);
+}
+
+void
+html_image_factory_ref_image_ptr (HTMLImageFactory *factory, const gchar *url)
+{
+	HTMLImagePointer *ptr;
+
+	if (!factory->loaded_images)
+		return;
+
+	ptr = HTML_IMAGE_POINTER (g_hash_table_lookup (factory->loaded_images, url));
+	if (ptr)
+		html_image_pointer_ref (ptr);
+}
+
+void
+html_image_factory_unref_image_ptr (HTMLImageFactory *factory, const gchar *url)
+{
+	HTMLImagePointer *ptr;
+
+	if (!factory->loaded_images)
+		return;
+
+	ptr = HTML_IMAGE_POINTER (g_hash_table_lookup (factory->loaded_images, url));
+	if (ptr)
+		html_image_pointer_unref (ptr);
+}
