@@ -23,6 +23,8 @@
 
 #include <config.h>
 #include <string.h>
+#include <glib/gdataset.h>
+#include <glib/gquark.h>
 
 #include "htmlclue.h"
 #include "htmlclueflow.h"
@@ -50,9 +52,6 @@ HTMLObjectClass html_object_class;
 
 #define HO_CLASS(x) HTML_OBJECT_CLASS (HTML_OBJECT (x)->klass)
 
-
-/* HTMLObject virtual methods.  */
-
 static void
 destroy (HTMLObject *self)
 {
@@ -795,29 +794,35 @@ html_object_calc_abs_position (HTMLObject *o,
 	for (p = o->parent; p != NULL; p = p->parent) {
 		*x_return += p->x;
 		*y_return += p->y - p->ascent;
-		if (html_object_is_frame (p)) {
-			*x_return += GTK_HTML (HTML_FRAME (p)->html)->engine->leftBorder;
-			*y_return += GTK_HTML (HTML_FRAME (p)->html)->engine->topBorder;
-		}
 	}
 }
 
-void
-html_object_calc_intersection (HTMLObject *o, ArtIRect *intersection, gint x, gint y, gint width, gint height)
+GdkRectangle *
+html_object_get_bounds (HTMLObject *o, GdkRectangle *bounds)
 {
-	ArtIRect clip, area;
-	
-	area.x0 = o->x;
-	area.x1 = o->x + o->width;
-	area.y0 = o->y - o->ascent;
-	area.y1 = o->y + o->descent;
+	if (!bounds)
+		bounds = g_new (GdkRectangle, 1);
 
-	clip.x0 = x;
-	clip.x1 = x + width;
-	clip.y0 = y;
-	clip.y1 = y + height;
+	bounds->x = o->x;
+	bounds->y = o->y - o->ascent;
+	bounds->width = o->width;
+	bounds->height = o->ascent + o->descent;
 
-	art_irect_intersect (intersection, &clip, &area);
+	return bounds;
+}
+
+gboolean
+html_object_intersect (HTMLObject *o, GdkRectangle *result, gint x, gint y, gint width, gint height)
+{
+	GdkRectangle b;
+	GdkRectangle a;
+
+	a.x = x;
+	a.y = y;
+	a.width = width;
+	a.height = height;
+
+	return gdk_rectangle_intersect (html_object_get_bounds (o, &b), &a, result);
 }
 
 
@@ -1493,10 +1498,7 @@ next_object_uptree_cursor (HTMLObject *obj, HTMLObject * (*next_fn ) (HTMLObject
 static HTMLObject *
 move_object_downtree_cursor (HTMLObject *obj, HTMLObject * (*down_fn ) (HTMLObject *))
 {
-	HTMLObject *down;
-
-	while ((down = (*down_fn) (obj))) {
-		obj = down;
+	while ((obj = (*down_fn) (obj))) {
 		if (html_object_accepts_cursor (obj))
 			break;
 	}
@@ -1511,6 +1513,7 @@ move_object_cursor (HTMLObject *obj, gint *offset, gboolean forward,
 	HTMLObject *down, *before;
 
 	do {
+		gboolean found = FALSE;
 		if (((*offset == 0 && forward) || (*offset && !forward)) && html_object_is_container (obj))
 			if ((down = (*down_fn) (obj))) {
 				down = move_object_downtree_cursor (down, down_fn);
@@ -1522,19 +1525,27 @@ move_object_cursor (HTMLObject *obj, gint *offset, gboolean forward,
 			}
 
 		before = obj;
-		obj = next_object_uptree_cursor (obj, next_fn);
-		if (obj) {
-			if (html_object_accepts_cursor (obj)) {
-				if (html_object_is_container (obj))
-					*offset = before->parent == obj->parent
-						? forward ? 0 : 1
-						: forward ? 1 : 0;
-			} else {
-				obj = move_object_downtree_cursor (obj, down_fn);
-				if (html_object_is_container (obj))
-					*offset = forward ? 0 : 1;
+		do {
+			obj = next_object_uptree_cursor (obj, next_fn);
+			if (obj) {
+				if (html_object_accepts_cursor (obj)) {
+					if (html_object_is_container (obj))
+						*offset = before->parent == obj->parent
+							? forward ? 0 : 1
+							: forward ? 1 : 0;
+					found = TRUE;
+				} else {
+					HTMLObject *down;
+					down = move_object_downtree_cursor (obj, down_fn);
+					if (down) {
+						if (html_object_is_container (down))
+							*offset = forward ? 0 : 1;
+						obj = down;
+						found = TRUE;
+					}
+				}
 			}
-		}
+		} while (obj && !found);
 	} while (obj && !html_object_accepts_cursor (obj));
 
 	return obj;
@@ -1572,6 +1583,12 @@ html_object_set_data (HTMLObject *object, const gchar *key, const gchar *value)
 	g_datalist_set_data_full (&object->object_data, key, g_strdup (value), g_free);
 }
 
+void
+html_object_set_data_full (HTMLObject *object, const gchar *key, const gpointer value, GDestroyNotify func)
+{
+	g_datalist_set_data_full (&object->object_data, key, value, func);
+}
+
 gpointer
 html_object_get_data (HTMLObject *object, const gchar *key)
 {
@@ -1583,10 +1600,9 @@ copy_data (GQuark key_id, gpointer data, gpointer user_data)
 {
 	HTMLObject *o = HTML_OBJECT (user_data);
 
-	/* printf ("copy key: '%s'\n", g_quark_to_string (key_id)); */
-	g_datalist_set_data_full (&o->object_data,
-				  g_strdup (g_quark_to_string (key_id)),
-				  g_strdup ((gchar *) data), g_free);
+	g_datalist_id_set_data_full (&o->object_data,
+				     key_id,
+				     g_strdup ((gchar *) data), g_free);
 }
 
 void
@@ -1623,7 +1639,6 @@ handle_object_data (gpointer key, gpointer value, gpointer data)
 		/* printf ("clear\n"); */
 		html_engine_save_output_string (state, "<!--+GtkHTML:<DATA class=\"%s\" clear=\"%s\">",
 						state->save_data_class_name, key);
-		
 		state->data_to_remove = g_slist_prepend (state->data_to_remove, key);
 	} else if (strcmp (value, str)) {
 		/* printf ("change\n"); */
@@ -1910,9 +1925,8 @@ html_object_engine_translation (HTMLObject *o, HTMLEngine *e, gint *tx, gint *ty
 		*ty += p->y - p->ascent;
 	}
 
-	*tx = *tx + e->leftBorder - e->x_offset;
-	*ty = *ty + e->topBorder - e->y_offset;
-
+	//*tx += e->leftBorder;
+	//*ty += e->topBorder;
 }
 
 gboolean
@@ -1951,6 +1965,25 @@ html_object_add_to_changed (GList **changed_objs, HTMLObject *o)
 	*changed_objs = g_list_prepend (*changed_objs, o);
 }
 
+gint
+html_object_get_n_children (HTMLObject *self)
+{
+	return HO_CLASS (self)->get_n_children ? (* HO_CLASS (self)->get_n_children) (self) : 0;
+}
+
+HTMLObject *
+html_object_get_child (HTMLObject *self, gint index)
+{
+	return HO_CLASS (self)->get_child ? (* HO_CLASS (self)->get_child) (self, index) : NULL;
+}
+
+
+gint
+html_object_get_child_index (HTMLObject *self, HTMLObject *child)
+{
+	return HO_CLASS (self)->get_child_index ? (* HO_CLASS (self)->get_child_index) (self, child) : -1;
+}
+
 HTMLClearType
 html_object_get_clear (HTMLObject *self)
 {
