@@ -25,7 +25,8 @@
 #define _HTMLPAINTER_H_
 
 #include <gdk-pixbuf/gdk-pixbuf.h>
-#include <gtk/gtkobject.h>
+#include <gdk/gdktypes.h>
+#include <glib-object.h>
 
 #include "gtkhtml-enums.h"
 #include "htmltypes.h"
@@ -33,14 +34,14 @@
 
 
 #define HTML_TYPE_PAINTER                 (html_painter_get_type ())
-#define HTML_PAINTER(obj)                 (GTK_CHECK_CAST ((obj), HTML_TYPE_PAINTER, HTMLPainter))
-#define HTML_PAINTER_CLASS(klass)         (GTK_CHECK_CLASS_CAST ((klass), HTML_TYPE_PAINTER, HTMLPainterClass))
-#define HTML_IS_PAINTER(obj)              (GTK_CHECK_TYPE ((obj), HTML_TYPE_PAINTER))
-#define HTML_IS_PAINTER_CLASS(klass)      (GTK_CHECK_CLASS_TYPE ((klass), HTML_TYPE_PAINTER))
+#define HTML_PAINTER(obj)                 (G_TYPE_CHECK_INSTANCE_CAST ((obj), HTML_TYPE_PAINTER, HTMLPainter))
+#define HTML_PAINTER_CLASS(klass)         (G_TYPE_CHECK_CLASS_CAST ((klass), HTML_TYPE_PAINTER, HTMLPainterClass))
+#define HTML_IS_PAINTER(obj)              (G_TYPE_CHECK_INSTANCE_TYPE ((obj), HTML_TYPE_PAINTER))
+#define HTML_IS_PAINTER_CLASS(klass)      (G_TYPE_CHECK_CLASS_TYPE ((klass), HTML_TYPE_PAINTER))
 
 
 struct _HTMLPainter {
-	GtkObject base;
+	GObject base;
 
 	HTMLFontManager     font_manager;
 	HTMLColorSet       *color_set;
@@ -51,7 +52,7 @@ struct _HTMLPainter {
 };
 
 struct _HTMLPainterClass {
-	GtkObjectClass   base;
+	GObjectClass   base;
 
 	void (* begin)            (HTMLPainter *painter, int x1, int y1, int x2, int y2);
 	void (* end)              (HTMLPainter *painter);
@@ -63,17 +64,17 @@ struct _HTMLPainterClass {
 	void (* alloc_color)      (HTMLPainter *painter, GdkColor *color);
 	void (* free_color)       (HTMLPainter *painter, GdkColor *color);
 
-	guint (* calc_ascent)     (HTMLPainter *p, GtkHTMLFontStyle f, HTMLFontFace *face);
-	guint (* calc_descent)    (HTMLPainter *p, GtkHTMLFontStyle f, HTMLFontFace *face);
-	guint (* calc_text_width) (HTMLPainter *p, const gchar *text, guint len,
-				   GtkHTMLFontStyle font_style, HTMLFontFace *face);
+	void (* calc_text_size) (HTMLPainter *p, const gchar *text, guint len, GList *items, GList *glyphs,
+				 GtkHTMLFontStyle font_style, HTMLFontFace *face, gint *width, gint *asc, gint *dsc);
+	void (* calc_text_size_bytes) (HTMLPainter *p, const gchar *text, guint len, GList *items, GList *glyphs,
+				       HTMLFont *font, GtkHTMLFontStyle font_style, gint *width, gint *asc, gint *dsc);
 
 	void (* set_pen)          (HTMLPainter *painter, const GdkColor *color);
 	const GdkColor * (* get_black) (const HTMLPainter *painter);
 	void (* draw_line)        (HTMLPainter *painter, gint x1, gint y1, gint x2, gint y2);
 	void (* draw_rect)        (HTMLPainter *painter, gint x, gint y, gint width, gint height);
-	void (* draw_text)        (HTMLPainter *painter, gint x, gint y, const gchar *text, gint len);
-	gint (* draw_spell_error) (HTMLPainter *painter, gint x, gint y, const gchar *text, gint len);
+	void (* draw_text)        (HTMLPainter *painter, gint x, gint y, const gchar *text, gint len, GList *items, GList *glyphs);
+	gint (* draw_spell_error) (HTMLPainter *painter, gint x, gint y, const gchar *text, gint len, GList *items, GList *glyphs);
 	void (* fill_rect)        (HTMLPainter *painter, gint x, gint y, gint width, gint height);
 	void (* draw_pixmap)      (HTMLPainter *painter, GdkPixbuf *pixbuf, 
 				   gint x, gint y,
@@ -100,7 +101,7 @@ struct _HTMLPainterClass {
 
 
 /* Creation.  */
-GtkType           html_painter_get_type                                (void);
+GType             html_painter_get_type                                (void);
 HTMLPainter      *html_painter_new                                     (void);
 
 /* Functions to drive the painting process.  */
@@ -138,18 +139,28 @@ void              html_painter_set_font_face                           (HTMLPain
 gpointer          html_painter_get_font                                (HTMLPainter       *painter,
 									HTMLFontFace      *face,
 									GtkHTMLFontStyle   style);
-guint             html_painter_calc_ascent                             (HTMLPainter       *p,
-									GtkHTMLFontStyle   f,
-									HTMLFontFace      *face);
-guint             html_painter_calc_descent                            (HTMLPainter       *p,
-									GtkHTMLFontStyle   f,
-									HTMLFontFace      *face);
-guint             html_painter_calc_text_width                         (HTMLPainter       *p,
+void              html_painter_calc_text_size                          (HTMLPainter       *p,
 									const gchar       *text,
 									guint              len,
+									GList             *items,
+									GList             *glyphs,
 									gint              *line_offset,
 									GtkHTMLFontStyle   font_style,
-									HTMLFontFace      *face);
+									HTMLFontFace      *face,
+									gint              *width,
+									gint              *asc,
+									gint              *dsc);
+void              html_painter_calc_text_size_bytes                    (HTMLPainter       *p,
+									const gchar       *text,
+									guint              len,
+									GList             *items,
+									GList             *glyphs,
+									gint              *line_offset,
+									HTMLFont          *font,
+									GtkHTMLFontStyle   font_style,
+									gint              *width,
+									gint              *asc,
+									gint              *dsc);
 
 /* The actual paint operations.  */
 void              html_painter_set_pen                                 (HTMLPainter       *painter,
@@ -169,6 +180,8 @@ gint              html_painter_draw_text                               (HTMLPain
 									gint               y,
 									const gchar       *text,
 									gint               len,
+									GList             *items,
+									GList             *glyphs,
 									gint               line_offset);
 void              html_painter_fill_rect                               (HTMLPainter       *painter,
 									gint               x,
@@ -225,7 +238,9 @@ gint              html_painter_draw_spell_error                        (HTMLPain
 									gint               x,
 									gint               y,
 									const gchar       *text,
-									gint               len);
+									gint               len,
+									GList             *items,
+									GList             *glyphs);
 HTMLFont         *html_painter_alloc_font                              (HTMLPainter       *painter,
 									gchar             *face_name,
 									gdouble            size,
@@ -238,6 +253,12 @@ void              html_painter_unref_font                              (HTMLPain
 guint             html_painter_get_space_width                         (HTMLPainter       *painter,
 									GtkHTMLFontStyle   font_style,
 									HTMLFontFace      *face);
+guint             html_painter_get_block_indent_width                  (HTMLPainter       *painter,
+									GtkHTMLFontStyle   font_style,
+									HTMLFontFace      *face);
+guint             html_painter_get_block_cite_width                    (HTMLPainter       *painter,
+									GtkHTMLFontStyle   font_style,
+									HTMLFontFace      *face);
 void              html_painter_draw_embedded                           (HTMLPainter       *painter,
 									HTMLEmbedded      *element,
 									gint               x,
@@ -248,4 +269,12 @@ guint             html_painter_get_page_height                         (HTMLPain
 								        HTMLEngine        *e);
 void              html_painter_set_focus                               (HTMLPainter       *painter,
 									gboolean           focus);
+void              html_replace_tabs                                    (const gchar       *text,
+									gchar             *translated,
+									guint              bytes);
+
+#define HTML_BLOCK_INDENT   "        "
+#define HTML_BLOCK_CITE     "> "
+#define HTML_ALLOCA_MAX     2048
+
 #endif /* _HTMLPAINTER_H_ */
