@@ -41,91 +41,10 @@
 HTMLTextSlaveClass html_text_slave_class;
 static HTMLObjectClass *parent_class = NULL;
 
-#ifndef GAL_UTF8_NOT_SLOW
-/*
- * NOTE:
- * currently the gal utf8 functions check the entire string you pass to them for
- * invalid utf8, the expense of those calls is massive inside gtkhtml when a paragraph
- * is long.  Until gal turns off the checks we will use versions without checks.
- * And yes, it really is worth it.
- *   
- * --Larry
- */
-gint h_utf8_pointer_to_offset (const gchar *str, const gchar *pos);
-gint h_utf8_strlen (const gchar *p, gint max);
-gchar *h_utf8_offset_to_pointer  (const gchar *str, gint         offset);
+static GList * get_items (HTMLTextSlave *slave, HTMLPainter *painter);
+static GList * get_glyphs (HTMLTextSlave *slave, HTMLPainter *painter, gint line_offset);
+static GList * get_glyphs_part (HTMLTextSlave *slave, HTMLPainter *painter, guint offset, guint len, gint line_offset, GList **items);
 
-gint
-h_utf8_pointer_to_offset (const gchar *str,
-			  const gchar *pos)
-{
-  const gchar *s = str;
-  gint offset = 0;
-
-  while (s < pos)
-    {
-      s = g_utf8_next_char (s);
-      offset++;
-    }
-
-  return offset;
-}
-
-gint
-h_utf8_strlen (const gchar *p, gint max)
-{
-  int len = 0;
-  const gchar *start = p;
-
-  if (max < 0)
-    {
-      while (*p)
-        {
-          p = g_utf8_next_char (p);
-          ++len;
-        }
-    }
-  else
-    {
-      if (max == 0 || !*p)
-        return 0;
-      
-      p = g_utf8_next_char (p);          
-
-      while (p - start < max && *p)
-        {
-          ++len;
-          p = g_utf8_next_char (p);          
-        }
-
-      /* only do the last len increment if we got a complete
-       * char (don't count partial chars)
-       */
-      if (p - start == max)
-        ++len;
-    }
-
-  return len;
-}
-
-gchar *
-h_utf8_offset_to_pointer  (const gchar *str,
-			   gint         offset)
-{
-  const gchar *s = str;
-
-  while (offset--)
-    s = g_utf8_next_char (s);
-  
-  return (gchar *)s;
-}
-
-#define g_utf8_strlen h_utf8_strlen
-#define g_utf8_offset_to_pointer h_utf8_offset_to_pointer
-#define g_utf8_pointer_to_offset h_utf8_pointer_to_offset
-#endif
-
-
 char *
 html_text_slave_get_text (HTMLTextSlave *slave)
 {
@@ -176,20 +95,26 @@ copy (HTMLObject *self,
 static guint
 get_words_width (HTMLText *text, HTMLPainter *p, guint start_word, guint words)
 {
-	return text->word_width [start_word + words - 1]
+	return words > 0
+		? text->word_width [start_word + words - 1]
 		- (start_word ? text->word_width [start_word - 1]
-		   + html_painter_get_space_width (p, html_text_get_font_style (text), text->face) : 0);
+		   + html_painter_get_space_width (p, html_text_get_font_style (text), text->face) : 0)
+		: 0;
 }
 
 static guint
-calc_width (HTMLTextSlave *slave, HTMLPainter *painter)
+calc_width (HTMLTextSlave *slave, HTMLPainter *painter, gint *asc, gint *dsc)
 {
 	HTMLText *text = slave->owner;
 	HTMLObject *next, *prev;
 
 	html_text_request_word_width (text, painter);
-	if (slave->posStart == 0 && slave->posLen == text->text_len)
+	if (slave->posStart == 0 && slave->posLen == text->text_len) {
+		*asc = HTML_OBJECT (text)->ascent;
+		*dsc = HTML_OBJECT (text)->descent;
+
 		return text->word_width [text->words - 1];
+	}
 
 	next = HTML_OBJECT (slave)->next;
 	prev = HTML_OBJECT (slave)->prev;
@@ -198,14 +123,33 @@ calc_width (HTMLTextSlave *slave, HTMLPainter *painter)
 	    || (next && HTML_OBJECT_TYPE (next) == HTML_TYPE_TEXTSLAVE
 		&& slave->start_word == HTML_TEXT_SLAVE (next)->start_word)) {
 		gint line_offset = -1;
+		gint width;
 
-		return html_painter_calc_text_width (painter, html_text_slave_get_text (slave), slave->posLen,
-						     &line_offset, html_text_get_font_style (text),
-						     text->face);
-	} else
-		return get_words_width (text, painter, slave->start_word,
-					(next && HTML_OBJECT_TYPE (next) == HTML_TYPE_TEXTSLAVE
-					 ? HTML_TEXT_SLAVE (next)->start_word : text->words) - slave->start_word);
+		html_painter_calc_text_size (painter, html_text_slave_get_text (slave), slave->posLen, get_items (slave, painter), get_glyphs (slave, painter, line_offset),
+					     &line_offset, html_text_get_font_style (text),
+					     text->face, &width, asc, dsc);	
+		return width;
+	} else {
+		gint width;
+
+		width = get_words_width (text, painter, slave->start_word,
+					 (next && HTML_OBJECT_TYPE (next) == HTML_TYPE_TEXTSLAVE
+					  ? HTML_TEXT_SLAVE (next)->start_word : text->words) - slave->start_word);
+		*asc = HTML_OBJECT (text)->ascent;
+		*dsc = HTML_OBJECT (text)->descent;
+
+		return width;
+	}
+}
+
+inline static void
+glyphs_destroy (GList *glyphs)
+{
+	GList *l;
+
+	for (l = glyphs; l; l = l->next)
+		pango_glyph_string_free ((PangoGlyphString *) l->data);
+	g_list_free (glyphs);
 }
 
 static gint
@@ -214,14 +158,14 @@ get_offset_for_bounded_width (HTMLTextSlave *slave, HTMLPainter *painter, gint *
 	HTMLText *text = slave->owner;
 	gint upper = slave->posLen;
 	gint lower = 0;
-	gint len, width;
+	gint len, width, asc, dsc;
 	gint line_offset = -1;
 	gchar *sep, *str;
 	char *buffer = html_text_slave_get_text (slave);
 
 	len = (lower + upper) / 2;
-	width = html_painter_calc_text_width (painter, buffer, len, &line_offset,
-					      html_text_get_font_style (text), text->face);
+	html_painter_calc_text_size (painter, buffer, len, get_items (slave, painter), get_glyphs (slave, painter, line_offset), &line_offset,
+				     html_text_get_font_style (text), text->face, &width, &asc, &dsc);
 	while (lower < upper) {
 		if (width > max_width)
 			upper = len - 1;
@@ -229,8 +173,8 @@ get_offset_for_bounded_width (HTMLTextSlave *slave, HTMLPainter *painter, gint *
 			lower = len + 1;
 		len = (lower + upper) / 2;
 		line_offset = -1;
-		width = html_painter_calc_text_width (painter, buffer, len, &line_offset, 
-						      html_text_get_font_style (text), text->face);
+		html_painter_calc_text_size (painter, buffer, len, get_items (slave, painter), get_glyphs (slave, painter, line_offset), &line_offset, 
+					     html_text_get_font_style (text), text->face, &width, &asc, &dsc);
 	}
 
 	if (width > max_width && len > 1)
@@ -251,7 +195,7 @@ get_offset_for_bounded_width (HTMLTextSlave *slave, HTMLPainter *painter, gint *
 }
 
 static void
-slave_split_if_too_long (HTMLTextSlave *slave, HTMLPainter *painter, gint *width)
+slave_split_if_too_long (HTMLTextSlave *slave, HTMLPainter *painter, gint *width, gint *asc, gint *dsc)
 {
 	gint x, y;
 
@@ -263,7 +207,7 @@ slave_split_if_too_long (HTMLTextSlave *slave, HTMLPainter *painter, gint *width
 		pos = get_offset_for_bounded_width (slave, painter, &words, MAX_WIDGET_WIDTH - x);
 		if (pos > 0 && pos < slave->posLen) {
 			split (slave, pos, slave->start_word + words, NULL);
-			*width = MAX (1, calc_width (slave, painter));
+			*width = MAX (1, calc_width (slave, painter, asc, dsc));
 		}
 	}
 }
@@ -282,8 +226,14 @@ calc_size (HTMLObject *self, HTMLPainter *painter, GList **changed_objs)
 	owner = HTML_TEXT (slave->owner);
 	font_style = html_text_get_font_style (owner);
 
-	new_ascent = html_painter_calc_ascent (painter, font_style, owner->face);
-	new_descent = html_painter_calc_descent (painter, font_style, owner->face);
+	new_width = MAX (1, calc_width (slave, painter, &new_ascent, &new_descent));
+
+	next = HTML_OBJECT (slave)->next;
+	if ((slave->start_word == owner->words - 1
+	     && (!next || (HTML_IS_TEXT_SLAVE (next) && HTML_TEXT_SLAVE (next)->start_word == slave->start_word + 1)))
+	    && (HTML_IS_PLAIN_PAINTER (painter) || HTML_IS_GDK_PAINTER (painter))
+	    && new_width > HTML_OBJECT (owner)->max_width)
+		slave_split_if_too_long (slave, painter, &new_width, &new_ascent, &new_descent);
 
 	/* handle sub & super script */
 	if (font_style & GTK_HTML_FONT_STYLE_SUBSCRIPT || font_style & GTK_HTML_FONT_STYLE_SUPERSCRIPT) {
@@ -298,15 +248,6 @@ calc_size (HTMLObject *self, HTMLPainter *painter, GList **changed_objs)
 		}
 	}
 
-	new_width = MAX (1, calc_width (slave, painter));
-
-	next = HTML_OBJECT (slave)->next;
-	if ((slave->start_word == owner->words - 1
-	     && (!next || (HTML_IS_TEXT_SLAVE (next) && HTML_TEXT_SLAVE (next)->start_word == slave->start_word + 1)))
-	    && (HTML_IS_PLAIN_PAINTER (painter) || HTML_IS_GDK_PAINTER (painter))
-	    && new_width > HTML_OBJECT (owner)->max_width)
-		slave_split_if_too_long (slave, painter, &new_width);
-
 	changed = FALSE;
 
 	if (new_ascent != self->ascent) {
@@ -374,8 +315,10 @@ html_text_slave_get_line_offset (HTMLTextSlave *slave, gint line_offset, gint of
 		bol = HTML_TEXT_SLAVE (prev->next);
 		return html_text_text_line_length (html_text_slave_get_text (bol),
 						   0, offset - bol->posStart);
-	} else
-		return line_offset + html_text_text_line_length (slave->owner->text, line_offset, offset);
+	} else {
+		html_text_text_line_length (slave->owner->text, &line_offset, offset);
+		return line_offset;
+	}
 }
 
 static gint
@@ -462,15 +405,15 @@ hts_fit_line (HTMLObject *o, HTMLPainter *painter,
 	gchar *begin;
 	guint words = 0;
 
-	/* printf ("fit_line %d left: %d\n", firstRun, widthLeft); */
-
 	slave = HTML_TEXT_SLAVE (o);
 	text  = HTML_TEXT (slave->owner);
 
 	begin = html_text_slave_remove_leading_space (slave, painter, lineBegin);
 
+	/* printf ("fit_line %d left: %d lspacetext: \"%s\"\n", firstRun, widthLeft, begin); */
+
 	sep = begin;
-	while (sep
+	while (sep && *sep
 	       && widthLeft >= html_text_slave_nb_width (slave, painter, words + 1)) {
 		words ++;
 		lsep   = sep;
@@ -498,8 +441,11 @@ hts_fit_line (HTMLObject *o, HTMLPainter *painter,
 			rv = next_to_floating ? HTML_FIT_NONE : HTML_FIT_COMPLETE;
 		else {
 			words ++;
-			sep    = strchr (sep + (words > 1 ? 0 : 1), ' ');
-			pos    = sep ? g_utf8_pointer_to_offset (begin, sep) : g_utf8_strlen (begin, -1);
+
+			if (sep)
+				sep = strchr (sep + (words > 1 ? 0 : 1), ' ');
+
+			pos = sep ? g_utf8_pointer_to_offset (begin, sep) : g_utf8_strlen (begin, -1);
 		}
 	}
 
@@ -540,8 +486,9 @@ get_ys (HTMLText *text, HTMLPainter *p)
 	if (text->font_style & GTK_HTML_FONT_STYLE_SUBSCRIPT || text->font_style & GTK_HTML_FONT_STYLE_SUPERSCRIPT) {
 		gint height2;
 
-		height2 = (html_painter_calc_ascent (p, text->font_style, text->face)
-			   + html_painter_calc_descent (p, text->font_style, text->face)) >> 1;
+		height2 = (HTML_OBJECT (text)->ascent + HTML_OBJECT (text)->descent) / 2;
+		/* FIX2? (html_painter_calc_ascent (p, text->font_style, text->face)
+		   + html_painter_calc_descent (p, text->font_style, text->face)) >> 1; */
 		return (text->font_style & GTK_HTML_FONT_STYLE_SUBSCRIPT) ? height2 : -height2;
 			
 	} else 
@@ -565,22 +512,28 @@ draw_spell_errors (HTMLTextSlave *slave, HTMLPainter *p, gint tx, gint ty, gint
 		ma = MAX (se->off, slave->posStart);
 		mi = MIN (se->off + se->len, slave->posStart + slave->posLen);
 		if (ma < mi) {
+			GList *items, *glyphs;
 			guint off = ma - slave->posStart;
 			guint len = mi - ma;
-			gint lo;
+			gint lo, width, asc, dsc;
 
 			html_painter_set_pen (p, &html_colorset_get_color_allocated (p, HTMLSpellErrorColor)->color);
 			/* printf ("spell error: %s\n", html_text_get_text (slave->owner, off)); */
 			lo = line_offset;
 			
-			x_off += html_painter_calc_text_width (p, text,
-							       off - last_off, &line_offset,
-							       p->font_style,
-							       p->font_face);
+			glyphs = get_glyphs_part (slave, p, last_off, off - last_off, line_offset, &items);
+			html_painter_calc_text_size (p, text,
+						     off - last_off, items, glyphs, &line_offset,
+						     p->font_style,
+						     p->font_face, &width, &asc, &dsc);
+			glyphs_destroy (glyphs);
+			x_off += width;
 			text = g_utf8_offset_to_pointer (text, off - last_off);
+			glyphs = get_glyphs_part (slave, p, off, len, line_offset, &items);
 			x_off += html_painter_draw_spell_error (p, obj->x + tx + x_off,
 								obj->y + ty + get_ys (HTML_TEXT (slave->owner), p),
-								text, len);
+								text, len, items, glyphs);
+			glyphs_destroy (glyphs);
 			last_off = off + len;
 			line_offset += len;
 			text = g_utf8_offset_to_pointer (text, len);
@@ -591,6 +544,124 @@ draw_spell_errors (HTMLTextSlave *slave, HTMLPainter *p, gint tx, gint ty, gint
 	}
 }
 
+static GList *
+get_items (HTMLTextSlave *slave, HTMLPainter *painter)
+{
+	if (!slave->items) {
+		PangoItem *item;
+		gint start_offset = html_text_slave_get_text (slave) - slave->owner->text;
+
+		slave->items = html_text_get_items (slave->owner, painter);
+		if (slave->items) {
+			item = (PangoItem *) slave->items->data;
+
+			while (slave->items && start_offset >= item->offset + item->length) {
+				slave->items = slave->items->next;
+				item = (PangoItem *) slave->items->data;
+			}
+		}
+	}
+
+	return slave->items;
+}
+
+static GList *
+get_glyphs_part (HTMLTextSlave *slave, HTMLPainter *painter, guint offset, guint len, gint line_offset, GList **items)
+{
+	GList *glyphs = NULL;
+	const gchar *text;
+
+	*items = get_items (slave, painter);
+	if (items) {
+		PangoGlyphString *str;
+		PangoItem *item;
+		GList *il;
+		gint bytes, index, current_len, index_in_item;
+		const gchar *slave_text, *item_text;
+		gchar *translated_text, *heap;
+
+		/* printf ("get_glyphs_part %d,%d\n", offset, len); */
+
+		slave_text = html_text_slave_get_text (slave);
+		text = g_utf8_offset_to_pointer (slave_text, offset);
+		index = offset;
+		while (text - slave->owner->text >= ((PangoItem *) (*items)->data)->offset + ((PangoItem *) (*items)->data)->length) {
+			*items = (*items)->next;
+			/* printf ("skip item, new num_chars: %d\n", ((PangoItem *) (*items)->data)->num_chars); */
+		}
+		il = *items;
+		while (il && index - offset < len) {
+			item = (PangoItem *) il->data;
+			str = pango_glyph_string_new ();
+			item_text = slave->owner->text + item->offset;
+			index_in_item = g_utf8_pointer_to_offset (item_text, text);
+			current_len = MIN (item->num_chars - index_in_item, len - (index - offset));
+			/* printf ("iii: %d oii: %d\n", index_in_item, text - item_text);
+			   printf ("cl: %d offset: %d idx: %d\n", current_len, offset, index); */
+
+			bytes = g_utf8_offset_to_pointer (text, current_len) - text;
+			heap = NULL;
+			if (bytes > HTML_ALLOCA_MAX)
+				heap = translated_text = g_malloc (bytes);
+			else 
+				translated_text = alloca (bytes);
+			html_replace_tabs (text, translated_text, bytes);
+			pango_shape (translated_text, bytes, &item->analysis, str);
+			g_free (heap);
+
+			glyphs = g_list_prepend (glyphs, str);
+			text = g_utf8_offset_to_pointer (text, current_len);
+			index += current_len;
+			il = il->next;
+		}
+		glyphs = g_list_reverse (glyphs);
+	}
+
+	return glyphs;
+}
+
+static GList *
+get_glyphs (HTMLTextSlave *slave, HTMLPainter *painter, gint line_offset)
+{
+	if (!slave->glyphs) {
+		GList *items = get_items (slave, painter);
+		const gchar *text;
+
+		if (items) {
+			PangoGlyphString *str;
+			PangoItem *item;
+			GList *il = items;
+			gint bytes, index, len;
+			gchar *translated_text, *heap;
+
+			text = html_text_slave_get_text (slave);
+			index = 0;
+			while (il && index < slave->posLen) {
+				item = (PangoItem *) il->data;
+				str = pango_glyph_string_new ();
+				len = MIN (item->num_chars, slave->posLen - index);
+
+				bytes = g_utf8_offset_to_pointer (text, len) - text;
+				heap = NULL;
+				if (bytes > HTML_ALLOCA_MAX)
+					heap = translated_text = g_malloc (bytes);
+				else 
+					translated_text = alloca (bytes);
+				html_replace_tabs (text, translated_text, bytes);
+				pango_shape (translated_text, bytes, &item->analysis, str);
+				g_free (heap);
+
+				slave->glyphs = g_list_prepend (slave->glyphs, str);
+				text = g_utf8_offset_to_pointer (text, len);
+				index += len;
+				il = il->next;
+			}
+			slave->glyphs = g_list_reverse (slave->glyphs);
+		}
+	}
+	return slave->glyphs;
+}
+
 static void
 draw_normal (HTMLTextSlave *self,
 	     HTMLPainter *p,
@@ -600,18 +671,35 @@ draw_normal (HTMLTextSlave *self,
 	     gint tx, gint ty, gint line_offset)
 {
 	HTMLObject *obj;
+	HTMLText *text = self->owner;
+	gchar *str;
+	gint lo;
 
 	obj = HTML_OBJECT (self);
 
-	html_painter_set_font_style (p, font_style);
-	html_painter_set_font_face  (p, HTML_TEXT (self->owner)->face);
-	html_color_alloc (HTML_TEXT (self->owner)->color, p);
-	html_painter_set_pen (p, &HTML_TEXT (self->owner)->color->color);
-	html_painter_draw_text (p,
-				obj->x + tx, obj->y + ty + get_ys (HTML_TEXT (self->owner), p),
-				html_text_slave_get_text (self),
-				self->posLen,
-				html_text_slave_get_line_offset (self, line_offset, self->posStart, p));
+	str = html_text_slave_get_text (self);
+	if (*str) {
+		GList *glyphs, *items;
+
+		html_painter_set_font_style (p, font_style);
+		html_painter_set_font_face  (p, HTML_TEXT (self->owner)->face);
+		html_color_alloc (HTML_TEXT (self->owner)->color, p);
+		html_painter_set_pen (p, &HTML_TEXT (self->owner)->color->color);
+		lo = html_text_slave_get_line_offset (self, line_offset, self->posStart, p);
+
+		if (self->posStart > 0)
+			glyphs = get_glyphs_part (self, p, 0, self->posLen, lo, &items);
+		else {
+			items = get_items (self, p);
+			glyphs = get_glyphs (self, p, lo);
+		}
+
+		html_painter_draw_text (p, obj->x + tx, obj->y + ty + get_ys (text, p),
+					str, self->posLen, items, glyphs, lo);
+
+		if (self->posStart > 0)
+			glyphs_destroy (glyphs);
+	}
 }
 
 static void
@@ -624,8 +712,9 @@ draw_highlighted (HTMLTextSlave *slave,
 {
 	HTMLText *owner;
 	HTMLObject *obj;
+	GList *glyphs1, *glyphs2, *glyphs3, *items1, *items2, *items3;
 	guint start, end, len;
-	gint offset_width, text_width, lo, lo_start, lo_sel;
+	gint offset_width, text_width, lo, lo_start, lo_sel, asc, dsc;
 	const gchar *text;
 	
 	char *slave_begin;
@@ -648,15 +737,23 @@ draw_highlighted (HTMLTextSlave *slave,
 	highlight_begin = g_utf8_offset_to_pointer (slave_begin, start - slave->posStart);
 
 	lo_start = lo = html_text_slave_get_line_offset (slave, line_offset, slave->posStart, p);
-	offset_width = html_painter_calc_text_width (p, slave_begin,
-						     start - slave->posStart,
-						     &lo,
-						     font_style, HTML_TEXT (owner)->face);
+
+	if (start > slave->posStart) {
+		glyphs1 = get_glyphs_part (slave, p, 0, start - slave->posStart, lo_start, &items1);
+		html_painter_calc_text_size (p, slave_begin, start - slave->posStart, items1, glyphs1, &lo,
+					     font_style, HTML_TEXT (owner)->face, &offset_width, &asc, &dsc);
+	} else
+		offset_width = 0;
 	lo_sel = lo;
-	text_width = html_painter_calc_text_width (p, highlight_begin,
-						   len, &lo,
-						   font_style, HTML_TEXT (owner)->face);
-	/* printf ("s: %d l: %d - %d %d\n", start, len, offset_width, text_width); */
+
+	if (len) {
+		glyphs2 = get_glyphs_part (slave, p, start - slave->posStart, len, lo_sel, &items2);
+
+		html_painter_calc_text_size (p, highlight_begin, len, items2, glyphs2, &lo,
+					     font_style, HTML_TEXT (owner)->face, &text_width, &asc, &dsc);
+		/* printf ("s: %d l: %d - %d %d\n", start, len, offset_width, text_width); */
+	} else
+		text_width = 0;
 
 	html_painter_set_font_style (p, font_style);
 	html_painter_set_font_face  (p, HTML_TEXT (owner)->face);
@@ -670,33 +767,40 @@ draw_highlighted (HTMLTextSlave *slave,
 				text_width, obj->ascent + obj->descent);
 	html_painter_set_pen (p, &html_colorset_get_color_allocated
 			      (p, p->focus ? HTMLHighlightTextColor : HTMLHighlightTextNFColor)->color);
-	html_painter_draw_text (p, obj->x + tx + offset_width, 
-				obj->y + ty + get_ys (HTML_TEXT (slave->owner), p),
-				highlight_begin, len,
-				lo_sel);
 
+	if (len) {
+		html_painter_draw_text (p, obj->x + tx + offset_width, 
+					obj->y + ty + get_ys (HTML_TEXT (slave->owner), p),
+					highlight_begin, len, items2, glyphs2,
+					lo_sel);
+		if (glyphs2)
+			glyphs_destroy (glyphs2);
+	}
 	/* Draw the non-highlighted part.  */
-
 	html_painter_set_pen (p, &HTML_TEXT (owner)->color->color);
 
 	/* 1. Draw the leftmost non-highlighted part, if any.  */
-
-	if (start > slave->posStart)
+	if (start > slave->posStart) {
 		html_painter_draw_text (p,
 					obj->x + tx, obj->y + ty + get_ys (HTML_TEXT (slave->owner), p),
 					slave_begin,
-					start - slave->posStart,
+					start - slave->posStart, items1, glyphs1,
 					lo_start);
+		if (glyphs1)
+			glyphs_destroy (glyphs1);
+	}
 
 	/* 2. Draw the rightmost non-highlighted part, if any.  */
-
-	if (end < slave->posStart + slave->posLen)
+	if (end < slave->posStart + slave->posLen) {
+		glyphs3 = get_glyphs_part (slave, p, start + len - slave->posStart, slave->posLen - start - len + slave->posStart, lo, &items3);
 		html_painter_draw_text (p,
 					obj->x + tx + offset_width + text_width,
 					obj->y + ty + get_ys (HTML_TEXT (slave->owner), p),
 					g_utf8_offset_to_pointer (highlight_begin, end - start),
-					slave->posStart + slave->posLen - end,
-					lo);
+					slave->posStart + slave->posLen - end, items3, glyphs3, lo);
+		if (glyphs3)
+			glyphs_destroy (glyphs3);
+	}
 }
 
 static void
@@ -711,16 +815,15 @@ draw (HTMLObject *o,
 	HTMLText *ownertext;
 	GtkHTMLFontStyle font_style;
 	guint end;
-	ArtIRect paint;
+	GdkRectangle paint;
 	gint line_offset;
 
 	/* printf ("slave draw %p\n", o); */
 
-	html_object_calc_intersection (o, &paint, x, y, width, height);
-	if (art_irect_empty (&paint))
+	textslave = HTML_TEXT_SLAVE (o);
+	if (!html_object_intersect (o, &paint, x, y, width, height) || textslave->posLen == 0)
 		return;
 	
-	textslave = HTML_TEXT_SLAVE (o);
 	owner = textslave->owner;
 	ownertext = HTML_TEXT (owner);
 	font_style = html_text_get_font_style (ownertext);
@@ -802,9 +905,13 @@ get_offset_for_pointer (HTMLTextSlave *slave, HTMLPainter *painter, gint x, gint
 	}
 
 	for ( ; i <= slave->posLen; i++) {
+		GList *items;
+		GList *glyphs;
+		gint asc, dsc;
+
 		lo = line_offset;
-		width = html_painter_calc_text_width (painter, text, 
-						      i, &lo, font_style, owner->face);
+		glyphs = get_glyphs_part (slave, painter, 0, i, lo, &items);
+		html_painter_calc_text_size (painter, text, i, items, glyphs, &lo, font_style, owner->face, &width, &asc, &dsc);
 
 		if ((width + prev_width) / 2 >= x)
 			return i - 1;
@@ -838,6 +945,19 @@ check_point (HTMLObject *self,
 	return NULL;
 }
 
+static void
+destroy (HTMLObject *obj)
+{
+	HTMLTextSlave *slave = HTML_TEXT_SLAVE (obj);
+
+	if (slave->glyphs) {
+		glyphs_destroy (slave->glyphs);
+		slave->glyphs = NULL;
+	}
+
+	HTML_OBJECT_CLASS (parent_class)->destroy (obj);
+}
+
 void
 html_text_slave_type_init (void)
 {
@@ -857,6 +977,7 @@ html_text_slave_class_init (HTMLTextSlaveClass *klass,
 
 	object_class->select_range = select_range;
 	object_class->copy = copy;
+	object_class->destroy = destroy;
 	object_class->draw = draw;
 	object_class->calc_size = calc_size;
 	object_class->fit_line = hts_fit_line;
@@ -893,7 +1014,10 @@ html_text_slave_init (HTMLTextSlave *slave,
 	slave->start_word = start_word;
 	slave->owner      = owner;
 	slave->charStart  = NULL;
-	/* text slaves has always min_width 0 */
+	slave->items      = NULL;
+	slave->glyphs     = NULL;
+
+	/* text slaves have always min_width 0 */
 	object->min_width = 0;
 	object->change   &= ~HTML_CHANGE_MIN_WIDTH;
 }
